/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package distributions;

import java.math.BigDecimal;
import java.util.Arrays;
import java.util.Set;
import java.util.HashSet;
import java.util.Map;
import java.util.LinkedHashMap;
import java.lang.Math;

import support.Contexter;
import support.Logger;

/**
 *
 * @author  
 * @author  
 */
public class JointDistribution {

    protected static final Contexter contexter = Contexter.getInstance();
    private static final Logger logger = Logger.getInstance();

    /**
     * set of labels (identifiers) on secrets of the system
     */
    private Set<String> secretLabels = null;

    /**
     * set of labels (identifiers) on observables of the system
     */
    private Set<String> observableLabels = null;

    /**
     * map that returns the joint probability of given secret and observable
     */
    private Map<String, Map<String, BigDecimal>> jointProbMap;

    /**
     * Constructs the empty distribution.
     */
    public JointDistribution() {
        this.secretLabels = new HashSet<String>();
        this.observableLabels = new HashSet<String>();
        this.jointProbMap = new LinkedHashMap<>();
    }

    /**
     * Constructs a single distribution with initial values.
     *
     * @param secretLabels set of labels (identifiers) on secrets of the system
     * @param observableLabels set of labels (identifiers) on observables of the
     * system
     * @param jointProbMap hashmap that represents a single distribution
     */
    public JointDistribution(Set<String> secretLabels, Set<String> observableLabels, Map<String, Map<String, BigDecimal>> jointProbMap) {
        this.secretLabels = new HashSet<String>();
        this.secretLabels = secretLabels;
        this.observableLabels = new HashSet<String>();
        this.observableLabels = observableLabels;
        this.jointProbMap = new LinkedHashMap<>();
        this.jointProbMap = jointProbMap;
    }

    /**
     * Constructs a joint distribution with initial values.
     *
     * @param secretLabels array of labels (identifiers) on secrets of the
     * system
     * @param observableLabels array of labels (identifiers) on observables of
     * the system
     * @param jointProbMap hashmap that represents a joint distribution
     */
    public JointDistribution(String[] secretLabels, String[] observableLabels, Map<String, Map<String, BigDecimal>> jointProbMap) {
        this.secretLabels = new HashSet<>();
        this.secretLabels.addAll(Arrays.asList(secretLabels));
        this.observableLabels = new HashSet<>();
        this.observableLabels.addAll(Arrays.asList(observableLabels));
        this.jointProbMap = new LinkedHashMap<>();
        this.jointProbMap.putAll(jointProbMap);
    }

    /**
     * Constructor that copies this joint distribution.
     *
     * @param jointDist joint distribution
     */
    public JointDistribution(JointDistribution jointDist) {
        if (jointDist != null) {
            this.secretLabels = new HashSet<>();
            this.secretLabels.addAll(jointDist.secretLabels);
            this.observableLabels = new HashSet<>();
            this.observableLabels.addAll(jointDist.observableLabels);
            this.jointProbMap = new LinkedHashMap<>();
            this.jointProbMap.putAll(jointDist.jointProbMap);
        }
    }

	////////////////////////////////////////////////////////////////
    // Methods for taking components of this joint distributions
    /**
     * Returns the set of the secret labels of this joint distribution.
     *
     * @return set of the secret labels of this joint distribution
     */
    public Set<String> getSecretLabels() {
        if (this.secretLabels == null) {
            log("ERROR(JD88): Cannot get secret labels. This joint distribution is not initialized. secretLabels is null.", 1);
            return null;
        }
        return this.secretLabels;
    }

    /**
     * Computes the size of the secret labels of this joint distribution.
     *
     * @return the size of the secret labels of this joint distribution
     */
    public int getSizeOfDomainSecrets() {
        if (this.secretLabels == null) {
            log("ERROR(JD100): Cannot get the number of secrets. This joint distribution is not initialized. secretLabels is null.", 1);
            return -1;
        }
        return this.secretLabels.size();
    }

    /**
     * Returns the set of the observable labels of this joint distribution.
     *
     * @return set of the observable labels of this joint distribution
     */
    public Set<String> getObservableLabels() {
        if (this.observableLabels == null) {
            log("ERROR(JD113): Cannot get observable labels. This joint distribution is not initialized. observableLabels is null.", 1);
            return null;
        }
        return this.observableLabels;
    }

    /**
     * Computes the size of the set of observables of this joint distribution.
     *
     * @return the size of the set of observables of this joint distribution
     */
    public int getSizeOfDomainObservables() {
        if (this.observableLabels == null) {
            log("ERROR(JD125): Cannot get the number of observables. This joint distribution is not initialized. observableLabels is null.", 1);
            return -1;
        }
        return this.observableLabels.size();
    }

    /**
     * Returns the mapping from a pair of secret and observable to their joint
     * probability
     *
     * @return the mapping from a pair of secret and observable to their joint
     * probability
     */
    public Map<String, Map<String, BigDecimal>> getJointProbMap() {
        if (this.jointProbMap == null) {
            log("ERROR(JD134): Cannot get the joint probability matrix map. This joint distribution is not initialized. jointProbMap is null.", 1);
            System.exit(1);
        }
        return this.jointProbMap;
    }

	////////////////////////////////////////////////////////////////
    // Methods for operations on probability distributions
    /**
     * Computes the probability of having the given keys. (Returns 0 if the keys
     * are not included in the domains.)
     *
     * @param keySecretLabel key for the domain of secrets
     * @param keyObservableLabel key for the domain of observables
     * @return the joint probability of having the given keys
     */
    public BigDecimal getProbability(String keySecretLabel, String keyObservableLabel) {
        if (this.jointProbMap == null) {
            log("ERROR(JD156): Cannot get the joint probability matrix map. This joint distribution is not initialized. jointProbMap is null.", 1);
            System.exit(1);
        }
        Map<String, BigDecimal> observableProbMap = this.jointProbMap.get(keySecretLabel);
        if (observableProbMap != null) {
            BigDecimal prob = observableProbMap.get(keyObservableLabel);
            if (prob != null) {
                return prob;
            }
        }
        return BigDecimal.ZERO;
    }

    /**
     * Sets the probability of having the given keys.
     *
     * @param keySecretLabel key for the domain of secrets
     * @param keyObservableLabel key for the domain of observables
     * @param probability the joint probability of having the given keys
     */
    synchronized public void setProbability(String keySecretLabel, String keyObservableLabel, BigDecimal probability) {
        if (this.jointProbMap == null) {
            log("ERROR(JD178): Cannot get the joint probability matrix map. This joint distribution is not initialized. jointProbMap is null.", 1);
            System.exit(1);
        }
        if (probability == null) {
            log("ERROR(JD182): The given probability is null.", 1);
            System.exit(1);
        }
        if (!this.jointProbMap.containsKey(keySecretLabel)) {
            Map<String, BigDecimal> observableProbMap = new LinkedHashMap<>();
            observableProbMap.put(keyObservableLabel, probability);
            this.jointProbMap.put(keySecretLabel, observableProbMap);
        } else {
            Map<String, BigDecimal> observableProbMap = this.jointProbMap.get(keySecretLabel);
            observableProbMap.put(keyObservableLabel, probability);
            this.jointProbMap.put(keySecretLabel, observableProbMap);
        }
    }

    /**
     * Adds a given joint (sub-)distribution to this joint (sub-)distribution.
     *
     * @param jointDist given joint (sub-)distribution
     */
    synchronized public void add(JointDistribution jointDist) {
        logger.changeVerbosity("JD_ADD");
        if (jointDist == null) {
            log("CAUTION(JD216): The added joint distribution is not initialized.", 1);
        } else if (jointDist.getJointProbMap() == null) {
            log("CAUTION(JD218): Cannot get the joint probability matrix map. The given joint distribution is not initialized and failed to be added to this distribution.", 3);
        } else if (this.jointProbMap == null) {
            log("CAUTION(JD220): Cannot get the joint probability matrix map. This joint distribution is not initialized.", 3);
            if (this.secretLabels == null) {
                this.secretLabels = new HashSet<>();
            }
            this.secretLabels.addAll(jointDist.getSecretLabels());
            if (this.observableLabels == null) {
                this.observableLabels = new HashSet<>();
            }
            this.observableLabels.addAll(jointDist.getObservableLabels());
            this.jointProbMap = jointDist.getJointProbMap();
        } else {
            this.secretLabels.addAll(jointDist.getSecretLabels());
            this.observableLabels.addAll(jointDist.getObservableLabels());
            Set<String> givenSecretLabels = jointDist.getSecretLabels();
            Map<String, Map<String, BigDecimal>> givenJointProbMap = jointDist.getJointProbMap();
            for (String secretLabel : givenSecretLabels) {
                log("INFO(JD237): secretLabel = " + secretLabel, 4);
                if (givenJointProbMap.get(secretLabel) != null) {
                    for (String observableLabel : givenJointProbMap.get(secretLabel).keySet()) {
                        if (!this.jointProbMap.containsKey(secretLabel)) {
                            Map<String, BigDecimal> observableProbMap = new LinkedHashMap<>();
                            BigDecimal givenProb = givenJointProbMap.get(secretLabel).get(observableLabel);
                            log("INFO(JD243): observableLabel = " + observableLabel + " givenProb = " + givenProb,4);
                            observableProbMap.put(observableLabel, givenProb);
                            this.jointProbMap.put(secretLabel, observableProbMap);
                            log("INFO(JD246): putProb = " + this.jointProbMap.get(secretLabel).get(observableLabel), 4);
                        } else if (!this.jointProbMap.get(secretLabel).containsKey(observableLabel)) {
                            Map<String, BigDecimal> observableProbMap = this.jointProbMap.get(secretLabel);
                            BigDecimal givenProb = givenJointProbMap.get(secretLabel).get(observableLabel);
                            log("INFO(JD250): observableLabel = " + observableLabel + " givenProb = " + givenProb, 4);
                            observableProbMap.put(observableLabel, givenProb);
                            this.jointProbMap.put(secretLabel, observableProbMap);
                            log("INFO(JD253): putProb = " + this.jointProbMap.get(secretLabel).get(observableLabel), 4);
                        } else {
                            BigDecimal givenProb = givenJointProbMap.get(secretLabel).get(observableLabel);
                            BigDecimal thisProb = this.jointProbMap.get(secretLabel).get(observableLabel);
                            BigDecimal sumProb = givenProb.add(thisProb);
                            log("INFO(JD258): observableLabel = " + observableLabel + " sumProb = " + sumProb + " (= " + givenProb + "+" + thisProb + ")", 4);
                            this.jointProbMap.get(secretLabel).put(observableLabel, sumProb);
                        }
                    }
                }
            }
        }
        logger.restoreVerbosity();
    }
    //TODO: Add more methods for synchronized operations on distributions

    ////////////////////////////////////////////////////////////////
    // Methods for computing projection
    /**
     * Computes the distribution on secrets from the joint distribution.
     *
     * @return the distribution on secrets
     */
    public SingleDistribution getProjectionSecrets() {
        if (this.secretLabels == null) {
            log("ERROR(JD263): Cannot get the secret domain. This joint distribution is not initialized. secretLabels is null.", 1);
            //System.exit(1);
        }
        if (this.observableLabels == null) {
            log("ERROR(JD267): Cannot get the observable domain. This joint distribution is not initialized. observableLabels is null.", 1);
            //System.exit(1);
        }
        Map<String, BigDecimal> probMap = new LinkedHashMap<>();
        for (String secretLabel : this.secretLabels) {
            BigDecimal prob = BigDecimal.ZERO;
            for (String observableLabel : this.observableLabels) {
                if (this.jointProbMap.get(secretLabel) != null) {
                    BigDecimal jointProb = this.jointProbMap.get(secretLabel).get(observableLabel);
                    if (jointProb != null) {
                        prob = prob.add(jointProb);
                    }
                }
            }
            probMap.put(secretLabel, prob);
        }
        SingleDistribution dist = new SingleDistribution(secretLabels, probMap);
        return dist;
    }

    /**
     * Computes the distribution on observables from the joint distribution.
     *
     * @return the distribution on observables
     */
    public SingleDistribution getProjectionObservables() {
        if (this.secretLabels == null) {
            log("ERROR(JD291): Cannot get the secret domain. This joint distribution is not initialized. secretLabels is null.", 1);
            //System.exit(1);
        }
        if (this.observableLabels == null) {
            log("ERROR(JD295): Cannot get the observable domain. This joint distribution is not initialized. observableLabels is null.", 1);
            //System.exit(1);
        }
        if (this.secretLabels == null || this.secretLabels == null) {
            return null;
        } else {
            Map<String, BigDecimal> probMap = new LinkedHashMap<>();
            for (String observableLabel : this.observableLabels) {
                BigDecimal prob = BigDecimal.ZERO;
                for (String secretLabel : this.secretLabels) {
                    if (this.jointProbMap.get(secretLabel) != null) {
                        BigDecimal jointProb = this.jointProbMap.get(secretLabel).get(observableLabel);
                        if (jointProb != null) {
                            prob = prob.add(jointProb);
                        }
                    }
                }
                probMap.put(observableLabel, prob);
            }
            SingleDistribution dist = new SingleDistribution(observableLabels, probMap);
            return dist;
        }
    }

    ////////////////////////////////////////////////////////////////
    // Methods for obtaining the sub-distribution representing
    // a row/column of this joint distribution
    /**
     * Computes the sub-distribution over observables
     * that is the specified row of this joint distribution.
     *
     * @param secretLabel secret label
     * @return specified row of this joint distribution
     */
    public SingleDistribution getRow(String secretLabel) {
        if (!this.jointProbMap.containsKey(secretLabel)) {
            log("ERROR(JD362): This joint distribution does not contain the secret label "+secretLabel+".", 1);
        	return null;
        }
        if (this.observableLabels == null) {
            log("ERROR(JD366): Cannot get the observable domain. This joint distribution is not initialized. observableLabels is null.", 1);
            System.exit(1);
            return null;
        } else {
            Map<String, BigDecimal> probMap = new LinkedHashMap<>();
            for (String observableLabel : this.observableLabels) {
                if (this.jointProbMap.get(secretLabel) != null) {
                    BigDecimal jointProb = this.jointProbMap.get(secretLabel).get(observableLabel);
                    if (jointProb != null) {
                        probMap.put(observableLabel, jointProb);
                    }
                }
            }
        	SingleDistribution dist = new SingleDistribution(observableLabels, probMap);
            return dist;
        }
    }
    
    /**
     * Computes the sub-distribution over secrets
     * that is the specified column of this joint distribution.
     *
     * @param observableLabel observable label
     * @return specified row of this joint distribution
     */
    public SingleDistribution getColumn(String observableLabel) {
        if (!this.jointProbMap.containsValue(observableLabel)) {
            log("ERROR(JD393): This joint distribution does not contain the observable label "+observableLabel+".", 1);
        	return null;
        }
        if (this.secretLabels == null) {
            log("ERROR(JD397): Cannot get the secret domain. This joint distribution is not initialized. secretLabels is null.", 1);
            System.exit(1);
            return null;
        } else {
            Map<String, BigDecimal> probMap = new LinkedHashMap<>();
            for (String secretLabel : this.secretLabels) {
                if (this.jointProbMap.get(observableLabel) != null) {
                    BigDecimal jointProb = this.jointProbMap.get(secretLabel).get(observableLabel);
                    if (jointProb != null) {
                        probMap.put(secretLabel, jointProb);
                    }
                }
            }
        	SingleDistribution dist = new SingleDistribution(secretLabels, probMap);
            return dist;
        }
    }
    
    ////////////////////////////////////////////////////////////////
    // Other methods
    /**
     * Computes the summation of all probabilities in this joint
     * (sub-)distribution
     *
     * @return the summation of all probabilities in this joint
     * (sub-)distribution
     */
    public BigDecimal getSumOfProbabilities() {
        if (this.secretLabels == null || this.observableLabels == null) {
            log("ERROR(JD426): This joint distribution is not initialized.", 1);
            System.exit(1);
        }
        BigDecimal sum = BigDecimal.ZERO;
        for (String secretLabel : this.secretLabels) {
            for (String observableLabel : this.observableLabels) {
                Map<String, BigDecimal> observableProbMap = this.jointProbMap.get(secretLabel);
                if (observableProbMap != null) {
                    BigDecimal prob = observableProbMap.get(observableLabel);
                    if (prob != null) {
                        sum = sum.add(prob);
                    }
                }
            }
        }
        return sum;
    }
    
    /**
     * Computes properties of this distribution (the size of the matrix,
     * the number of entries with non-zero probabilities and minmum sample size).
     * 
     * @return array of properties of this distribution
     */
    public int[] getSizeProperties() {
        logger.changeVerbosity("JD_SIZEPROP");
    	// get the matrix size
    	int sizeOfSecretDomain = secretLabels.size();
    	int sizeOfObservableDomain = observableLabels.size();
        log("Matrix Size = ("+sizeOfSecretDomain+", "+sizeOfObservableDomain+")", 3);
    	// get the number of entries with non-zero probabilities
    	int countNonZeroFull = 0;
    	int countNonZeroAbs = 0;
    	boolean countAbs = true;
        for (String secretLabel : this.secretLabels) {
            for (String observableLabel : this.observableLabels) {
                Map<String, BigDecimal> observableProbMap = this.jointProbMap.get(secretLabel);
                if (observableProbMap != null) {
                    BigDecimal prob = observableProbMap.get(observableLabel);
                    if(prob != null && prob.compareTo(BigDecimal.ZERO) > 0) {
                        countNonZeroFull++;
                        if(countAbs) {
                            countNonZeroAbs++;
                        }
                    }
                }
            }
            countAbs = false;
        }
        log("#EntryWithNon-zeroProbability for full = "+countNonZeroFull, 3);
        log("#EntryWithNon-zeroProbability for abs  = "+countNonZeroAbs, 3);
    	// get the minimum sample size
        int minSampleSizeFull = Math.max(0,(countNonZeroFull-1) * 4);
        int minSampleSizeAbs  = Math.max(0,(countNonZeroAbs-1) * 4);
        log("Minimum Sample Size for full-sim (roughly) = "+minSampleSizeFull, 3);
        log("Minimum Sample Size for abs-sim (roughly)  = "+minSampleSizeAbs, 3);
    	// store the results
        int[] res = new int[6];
        res[0] = sizeOfSecretDomain;
        res[1] = sizeOfObservableDomain;
        res[2] = countNonZeroFull;
        res[3] = minSampleSizeFull;
        res[4] = countNonZeroAbs;
        res[5] = minSampleSizeAbs;
        logger.restoreVerbosity();
        return res;
    }

    /**
     * Print this joint distribution.
     */
    public void print() {
        print(3);
    }

    /**
     * Print this joint distribution.
     */
    public void print(int level) {
        if (Logger.getInstance().getVerbosity()<level){
            return;
        }
        log("{", level);
        printLabels(level);
        printProbs(level);
        log("}", level);
    }

    /**
     * Print this joint distribution.
     */
    public void print(String string) {
        print(string, 3);
    }

    /**
     * Print this joint distribution.
     */
    public void print(String string, int level) {
        if (Logger.getInstance().getVerbosity()<level){
            return;
        }
        log(string + " = {", level);
        printLabels(level);
        printProbs(level);
        log("}", level);
    }

    private void printLabels(int level) {
        if (Logger.getInstance().getVerbosity()<level){
            return;
        }
        log("  secret labels = {", level);
        if (this.secretLabels != null) {
            for (String secretLabel : this.secretLabels) {
                log("    " + secretLabel, level);
            }
        }
        log("  }", level);
        log("  observable labels = {", level);
        if (this.observableLabels != null) {
            for (String observableLabel : this.observableLabels) {
                log("    " + observableLabel, level);
            }
        }
        log("  }", level);
    }

    private void printProbs(int level) {
        if (Logger.getInstance().getVerbosity()<level){
            return;
        }
        if (this.secretLabels != null && this.observableLabels != null) {
            for (String secretLabel : this.secretLabels) {
                for (String observableLabel : this.observableLabels) {
                    Map<String, BigDecimal> observableProbMap = this.jointProbMap.get(secretLabel);
                    if (observableProbMap != null) {
                        BigDecimal prob = observableProbMap.get(observableLabel);
                        if (prob != null) {
                            log("  Pr[" + secretLabel + ", " + observableLabel + "] = " + prob, level);
                        } else {
                            //log("  Pr["+secretLabel+", "+observableLabel+"] = null",level);
                        }
                    }
                }
            }
        }
    }

    private static void log(Object aObject, int level) {
        Logger loggy = Logger.getInstance();
        loggy.log(String.valueOf(aObject), level);
    }
}
