/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package distributions;

import java.math.BigInteger;
import java.util.Arrays;
import java.util.Set;
import java.util.HashSet;
import java.util.Map;
import java.util.LinkedHashMap;
import hyleak.Interval;
import hyleak.PrivateVariable;
import hyleak.SimulationSignature;
import hyleak.VariableList;

import support.Contexter;
import support.Logger;
import theory.BigIntegerCartesianIterator;

/**
 * @author  
 */
public class JointFrequencyDistribution {
    protected static final Contexter contexter = Contexter.getInstance();
    private static final Logger logger = Logger.getInstance();

	/**
	 * set of labels (identifiers) on secrets of the system
	 */
	private Set<String> secretLabels;
	
	/**
	 * set of labels (identifiers) on observables of the system
	 */
	private Set<String> observableLabels;

	/**
	 * map that returns the joint frquency of given secret and observable
	 */
	private Map<String,Map<String,Integer>> jointFreqMap;

	/**
	 * Constructs the empty distribution.
	 */
	public JointFrequencyDistribution() {
        this.secretLabels = new HashSet<String>();
        this.observableLabels = new HashSet<String>();
        this.jointFreqMap = new LinkedHashMap<String,Map<String,Integer>>();
    }

	/**
	 * Constructs a joint frquency distribution with initial values.
	 * 
	 * @param secretLabels array of labels (identifiers) on secrets of the system
	 * @param observableLabels array of labels (identifiers) on observables of the system
	 * @param jointFreqMap hashmap that represents a joint frquency distribution
	 */
	public JointFrequencyDistribution(String[] secretLabels, String[] observableLabels, Map<String,Map<String,Integer>> jointFreqMap) {
		this.secretLabels = new HashSet<String>();
		this.secretLabels.addAll(Arrays.asList(secretLabels));
        this.observableLabels = new HashSet<String>();
        this.observableLabels.addAll(Arrays.asList(observableLabels));
        this.jointFreqMap = new LinkedHashMap<String,Map<String,Integer>>();
		this.jointFreqMap.putAll(jointFreqMap);
	}
        
        public JointFrequencyDistribution(Set<String> secretLabels, Set<String> observableLabels, Map<String,Map<String,Integer>> jointFreqMap) {
		this.secretLabels = new HashSet<String>();
		this.secretLabels.addAll(secretLabels);
        this.observableLabels = new HashSet<String>();
        this.observableLabels.addAll(observableLabels);
        this.jointFreqMap = new LinkedHashMap<String,Map<String,Integer>>();
		this.jointFreqMap.putAll(jointFreqMap);
	}

	/**
	 * Constructor that copies this joint frquency distribution.
	 * 
	 * @param jointDist joint frquency distribution
	 */
	public JointFrequencyDistribution(JointFrequencyDistribution jointDist) {
		if(jointDist != null) {
			this.secretLabels = new HashSet<String>();
			this.secretLabels.addAll(jointDist.secretLabels);
	        this.observableLabels = new HashSet<String>();
			this.observableLabels.addAll(jointDist.observableLabels);
	        this.jointFreqMap = new LinkedHashMap<String,Map<String,Integer>>();
			this.jointFreqMap.putAll(jointDist.jointFreqMap);
		}
	}

	////////////////////////////////////////////////////////////////
	// Methods for taking components of this joint frquency distributions
	/**
	 * Returns the set of the secret labels of this joint frquency distribution.
	 * 
	 * @return set of the secret labels of this joint frquency distribution
	 */
	public Set<String> getSecretLabels() {
    	if(this.secretLabels == null) {
            log("ERROR(JFD81): Cannot get secret labels. This joint frquency distribution is not initialized. secretLabels is null.",1);
            return null;
    	}
    	/*
        String[] toreturn = new String[secretLabels.size()];
        int count = 0;
        for(String str : secretLabels ) {
            toreturn[count++] = str;
        }
        return toreturn;
        */
		return this.secretLabels;
	}
	
	/**
	 * Computes the size of the secret labels of this joint frquency distribution.
	 * @return the size of the secret labels of this joint frquency distribution
	 */
	public int getSizeOfDomainSecrets() {
    	if(this.secretLabels == null) {
            log("ERROR(JFD93): Cannot get the number of secrets. This joint frquency distribution is not initialized. secretLabels is null.",1);
            return -1;
    	}
		return this.secretLabels.size();
	}

	/**
	 * Returns the set of the observable labels of this joint frquency distribution.
	 * 
	 * @return set of the observable labels of this joint frquency distribution
	 */
	public Set<String> getObservableLabels() {
    	if(this.observableLabels == null) {
            log("ERROR(JFD106): Cannot get observable labels. This joint frquency distribution is not initialized. observableLabels is null.",1);
            return null;
    	}
    	/*
        String[] toreturn = new String[observableLabels.size()];
        int count = 0;
        for(String str : observableLabels ) {
            toreturn[count++] = str;
        }
        return toreturn;
        */
		return this.observableLabels;
	}
	
	/**
	 * Computes the size of the set of observables of this joint frquency distribution.
	 * @return the size of the set of observables of this joint frquency distribution
	 */
	public int getSizeOfDomainObservables() {
    	if(this.observableLabels == null) {
            log("ERROR(JFD118): Cannot get the number of observables. This joint frquency distribution is not initialized. observableLabels is null.",1);
            return -1;
    	}
		return this.observableLabels.size();
	}
	
	/**
	 * Returns the mapping from a pair of secret and observable to their joint frquency
	 * 
	 * @return the mapping from a pair of secret and observable to their joint frquency
	 */
	public Map<String,Map<String,Integer>> getJointFreqMap() {
    	if(this.jointFreqMap == null) {
            log("ERROR(JFD131): Cannot get the joint frquency matrix map. This joint frquency distribution is not initialized. jointFreqMap is null.",1);
            System.exit(1);
    	}
		return this.jointFreqMap;
	}

	////////////////////////////////////////////////////////////////
	// Methods for operations on frquency distributions
	/**
	 * Computes the frquency of having the given keys.
	 * (Returns 0 if the keys are not included in the domains.)
	 * 
	 * @param keySecretLabel key for the domain of secrets
	 * @param keyObservableLabel key for the domain of observables
	 * @return the joint frquency of having the given keys
	 */
    public int getFrequency(String keySecretLabel, String keyObservableLabel) {
    	if(this.jointFreqMap == null) {
            log("ERROR(JFD149): Cannot get the joint frquency matrix map. This joint frquency distribution is not initialized. jointFreqMap is null.",1);
            System.exit(1);
    	}
    	if(!this.jointFreqMap.containsKey(keySecretLabel)) {
            return 0;
        } else {
             Map<String,Integer> observableFreqMap = this.jointFreqMap.get(keySecretLabel);
             if(!observableFreqMap.containsKey(keyObservableLabel)) {
                 return 0;
             } else {
                 return Math.max(0,observableFreqMap.get(keyObservableLabel));
             }
        }
    }
	
	/**
	 * Sets the frquency of having the given keys.
	 * 
	 * @param keySecretLabel key for the domain of secrets
	 * @param keyObservableLabel key for the domain of observables
     * @param frquency the joint frquency of having the given keys
	 */
    public void setFrequency(String keySecretLabel, String keyObservableLabel, int frquency) {
    	if(this.jointFreqMap == null) {
            log("ERROR(JFD171): Cannot get the joint frequency matrix map. This joint frquency distribution is not initialized. jointFreqMap is null.",1);
            System.exit(1);
    	}
    	if(frquency < 0) {
            log("ERROR(JFD175): The given frquency is negative.",1);
            System.exit(1);
    	}
        if(!this.secretLabels.contains(keySecretLabel)) {
            this.secretLabels.add(keySecretLabel);
            log("Added new secret label "+keySecretLabel,3);
        }
        if(!this.observableLabels.contains(keyObservableLabel)) {
            this.observableLabels.add(keyObservableLabel);
            log("Added new observable label "+keyObservableLabel,3);
        }
        
        if(!this.jointFreqMap.containsKey(keySecretLabel)) {
    	    Map<String,Integer> observableFreqMap = new LinkedHashMap<String,Integer>();
    	    observableFreqMap.put(keyObservableLabel, frquency);
            this.jointFreqMap.put(keySecretLabel, observableFreqMap);
        } else {
        	Map<String,Integer> observableFreqMap = this.jointFreqMap.get(keySecretLabel);
        	observableFreqMap.put(keyObservableLabel, frquency);
            this.jointFreqMap.put(keySecretLabel, observableFreqMap);
        }
    }

	/**
	 * Adds a given joint (sub-)distribution to this joint (sub-)distribution.
	 * 
     * @param jointDist given joint (sub-)distribution
	 */
    synchronized public void add(JointFrequencyDistribution jointDist) {
    	if(jointDist == null) {
            log("ERROR(JFD225): The added joint frquency distribution is not initialized.",1);
    	} else if(jointDist.getJointFreqMap() == null) {
            log("CAUTION(JD227): Cannot get the joint frequency probability matrix map. The given joint frequency distribution is not initialized and failed to be added to this distribution.",2);
    	} else if(this.jointFreqMap == null) {
            log("ERROR(JFD229): Cannot get the joint frquency matrix map. This joint frquency distribution is not initialized.",2);
    		this.secretLabels.addAll(jointDist.getSecretLabels());
    		this.observableLabels.addAll(jointDist.getObservableLabels());
            this.jointFreqMap=jointDist.getJointFreqMap();
    	} else {
    		this.secretLabels.addAll(jointDist.getSecretLabels());
    		this.observableLabels.addAll(jointDist.getObservableLabels());
        	Set<String> givenSecretLabels = jointDist.getSecretLabels();
        	Map<String,Map<String,Integer>> givenJointFreqMap = jointDist.getJointFreqMap();
        	for(String secretLabel : givenSecretLabels) {
            	for(String observableLabel : givenJointFreqMap.get(secretLabel).keySet()) {
                    if(!this.jointFreqMap.containsKey(secretLabel)) {
                	    Map<String,Integer> observableFreqMap = new LinkedHashMap<String,Integer>();
                	    int givenFreq = givenJointFreqMap.get(secretLabel).get(observableLabel);
                	    observableFreqMap.put(observableLabel, givenFreq);
                        this.jointFreqMap.put(secretLabel, observableFreqMap);
                    } else if(!this.jointFreqMap.get(secretLabel).containsKey(observableLabel)) {
                	    Map<String,Integer> observableFreqMap = this.jointFreqMap.get(secretLabel);
                	    int givenFreq = givenJointFreqMap.get(secretLabel).get(observableLabel);
                	    observableFreqMap.put(observableLabel, givenFreq);
                        this.jointFreqMap.put(secretLabel, observableFreqMap);
                    } else{
                	    int givenFreq = givenJointFreqMap.get(secretLabel).get(observableLabel);
                	    int thisFreq = this.jointFreqMap.get(secretLabel).get(observableLabel);
                	    int sumFreq = givenFreq + thisFreq;
                    	this.jointFreqMap.get(secretLabel).put(observableLabel, sumFreq);
                    }
            	}
        	}
    	}
    }
    //TODO: Add more methods for synchronized operations on distributions

	////////////////////////////////////////////////////////////////
    // Methods for computing projection
	/**
	 * Computes the distribution on secrets from the joint frquency distribution.
	 * @return the distribution on secrets
	 */
	public SingleFrequencyDistribution getProjectionSecrets() {
    	if(this.secretLabels == null) {
            log("ERROR(JFD266): Cannot get the secret domain. This joint frquency distribution is not initialized. secretLabels is null.",1);
            System.exit(1);
    	}
		Map<String,Integer> freqMap = new LinkedHashMap<String,Integer>();
    	for(String secretLabel : this.secretLabels) {
    		int freq = 0;
        	for(String observableLabel : this.observableLabels) {
        		int jointFreq = this.jointFreqMap.get(secretLabel).get(observableLabel);
        		freq += jointFreq;
        	}
        	freqMap.put(secretLabel, freq);
    	}
		SingleFrequencyDistribution dist = new SingleFrequencyDistribution(this.getSecretLabels(), freqMap);
		return dist;
	}

	/**
	 * Computes the distribution on observables from the joint frquency distribution.
	 * @return the distribution on observables
	 */
	public SingleFrequencyDistribution getProjectionObservables() {
    	if(this.observableLabels == null) {
            log("ERROR(JFD288): Cannot get the observable domain. This joint frquency distribution is not initialized. observableLabels is null.",1);
            System.exit(1);
    	}
		Map<String,Integer> freqMap = new LinkedHashMap<String,Integer>();
    	for(String observableLabel : this.observableLabels) {
    		int freq = 0;
        	for(String secretLabel : this.secretLabels) {
        		int jointFreq = this.jointFreqMap.get(secretLabel).get(observableLabel);
        		freq += jointFreq;
        	}
        	freqMap.put(observableLabel, freq);
    	}
		SingleFrequencyDistribution dist = new SingleFrequencyDistribution(this.getObservableLabels(), freqMap);
		return dist;
	}
        
	////////////////////////////////////////////////////////////////
    // Other methods
    /**
     * Computes the summation of all frquencies in this joint (sub-)distribution
     * 
     * @return the summation of all frquencies in this joint (sub-)distribution
     */
    public int getSumOfFreqabilities() {
    	if(this.secretLabels == null || this.observableLabels == null) {
            log("ERROR(JFD313): This joint frquency distribution is not initialized.",1);
            System.exit(1);
    	}
    	int sum = 0;
    	for(String secretLabel : this.secretLabels) {
        	for(String observableLabel : this.observableLabels) {
            	Map<String,Integer> observableFreqMap = this.jointFreqMap.get(secretLabel);
            	if(observableFreqMap != null) {
            		int freq = observableFreqMap.get(observableLabel);
                	if(freq > 0) {
                		sum += freq;
                	}
            	}
        	}
    	}
    	return sum;
    }
    
    
    /**
     * Print this joint frequency distribution.
     */
    public void print() {
    	print(1);
    }
    
    /**
     * Print this joint frequency distribution.
     */
    public void print(int level) {
        if (Logger.getInstance().getVerbosity()<level){
            return;
        }
        log("{",level);
        printLabels(level);
        printFreqs(level);
        log("}",level);
    }
    
    /**
     * Print this joint frequency distribution.
     */
    public void print(String string) {
    	print(string, 1);
    }
    
    /**
     * Print this joint frequency distribution.
     */
    public void print(String string, int level) {
        if (Logger.getInstance().getVerbosity()<level){
            return;
        }
        log(string+" = {",level);
        printLabels(level);
        printFreqs(level);
        log("}",level);
    }

    private void printLabels(int level) {
        if (Logger.getInstance().getVerbosity()<level){
            return;
        }
        log("  secret labels = {",level);
        if(this.secretLabels != null) {
        	for(String secretLabel : this.secretLabels) {
                log("    "+secretLabel,level);
        	}
        }
        log("  }",level);
        log("  observable labels = {",level);
        if(this.observableLabels != null) {
        	for(String observableLabel : this.observableLabels) {
                log("    "+observableLabel,level);
        	}
        }
        log("  }",level);
    }

    private void printFreqs(int level) {
        if (Logger.getInstance().getVerbosity()<level){
            return;
        }
    	if(this.secretLabels != null && this.observableLabels != null) {
        	for(String secretLabel : this.jointFreqMap.keySet()) {
                Map<String,Integer> observableFreqMap = this.jointFreqMap.get(secretLabel);
                for(String observableLabel : observableFreqMap.keySet()) {
            	int freq = observableFreqMap.get(observableLabel);
                	if(freq > 0) {
                        log(" Cnt["+secretLabel+", "+observableLabel+"] = "+freq,level);
                	} else if(freq < 0) {
                        log(" ERROR(JFD376):Cnt["+secretLabel+", "+observableLabel+"] = "+freq,1);
                	}
                }
        	}
    	}
    }

    private static void log(Object aObject, int level){
        if (logger.getVerbosity() >= level) {
        	Logger loggy = Logger.getInstance();
        	loggy.log(String.valueOf(aObject),level);
        }
    }

    synchronized public void increaseFrequency(String secretId, String observId) {
        this.setFrequency(secretId, observId, this.getFrequency(secretId, observId)+1);
    }

    public JointFrequencyDistribution replicateSecretDistribution(String secretid, SimulationSignature sig) {
        logger.changeVerbosity("JFD_REPLICATE");
    	Map<String,Integer> toduplicate = this.jointFreqMap.get(secretid);
        HashSet<String> newObsLabels = new HashSet<String>();
        newObsLabels.addAll(toduplicate.keySet());
        HashSet<String> newSecLabels = new HashSet<String>();
        Map<String,Map<String,Integer>> newmap = new LinkedHashMap<String,Map<String,Integer>>();
        VariableList secVars = sig.getSecVars();
        Iterable[] iters = new Iterable[secVars.size()];
        for (int i = 0; i < secVars.size(); i++) {
            PrivateVariable pvar = (PrivateVariable) secVars.get(i);
            iters[i] = pvar.getInterval();
        }
        for (BigIntegerCartesianIterator cartiter = new BigIntegerCartesianIterator(iters); cartiter.hasNext();) {
            VariableList varList = new VariableList();
            log("Creating secret value:", 4);
            BigInteger values[] = cartiter.next();
            // giving a single value to each private variable
            for (int i = 0; i < secVars.size(); i++) {
                BigInteger value = values[i];
                String valueStr = new Interval(value).toString();
                log(secVars.getVariableName(i) + ":=" + valueStr, 3);
                varList.add(new PrivateVariable(secVars.getVariableName(i),valueStr,secVars.getVariableLength(i)));
            }
            String secLabel = varList.getSingleIdentifier();
            newmap.put(secLabel, toduplicate);
            newSecLabels.add(secLabel);
        }
        logger.restoreVerbosity();
        return new JointFrequencyDistribution(newSecLabels,newObsLabels,newmap);
    }
}
