/*  HyLeak is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2014-2017 - 

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: 
*/

package distributions;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.HashMap;
import java.util.Set;

import hyleak.HashFreqProb;
import hyleak.IntervalAssignments;
import hyleak.VariableList;
import support.Contexter;
import support.Logger;
import theory.BigFunctions;


/**
 *
 * @author  
 */
public class MultiIntervalProbabilityDistribution extends HashMap< IntervalAssignments , HashFreqProb > {

    protected static final Contexter contexter = Contexter.getInstance();
    protected static final Logger logger = Logger.getInstance();
  
    public String printNice() {
        String toreturn = "";
        Set<IntervalAssignments> keys = this.keySet();
        for (IntervalAssignments key : keys){
            HashFreqProb hfp = this.get(key);
            toreturn +="assignment "+key.printNice()+" with frequency "+hfp.getTotalFrequency()+" and probability "+hfp.getTotalProbability().toEngineeringString()+"\n";
        }
        return toreturn;
    }

    private static void log(Object aObject, int level){
        Logger loggy = Logger.getInstance();
        loggy.log(String.valueOf(aObject),level);
    }

    public void update(VariableList secrets, String variables, BigDecimal stateprob) {
        this.update(new IntervalAssignments(secrets),variables,stateprob);
    }
    
    public void update(IntervalAssignments secrets, String variables, BigDecimal stateprob) {
        IntervalAssignments newassignment = secrets;
        BigDecimal prob = stateprob;
        
        log("Updating with assignment "+secrets.printNice()+", variables "+variables+" and probability "+stateprob.toEngineeringString(),3);
        // if the new assignment already exists, just update its value
        if (this.containsKey(newassignment)){
            log("The assignment already exists; updating its probability and increasing frequency by 1",3);
            this.get(newassignment).updateProbability(variables, prob);
            return;
            // BUG: THIS NEVER GETS CHOSEN
        }
        
        HashMap< IntervalAssignments , HashFreqProb > newdist = new HashMap();
        
        Set<IntervalAssignments> keys = this.keySet();
        for (IntervalAssignments oldassignment : keys){
            if (newassignment.hasEmptyInterval()){
                newdist.put(oldassignment, this.get(oldassignment));
                continue;
            }
            HashFreqProb cloneoldHFP = (HashFreqProb) this.get(oldassignment).clone();
            //log("Old HFP: "+cloneoldHFP.printNice(),3);
            // we have the old assignment and the new assignment; we obtain their intersection, the rest of the old assignment without the intersection, and the rest of the new assignment without the interseciton
            //IntervalAssignments[] splitted = oldassignment.splitIntervals(newassignment);
            IntervalAssignments[] splitted = new IntervalAssignments[3];
            
            log("Splitting current assignment "+oldassignment.printNice()+" with new assignment "+newassignment.printNice(),3);
            IntervalAssignments intersection = splitted[0];
            IntervalAssignments restofold    = splitted[1];
            IntervalAssignments restofnew    = splitted[2];
            log("Result of split:\nintersection: "+intersection.printNice()+"\nrest of old: "+restofold.printNice()+"\nrestofnew: "+restofnew.printNice(),3);
        
            BigInteger oldsize = oldassignment.getSize();
            BigInteger intsize = intersection.getSize();
            BigInteger newsize = newassignment.getSize();
            
            // old interval: change probability to account for reduced size and put with same frequency
            if (!restofold.hasEmptyInterval()){
                
                BigInteger restofoldsize = restofold.getSize();
                HashFreqProb oldHFP = (HashFreqProb) cloneoldHFP.clone();
                if (oldsize!=restofoldsize){
                    BigDecimal tomult = new BigDecimal(restofoldsize,contexter.getContext());
                    tomult = tomult.divide(new BigDecimal(oldsize,contexter.getContext()), contexter.getContext());
                    
                    oldHFP.multiplyProbabilities(tomult);
                }
                log("The old interval minus the intersection is "+restofold.printNice()+", putting it back in with "+oldHFP.printNice(),3);
                newdist.put(restofold, oldHFP);
            }else{
                log("The old interval minus the intersection has an empty interval",3);
            }
            
            // intersection: take probability by size from old interval, then update with prob by size from new interval (freq is freq of old plus one)
            if (!intersection.hasEmptyInterval()){
                
                HashFreqProb intersHFP = (HashFreqProb) cloneoldHFP.clone();
                if (oldsize!=intsize){
                    BigDecimal tomult = new BigDecimal(intsize).divide(new BigDecimal(oldsize), contexter.getContext());
                    
                    intersHFP.multiplyProbabilities(tomult);
                }
                BigDecimal toupdate = prob;
                toupdate = toupdate.multiply(new BigDecimal(intsize,contexter.getContext()), contexter.getContext());
                toupdate = toupdate.divide(new BigDecimal(newsize,contexter.getContext()), contexter.getContext());
                
                intersHFP.updateProbability(variables, toupdate);
                
                log("The intersection is "+intersection.printNice()+", putting it in in with "+intersHFP.printNice(),3);
                newdist.put(intersection, intersHFP);
                
                // change remaining probability to reflect the fact that some of it has gone to the intersection
                prob = prob.subtract(toupdate, contexter.getContext());
                
            }else{
                log("The intersection has an empty interval",3);
            }
            
            
            // new interval: change probability to account for reduced size, wait for next iteration
            newassignment = restofnew;
            log("The remainder is "+newassignment.printNice()+" with probability "+prob.toEngineeringString(),3);
            
        }
        // if new interval is now not empty, put it with the remaining probability and frequency 1
        
        
        if (!newassignment.hasEmptyInterval()){
            log("Inserting the remainder "+newassignment.toString()+" with variables "+variables+" and frequency 1 and probability "+prob.toPlainString(),3);
            HashFreqProb newHFP = new HashFreqProb(variables,prob);
            newdist.put(newassignment, newHFP);
        }
        this.clear();
        this.putAll(newdist);
    }

    public void initialize(VariableList secrets, String varidentifier, BigDecimal stateprob) {
        IntervalAssignments newkey = new IntervalAssignments(secrets);
        
        HashFreqProb newelement = new HashFreqProb(varidentifier,stateprob);
        
        this.put(newkey, newelement);
        
    }

    public void pad(IntervalAssignments initialsecretintervals) {
        Integer totobs = this.getTotalFrequency();
        BigDecimal totprob = this.getTotalProbability();
        BigDecimal padprob = totprob.divide(new BigDecimal(totobs), contexter.getContext());
        this.update(initialsecretintervals, "", padprob);
    }

    

    public BigDecimal getEntropy() {
        BigDecimal entropy = contexter.getDecimalZero();
        for (IntervalAssignments inter : this.keySet()){
            BigInteger intervsize = inter.getSize();
            
            HashFreqProb hfp = this.get(inter);
            BigDecimal prob = hfp.getTotalProbability();
            BigDecimal localent = prob.multiply(log2(prob.divide(new BigDecimal(intervsize), contexter.getContext())),contexter.getContext());
            
            log("Local entropy: "+localent,4);
            entropy = entropy.subtract(localent, contexter.getContext());

        }
        return entropy;

    }
    
    protected BigDecimal log2(BigDecimal value) {
        if (value.compareTo(contexter.getDecimalOne())==0){
            return contexter.getDecimalZero();
        }
        log("Computing the logarithm in base 2 of "+value.toPlainString()+" with scale "+value.scale(),4);
        BigDecimal toreturn = BigFunctions.ln(value,contexter.getContext().getPrecision());
        log("The natural logarithm is "+toreturn,4);
        toreturn = toreturn.divide(BigFunctions.ln(new BigDecimal("2",contexter.getContext()),contexter.getContext().getPrecision()),contexter.getContext());
        log("The logarithm in base 2 is "+toreturn,4);
        
        return toreturn;
    }

    public void normalizeTo(BigDecimal prob) {
        BigDecimal currprob = this.getTotalProbability();
        if (currprob.compareTo(prob)!=0){
            BigDecimal multiplier = prob.divide(currprob, contexter.getContext());
            for (HashFreqProb map : this.values()){
                map.multiplyProbabilities(multiplier);
            }
        }
    }
    
    public void normalize() {
        this.normalizeTo(contexter.getDecimalOne());

    }

    private BigDecimal getTotalProbability() {
        BigDecimal toreturn = contexter.getDecimalZero();
        for (HashFreqProb hfp : this.values()){
            toreturn = toreturn.add(hfp.getTotalProbability(), contexter.getContext());
        }
        return toreturn;
    }

    private Integer getTotalFrequency() {
        Integer toreturn = 0;
        for (HashFreqProb hfp : this.values()){
            toreturn += hfp.getTotalFrequency();
        }
        return toreturn;
    }

    BigDecimal getHighestProbability() {
        BigDecimal max = contexter.getDecimalZero();
        for (HashFreqProb hfp: this.values()){
            BigDecimal prob = hfp.getTotalProbability();
            if (prob.compareTo(max)>0){
                max = prob;
            }
        }
        return max;
    }

    

    

    
}
