/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package distributions;

import java.math.BigDecimal;
import java.math.MathContext;
import java.util.Arrays;
import java.util.Set;
import java.util.HashSet;
import java.util.Map;
import java.util.LinkedHashMap;

import support.Contexter;
import support.Logger;
import theory.BigFunctions;

/**
*
* @author  
* @author  
*/
public class SingleDistribution implements ProbabilityDistribution, Cloneable {
    protected static final Contexter contexter = Contexter.getInstance();
    private static final Logger logger = Logger.getInstance();

	/**
	 * array of labels (identifiers) of the domain
	 */
	private Set<String> labels = null;

	/**
	 * map that returns the probability of a given label
	 */
	private Map<String,BigDecimal> probMap = new LinkedHashMap<>();

	/**
	 * Constructs the empty distribution.
	 */
	public SingleDistribution() {
        this.labels = new HashSet<>();
        this.probMap = new LinkedHashMap<>();
	}

	/**
	 * Constructs a single distribution with initial values.
	 * 
	 * @param labels set of labels
	 * @param probMap hashmap that represents a single distribution
	 */
	public SingleDistribution(Set<String> labels, Map<String,BigDecimal> probMap) {
        this.labels = new HashSet<>();
		this.labels.addAll(labels);
        this.probMap = new LinkedHashMap<>();
		this.probMap.putAll(probMap);
	}

	/**
	 * Constructs a single distribution with initial values.
	 * 
	 * @param labels array of labels
	 * @param probMap hashmap that represents a single distribution
	 */
	public SingleDistribution(String[] labels, Map<String,BigDecimal> probMap) {
		this.labels = new HashSet<>();
		this.labels.addAll(Arrays.asList(labels));
        this.probMap = new LinkedHashMap<>();
		this.probMap.putAll(probMap);
	}

	/**
	 * Constructor that copies this single distribution.
	 * 
	 * @param singleDist signle distribution
	 */
	public SingleDistribution(SingleDistribution singleDist) {
		if(singleDist != null) {
	        this.labels = new HashSet<>();
			this.labels.addAll(singleDist.labels);
	        this.probMap = new LinkedHashMap<>();
			this.probMap.putAll(singleDist.probMap);
		}
	}

	/**
	 * Returns the set of the labels.
	 * 
	 * @return set of the labels of this single distribution
	 */
	public Set<String> getLabels() {
    	if(this.labels == null) {
            log("ERROR(SD76): Cannot get the labels. This single distribution is not initialized.",1);
            return null;
    	}
		return this.labels;
	}
	
	/**
	 * Computes the size of the domain of this single distribution.
	 * @return the size of the domain of this single distribution
	 */
	public int getSizeOfDomain() {
    	if(this.labels == null) {
            log("ERROR(SD88): Cannot get the size of the domain. This single distribution is not initialized.",1);
            return -1;
    	}
		return this.labels.size();
	}

	/**
	 * Returns the mapping from a label to its joint probability
	 * 
	 * @return the mapping from a label to its joint probability
	 */
	public Map<String,BigDecimal> getProbMap() {
    	if(this.probMap == null) {
            log("ERROR(SD101): This single distribution is not initialized.",1);
            System.exit(1);
    	}
		return this.probMap;
	}

	/**
	 * Computes the probability of having the given label.
	 * 
	 * @param keyLabel label
	 * @return the probability of having the given label
	 */
    @Override
    public BigDecimal getProbability(String keyLabel) {
    	if(this.probMap == null) {
            log("ERROR(SD115): This single distribution is not initialized.",1);
            System.exit(1);
    	}
    	BigDecimal prob = this.probMap.get(keyLabel);
    	if(prob != null) {
    		return prob;
    	} else {
    		return BigDecimal.ZERO;
    	}
    }
    
	/**
	 * Set the probability of having the given keys.
	 * 
	 * @param keyLabel label
     * @param probability the probability of having the given keys
	 */
    synchronized public void setProbability(String keyLabel, BigDecimal probability) {
    	if(this.probMap == null) {
            log("ERROR(SD134): This single distribution is not initialized.",1);
            System.exit(1);
    	}
        this.probMap.put(keyLabel, probability);
    }

	/**
	 * Adds a given (sub-)distribution to this (sub-)distribution.
	 * 
     * @param dist given (sub-)distribution
	 */
    synchronized public void add(SingleDistribution dist) {
    	if(dist == null) {
            log("ERROR(SD158): The added single distribution is not initialized.",1);
    	} else if(dist.getProbMap() == null) {
            log("CAUTION(SD160): Cannot get the probability map. The given single distribution is not initialized and failed to be added to this distribution.",2);
    	} else if(this.getProbMap() == null) {
            log("CAUTION(SD162): Cannot get the probability map. This single distribution is not initialized.",2);
    		this.labels.addAll(dist.getLabels());
            this.probMap=dist.getProbMap();
    	} else {
    		this.labels.addAll(dist.getLabels());
        	Set<String> givenLabels = dist.getLabels();
        	Map<String,BigDecimal> givenProbMap = dist.getProbMap();
        	for(String label : givenLabels) {
                if(!this.probMap.containsKey(label)) {
             	    BigDecimal givenProb = givenProbMap.get(label);
                    this.probMap.put(label, givenProb);
                } else{
             	    BigDecimal givenProb = givenProbMap.get(label);
             	    BigDecimal thisProb = this.probMap.get(label);
             	    BigDecimal sumProb = givenProb.add(thisProb);
                 	this.probMap.put(label, sumProb);
                 }
         	}
    	}
    }
    
    //TODO: Add more methods for synchronized operations on distributions

    /**
     * Computes the summation of all probabilities in this joint (sub-)distribution
     * 
     * @return the summation of all probabilities in this joint (sub-)distribution
     */
    public BigDecimal getSumOfProbabilities() {
    	if(this.labels == null) {
            log("ERROR(SD192): This single distribution is not initialized.",1);
            System.exit(1);
    	}
    	BigDecimal sum = BigDecimal.ZERO;
    	for(String label : this.labels) {
    		sum = sum.add(this.probMap.get(label));
    	}
    	return sum;
    }

    @Override
    public BigDecimal getEntropy() {
        BigDecimal toreturn = contexter.getDecimalZero();
        for (String label : probMap.keySet()){
            BigDecimal prob = probMap.get(label);
            toreturn = toreturn.subtract(prob.multiply(BigFunctions.log2(prob, contexter.getContext().getPrecision()), contexter.getContext()),contexter.getContext());
        }
        return toreturn;
    }

    @Override
    public String printNice() {
        throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
    }
    
    /**
     * Print this single distribution.
     */
    public void print() {
    	print(1);
    }
    
    /**
     * Print this single distribution.
     */
    public void print(int level) {
        log("{",level);
        printLabels(level);
        printProbs(level);
        log("}",level);
    }

    /**
     * Print this single distribution.
     */
    public void print(String string) {
    	print(string, 1);
    }

    /**
     * Print this single distribution.
     */
    public void print(String string, int level) {
        log(string+" = {",level);
        printLabels(level);
        printProbs(level);
        log("}",level);
    }

    private void printLabels(int level) {
        log("  labels = {",level);
        if(this.labels != null) {
        	for(String label : this.labels) {
                log("    "+label,level);
        	}
        }
        log("  }",level);
    }

    private void printProbs(int level) {
    	if(this.labels != null) {
        	for(String label : this.labels) {
        		BigDecimal prob = this.probMap.get(label);
            	if(prob != null) {
                    log("  Pr["+label+"] = "+prob,level);
            	} else {
                    //log("  Pr["+label+"] = null",level);
                }
        	}
    	}
    }

    private static void log(Object aObject, int level){
        if (logger.getVerbosity() >= level) {
        	Logger loggy = Logger.getInstance();
        	loggy.log(String.valueOf(aObject),level);
        }
    }
}
