/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package distributions;

import java.math.BigDecimal;
import java.util.Arrays;
import java.util.Set;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.Map;

import support.Contexter;
import support.Logger;

/**
* @author  
*/
public class SingleFrequencyDistribution {
    protected static final Contexter contexter = Contexter.getInstance();
    private static final Logger logger = Logger.getInstance();

	/**
	 * array of labels (identifiers) of the domain
	 */
	private Set<String> labels = null;

	/**
	 * map that returns the frequency of a given label
	 */
	private Map<String,Integer> freqMap = new LinkedHashMap<String,Integer>();

	/**
	 * Constructs the empty frquency distribution.
	 */
	public SingleFrequencyDistribution() {
        this.labels = new HashSet<String>();
        this.freqMap = new LinkedHashMap<String,Integer>();
	}

	/**
	 * Constructs a single frquency distribution with initial values.
	 * 
	 * @param labels set of labels
	 * @param freqMap hashmap that represents a single frquency distribution
	 */
	public SingleFrequencyDistribution(Set<String> labels, Map<String,Integer> freqMap) {
        this.labels = new HashSet<String>();
		this.labels.addAll(labels);
        this.freqMap = new LinkedHashMap<String,Integer>();
		this.freqMap.putAll(freqMap);
	}

	/**
	 * Constructs a single frquency distribution with initial values.
	 * 
	 * @param labels array of labels
	 * @param freqMap hashmap that represents a single frquency distribution
	 */
	public SingleFrequencyDistribution(String[] labels, Map<String,Integer> freqMap) {
		this.labels = new HashSet<String>();
		this.labels.addAll(Arrays.asList(labels));
        this.freqMap = new LinkedHashMap<String,Integer>();
		this.freqMap.putAll(freqMap);
	}

	/**
	 * Constructor that copies this single frquency distribution.
	 * 
	 * @param singleDist signle frquency distribution
	 */
	public SingleFrequencyDistribution(SingleFrequencyDistribution singleDist) {
		if(singleDist != null) {
	        this.labels = new HashSet<String>();
			this.labels.addAll(singleDist.labels);
	        this.freqMap = new LinkedHashMap<String,Integer>();
			this.freqMap.putAll(singleDist.freqMap);
		}
	}

	/**
	 * Returns the set of the labels.
	 * 
	 * @return set of the labels of this single frquency distribution
	 */
	public Set<String> getLabels() {
    	if(this.labels == null) {
            log("ERROR(SFD80): Cannot get the labels. This single frquency distribution is not initialized.",1);
            return null;
    	}
		return this.labels;
	}
	
	/**
	 * Computes the size of the domain of this single frquency distribution.
	 * @return the size of the domain of this single frquency distribution
	 */
	public int getSizeOfDomain() {
    	if(this.labels == null) {
            log("ERROR(SFD92): Cannot get the size of the domain. This single frquency distribution is not initialized.",1);
            return -1;
    	}
		return this.labels.size();
	}

	/**
	 * Returns the mapping from a label to its joint frequency
	 * 
	 * @return the mapping from a label to its joint frequency
	 */
	public Map<String,Integer> getFreqMap() {
    	if(this.freqMap == null) {
            log("ERROR(SFD105): This single frquency distribution is not initialized.",1);
            System.exit(1);
    	}
		return this.freqMap;
	}

	/**
	 * Computes the frequency of having the given label.
	 * 
	 * @param keyLabel label
	 * @return the frequency of having the given label
	 */
    public int getFrequency(String keyLabel) {
    	if(this.freqMap == null) {
            log("ERROR(SFD119): This single frquency distribution is not initialized.",1);
            System.exit(1);
    	}
    	if(this.freqMap.get(keyLabel) != null) {
    		return this.freqMap.get(keyLabel);
    	} else {
    		return 0;
    	}
    }
    
	/**
	 * Set the frequency of having the given keys.
	 * 
	 * @param keyLabel label
     * @param frequency the frequency of having the given keys
	 */
    synchronized public void setFrequency(String keyLabel, int frequency) {
    	if(this.freqMap == null) {
            log("ERROR(SFD137): This single frquency distribution is not initialized.",1);
            System.exit(1);
    	}
        this.freqMap.put(keyLabel, frequency);
    }

	/**
	 * Adds a given frquency distribution to this frquency distribution.
	 * 
     * @param jointDist given frquency distribution
	 */
    synchronized public void add(SingleFrequencyDistribution dist) {
    	if(dist == null) {
            log("ERROR(SFD150): The added single frquency distribution is not initialized.",1);
    	} else if(dist.getFreqMap() == null) {
            log("CAUTION(SD153): Cannot get the freqency map. The given single freqency distribution is not initialized and failed to be added to this distribution.",2);
    	} else if(this.freqMap == null) {
            log("CAUTION(SFD155): This single frquency distribution is not initialized.",2);
    		this.labels.addAll(dist.getLabels());
            this.freqMap=dist.getFreqMap();
    	} else {
    		this.labels.addAll(dist.getLabels());
        	Set<String> givenLabels = dist.getLabels();
        	Map<String,Integer> givenFreqMap = dist.getFreqMap();
        	for(String label : givenLabels) {
                if(!this.freqMap.containsKey(label)) {
             	    int givenFreq = givenFreqMap.get(label);
                    this.freqMap.put(label, givenFreq);
                } else{
             	    int givenFreq = givenFreqMap.get(label);
             	    int thisFreq = this.freqMap.get(label);
             	    int sumFreq = givenFreq + thisFreq;
                 	this.freqMap.put(label, sumFreq);
                }
         	}
    	}
    }
    
    //TODO: Add more methods for synchronized operations on frquency distributions

    /**
     * Computes the summation of all frquencies in this joint frquency distribution
     * 
     * @return the summation of all frquencies in this joint frquency distribution
     */
    public int getSumOfFrquencies() {
    	if(this.labels == null) {
            log("ERROR(SFD184): This single frquency distribution is not initialized.",1);
            System.exit(1);
    	}
    	int sum = 0;
    	for(String label : this.labels) {
    		sum += this.freqMap.get(label);
    	}
    	return sum;
    }

    /**
     * Print this single freqency distribution.
     */
    public void print() {
    	print(1);
    }
    
    /**
     * Print this single freqency distribution.
     */
    public void print(int level) {
        if (Logger.getInstance().getVerbosity()<level){
            return;
        }
        log("{",level);
        printLabels(level);
        printFreqs(level);
        log("}",level);
    }
    
    /**
     * Print this single freqency distribution.
     */
    public void print(String string) {
    	print(string, 1);
    }
    
    /**
     * Print this single freqency distribution.
     */
    public void print(String string, int level) {
        if (Logger.getInstance().getVerbosity()<level){
            return;
        }
        log(string+" = {",level);
        printLabels(level);
        printFreqs(level);
        log("}",level);
    }

    private void printLabels(int level) {
        if (Logger.getInstance().getVerbosity()<level){
            return;
        }
        log("  labels = {",level);
        if(this.labels != null) {
        	for(String label : this.labels) {
                log("    "+label,level);
        	}
        }
        log("  }",level);
    }

    private void printFreqs(int level) {
        if (Logger.getInstance().getVerbosity()<level){
            return;
        }
    	if(this.labels != null) {
        	for(String label : this.labels) {
        		int freq = this.freqMap.get(label);
            	if(freq > 0) {
                    log("  Cnt["+label+"] = "+freq,level);
            	} else if(freq < 0) {
                    log("  ERROR(SFD229):Cnt["+label+"] = "+freq,1);
                }
        	}
    	}
    }

    private static void log(Object aObject, int level){
        if (logger.getVerbosity() >= level) {
        	Logger loggy = Logger.getInstance();
        	loggy.log(String.valueOf(aObject),level);
        }
    }
}
