/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package hyleak;

import java.io.FileNotFoundException;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.logging.Level;

import distributions.JointDistribution;

import support.Contexter;
import support.Logger;

/**
 *
 * @author  
 * @author  
 */
class ConcurrentHybridExplorer {

    private static final Contexter contexter = Contexter.getInstance();
    private static final Logger logger = Logger.getInstance();
    private final int waitTimeDefault = 2;

    private final int keepAliveTimeDefault = 10;
    private int numthreads = 2;
    private final int numSimulations;
    private final BigDecimal treshold;
    private final long timelimit;
    private final boolean drawcfg;

    public ConcurrentHybridExplorer(int iter, BigDecimal tresh, long tlimit, int threads, boolean docfg) {
        numSimulations = iter;
        treshold = tresh;
        numthreads = threads;
        timelimit = tlimit;
        drawcfg = docfg;
    }

    void explore(String filename) {
        ArrayList<String> codelines = null;

        logger.changeVerbosity("CODEREADER");
        long starttime = System.currentTimeMillis();
        CodeReader reader = new CodeReader(filename);

        try {
            codelines = reader.readLines();
        } catch (FileNotFoundException fnf) {
            log("Error while processing input file: file " + filename + " not found", 1);
            System.exit(1);

        }
        //initializing
        AtomicInteger count = new AtomicInteger(1);
        long keepAliveTime = keepAliveTimeDefault;
        boolean allexplored = false;
        boolean timeout = false;
        boolean alliterations = false;
        boolean alltraces = false;
        long time = System.currentTimeMillis();
        BigInteger numOfSims = new BigInteger("" + numSimulations);

        MarkovChainState firststate = new MarkovChainState("current", contexter.getDecimalOne());
        PublicVariable programcounter = new PublicVariable("pc", BigInteger.ONE, 32);
        firststate.addVariable(programcounter);

        logger.changeVerbosity("CFG");
        if (drawcfg) {
            CFGExtractor cfgExtr = new CFGExtractor(codelines);
            ControlFlowGraph cfg = cfgExtr.getCFG();
            log(cfg, 3);
            cfg.printToPNGFile("cfg");

        }

        logger.changeVerbosity("PRIORDIST");
        CodePriorDistributionManager priorDistMan = new CodePriorDistributionManager(codelines);

        BigDecimal priorentropy = priorDistMan.computePriorShannonEntropy();
        //BigDecimal priorminentropy = priordistman.computePriorMinEntropy();
        log("Prior distribution analysis: secret has " + priorDistMan.getSecretSize() + " different values, corresponding to:", 2);
        log("a prior Shannon entropy of " + priorentropy + " bits.", 2);
        //log("a prior min-entropy of "+priorminentropy+" bits.",2);

        JointPosteriorDistributionManager postdistman = new JointPosteriorDistributionManager(priorDistMan, numOfSims);
        //JointPosteriorDistributionManagerWithKnownPrior postdistman = new JointPosteriorDistributionManagerWithKnownPrior(priorDistMan, numOfSims);

        /* TEMPORARY: to prevent the postdistman from panicking when no trace analysis is completed*/
//        MarkovChainState bogusState = new MarkovChainState("fake",contexter.getDecimalZero());
//        PublicVariable obsVar = new PublicVariable("o",BigInteger.ONE,16);
//        obsVar.setObservable(true);
//        bogusState.addVariable(obsVar);
//        PrivateVariable secVar = new PrivateVariable("h","[0,3]",4);
//        secVar.setSecret(true);
//        bogusState.addVariable(secVar);
//        postdistman.storeExactSubDistToManager(bogusState.getJointSecretObservablesDistribution());
        // end temporary
        BigDecimal totalProb = contexter.getDecimalZero();
        Simulator simulator = new Simulator(numSimulations, postdistman, numthreads, keepAliveTime, codelines);
        logger.changeVerbosity("EXPLORER");
        log("Starting hybrid state space exploration at time " + (System.currentTimeMillis() - starttime) + " milliseconds using " + numthreads + " threads", 2);
        //ThreadPoolExecutor tpoolexplorer = Executors.newCachedThreadPool();
        ThreadPoolExecutor tpoolexplorer = new ThreadPoolExecutor(numthreads, numthreads, keepAliveTime, TimeUnit.MILLISECONDS, new LinkedBlockingQueue<Runnable>());
        tpoolexplorer.submit(new RunnableSingleHybridExplorer(firststate, codelines, postdistman, tpoolexplorer, count, simulator));
        while (!allexplored && !timeout && !alliterations && !alltraces) {
            try {
                Thread.sleep(waitTimeDefault);
            } catch (InterruptedException ex) {
                java.util.logging.Logger.getLogger(ConcurrentHybridExplorer.class.getName()).log(Level.SEVERE, null, ex);
            }
            totalProb = postdistman.getSumOfProbabilities();

            long currtime = System.currentTimeMillis();
            if (totalProb.compareTo(treshold) >= 0) {
                allexplored = true;
            }
            if (tpoolexplorer.isTerminated()) {
                alltraces = true;
            }
            if ((currtime - time) > timelimit && timelimit != 0) {
                timeout = true;
            }
            long completedTraces = tpoolexplorer.getCompletedTaskCount();
//            if (completedsim>=numSimulations){
//                alliterations = true;
//            }
            if ((completedTraces % 100) == 0) {
                log(completedTraces + " traces explored at time " + (currtime - time) + " milliseconds, total probability " + totalProb.toEngineeringString(), 2);
            }
        }
        tpoolexplorer.shutdownNow();
        if (allexplored) {
            log("We have reached the treshold of " + treshold.toEngineeringString() + ", no further simulation needed.", 2);
        }
        if (timeout) {
            log("We have reached the time limit of " + timelimit + " milliseconds, moving to leakage computation.", 2);
        }
//        if (alliterations){
//            log(numSimulations+" simulations completed, moving to leakage computation.",2);
//        }  
        if (alltraces) {
            log("All traces have been explored, moving to simulation.", 2);
        }
        try {
            tpoolexplorer.awaitTermination(1, TimeUnit.HOURS);
        } catch (InterruptedException ex) {
            log("ERROR: Explorer interrupted: " + ex.getLocalizedMessage() + ". Terminating", 1);
            System.exit(1);
        }

        log("Trace analysis terminated  at time " + (System.currentTimeMillis() - starttime) + " milliseconds after exploring " + tpoolexplorer.getCompletedTaskCount() + " traces", 2);
        if (totalProb.compareTo(BigDecimal.ZERO) <= 0) {
            postdistman.storeExactSubDistToManager(new JointDistribution());
        }
        logger.changeVerbosity("SIMULATOR");
        if (simulator.getQueueSize() > 0) {
            log(simulator.getQueueSize() + " simulations have been queued. Starting simulations...", 2);
            simulator.simulate();
        }
        log("Examining outputs...", 2);

        logger.changeVerbosity("POSTERIORDIST");
        // Computes the joint (total) distribution
        JointDistribution jointTotalDist = postdistman.getTotalJointDistribution();
        if (logger.getVerbosity() > 2) {
            jointTotalDist.print(3);
            jointTotalDist.getProjectionObservables().print(3);
            jointTotalDist.getProjectionSecrets().print(3);
        }

        // Computes the posterial Shannon entropy
        BigDecimal postentropy = postdistman.computePosteriorShannonEntropyWithUniformPrior();
        logger.changeVerbosity("POSTERIORENTROPY");
        log("\nThe prior Shannon entropy of the secret is " + priorentropy, 2);
        log("The posterior Shannon entropy is " + postentropy, 2);
        //postdistman.computeConfidenceIntervalOfMI();

       
        // Compute the mutual information
        BigDecimal MI = priorentropy.subtract(postentropy, contexter.getContext());
         logger.changeVerbosity("BIASEDLEAKAGE");
        if (logger.getVerbosity() > 1) {
            log("The Shannon leakage (with possible bias) is " + MI.max(BigDecimal.ZERO) + " bits", 1);
        } else if (logger.getVerbosity() == 1) {
            log(MI.max(BigDecimal.ZERO), 1);
        }

        // Compute the corrected mutual information
        BigDecimal correctedMI = postdistman.computeCorrectedMIPointEstimate();
        
        logger.changeVerbosity("CORRECTEDLEAKAGE");
        if (logger.getVerbosity() > 1) {
            log("The Shannon leakage (after bias correction) is " + correctedMI + " bits", 1);
        } else if (logger.getVerbosity() == 1) {
            log(correctedMI, 1);
        }

        // Compute the confidence interval of the estimated mutual information
        logger.changeVerbosity("CONFIDENCEINTERVAL");
        if (logger.getVerbosity() > 1) {
            BigDecimal[] confidenceInterval = postdistman.computeConfidenceIntervalOfMI();
            logger.changeVerbosity("CONFIDENCEINTERVAL");
            log("The confidence interval is [" + confidenceInterval[0].max(BigDecimal.ZERO) + ", " + confidenceInterval[1].max(BigDecimal.ZERO) + "]", 1);
        } else if (logger.getVerbosity() == 1) {
            BigDecimal[] confidenceInterval = postdistman.computeConfidenceIntervalOfMI();
            logger.changeVerbosity("CONFIDENCEINTERVAL");
            log("[" + confidenceInterval[0] + ", " + confidenceInterval[1] + "]", 1);
        }

        logger.changeVerbosity("TIME");
        if (logger.getVerbosity() > 1) {
            log("Time for exploration: " + (System.currentTimeMillis() - starttime) + " milliseconds", 2);
        } else if (logger.getVerbosity() == 1) {
            log ("T(ms):"+ (System.currentTimeMillis() - starttime),1);
        }
         
        // Log the time for exploration
        
    }

    private static void log(Object aObject, int level) {
        if (logger.getVerbosity() >= level) {
            logger.log(String.valueOf(aObject), level);
        }
    }
}
