/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package hyleak;

import java.io.FileNotFoundException;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.logging.Level;
import support.Contexter;
import support.Logger;

/**
 *
 * @author  
 */
class ConcurrentProbabilisticExplorer {
    private static final Contexter contexter = Contexter.getInstance();
    private static final Logger logger = Logger.getInstance();
    private final int waitTimeDefault = 2;

    private final int keepAliveTimeDefault = 1;
    private int numthreads = 2;
    private final int iterations;
    private final BigDecimal treshold;
    private final long timelimit;

    public ConcurrentProbabilisticExplorer(int iter, BigDecimal tresh, long tlimit, int threads){
        iterations = iter;
        treshold = tresh;
        numthreads = threads;
        timelimit = tlimit;
    }

    void explore(String filename) {
        ArrayList<String> codelines = null;

        long starttime = System.currentTimeMillis();
        CodeReader reader = new CodeReader(filename);
       
        try{
            codelines = reader.readLines();
        }catch (FileNotFoundException fnf)
        {
            log("Error while processing input file: file " + filename + " not found",1);
            System.exit(1);
            
        }
        //initializing
        AtomicInteger count = new AtomicInteger(1);
        long keepAliveTime = keepAliveTimeDefault;
        boolean allexplored = false;
        boolean timeout = false;
        boolean alliterations = false;
        boolean alltraces = false;
        long time = System.currentTimeMillis();
        
        MarkovChainState firststate = new MarkovChainState("current",contexter.getDecimalOne());
        PublicVariable programcounter = new PublicVariable("pc",BigInteger.ONE,Integer.valueOf(32));
        firststate.addVariable(programcounter);
        
        CodePriorDistributionManager priordistman = new CodePriorDistributionManager(codelines);
                
        BigDecimal priorentropy = priordistman.computePriorShannonEntropy();
        BigDecimal priorminentropy = priordistman.computePriorMinEntropy();
        log("Prior distribution analysis: secret has "+priordistman.getSecretSize()+" different values, corresponding to:",2);
        log("a prior Shannon entropy of "+priorentropy+" bits.",2);
        log("a prior min-entropy of "+priorminentropy+" bits.",2);
        
        ConcurrentSimplePosteriorDistributionManager postdistman = new ConcurrentSimplePosteriorDistributionManager();
        
        log("Starting hybrid state space exploration at time "+(System.currentTimeMillis()-starttime)+" milliseconds using "+numthreads+" threads",2);
        //ThreadPoolExecutor tpoolexplorer = Executors.newCachedThreadPool();
        ThreadPoolExecutor tpoolexplorer = new ThreadPoolExecutor(numthreads, numthreads, keepAliveTime, TimeUnit.MILLISECONDS, new LinkedBlockingQueue<Runnable>());
        tpoolexplorer.submit(new RunnableSingleProbabilisticExplorer(firststate,codelines,postdistman,tpoolexplorer,count));
        while (!allexplored && !timeout && !alliterations && !alltraces){
            try {
                Thread.sleep(waitTimeDefault);
            } catch (InterruptedException ex) {
                java.util.logging.Logger.getLogger(ConcurrentProbabilisticExplorer.class.getName()).log(Level.SEVERE, null, ex);
            }
            BigDecimal totprob = postdistman.getTotalProbability();
        
            long currtime = System.currentTimeMillis();
            if (totprob.compareTo(treshold)>=0 ){
                allexplored = true;
            }
            if (tpoolexplorer.isTerminated()){
                alltraces = true;
            }
            if ((currtime-time)>timelimit && timelimit!=0){
                timeout = true;
            }
            long completedsim = tpoolexplorer.getCompletedTaskCount();
            if (completedsim>=iterations){
                alliterations = true;
            }
            log(completedsim+" traces explored at time "+(currtime-time)+" milliseconds, total probability "+totprob.toEngineeringString(),2);
        }
        tpoolexplorer.shutdownNow();
        if (allexplored){
            log("We have reached the treshold of "+treshold.toEngineeringString()+", no further simulation needed.",2);
        }
        if (timeout){
            log("We have reached the time limit of "+timelimit+" milliseconds, moving to leakage computation.",2);
        }   
        if (alliterations){
            log(iterations+" simulations completed, moving to leakage computation.",2);
        }  
        if (alltraces){
            log("All traces have been explored, moving to leakage computation.",2);
        }  
        try {
            tpoolexplorer.awaitTermination(1, TimeUnit.HOURS);
        } catch (InterruptedException ex) {
            log("ERROR: Explorer interrupted: "+ex.getLocalizedMessage()+". Terminating",1);
            System.exit(1);
        }
        
        log("State space exploration terminated  at time "+(System.currentTimeMillis()-starttime)+" milliseconds after exploring "+tpoolexplorer.getCompletedTaskCount()+" traces",2);
        
        log("Output distributions:\n"+postdistman.getDistributionOnOutputs().printNice(),3);
        log("Examining outputs...",2);
        

        BigDecimal postentropy = postdistman.computePosteriorShannonEntropy();
        log("\nThe prior Shannon entropy of the secret is "+priorentropy,2);
        log("The posterior Shannon entropy is "+postentropy,2);
        
        
        
        
        BigDecimal leakage = priorentropy.subtract(postentropy);
        if (logger.getVerbosity()>0){
            log("The Shannon leakage is "+leakage+" bits",1);
        }else{
            log(leakage,0);
        }
        
        BigDecimal postminentropy = postdistman.computePosteriorMinEntropy();
        log("\nThe prior min-entropy of the secret is "+priorminentropy,2);
        log("The posterior min-entropy is "+postminentropy,2);
        BigDecimal minleakage = priorminentropy.subtract(postminentropy);
        log("The min-entropy leakage is "+minleakage+" bits",1);
//        double interppostentropy = getInterpolatedValue(interpx,interpy,1.0);
//        log("The interpolated posterior entropy estimate is "+interppostentropy,2);
//        BigDecimal interpleakage = secretbitsize.subtract(new BigDecimal(interppostentropy));
//        log("The interpolated leakage is "+interpleakage+" bits",2);
        
        log("\nTime for exploration: "+(System.currentTimeMillis()-starttime)+" milliseconds",2);
    }
    
    private static void log(Object aObject, int level){
        Logger loggy = Logger.getInstance();
        loggy.log(String.valueOf(aObject),level);
    }
}
