/*  HyLeak is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2014-2017

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: Anonymous team Fylma (real names revealed after acceptance of the paper)
*/
    
package hyleak;

import java.io.File;
import static java.lang.Math.max;
import java.math.BigDecimal;
import java.math.MathContext;
import java.math.RoundingMode;
import java.util.HashMap;
import preprocessor.Preprocessor;
import support.Contexter;
import support.Logger;


/**
 *
 * @author  
 */
public class HyLeak {
    
    protected static String DOT = "/usr/bin/dot";
    private static final Logger logger = Logger.getInstance();
    

    
    private static void log(Object aObject, int level){
        if (logger.getVerbosity() >= level) {
            logger.log(String.valueOf(aObject),level);
        }
    }
    
    
    private static void printHelp() {
        log("HyLeak 3.0 (Hybrid Abstract Analysis for Information Leakage)",0);
        log("",0);
        log("Description:",0);
        log("    HyLeak analyzes a program written in an imperative language with a quantitative security analysis",0);
        log("    that builds and evaluates observability models, and finally computes the information leakage",0);
        log("    of the program.",0);
        log("",0);
        log("Usage:",0);
        log("    hyleak [options] program.hyleak",0);
        log("",0);
        log("Options:",0);
        log("    -c | --clean                 Delete the temporary preprocessed file.",0);
        log("    --config=filename            Alternative location for the configuration file.",0);
        log("    -h | -? | --help             Print this help and terminate.",0);
        log("    -v level                     Set default verbosity between 0 (minimum) and 4 (maximum).",0);
        log("                                 Default value if not specified is 2.",0);
        log("    --verb:<vname>:=<vvalue>     Set verbosity for location <vname> to value <vvalue>",0);
        log("    --verb-locations             Print verbosity locations and terminate",0);
        log("",0);
        log("    -th number                   Sets the number of parallel threads HyLeak should use. Default is the number",0);
        log("                                 returned by Runtime.getRuntime().availableProcessors(), minimum 1",0);
        log("    -p value                     Define computation precision as a number of digits.",0);
        log("                                 Default value if not specified is 15.",0);
        log("    --const:<cname>:=<cvalue>    Replace the value of constant <cname> in the model with value <cvalue>",0);
        log("",0);
        log("    --preprocess                 Run only the preprocessor and terminate.",0);
        log("",0);
        log("    --no-preprocess              Do not run the preprocessor (input file must be in if-goto format).",0);
        log("",0);
        log("    -i iterations                Sets the number of simulations. Default: 50000",0);
//        log("                                 -l threshold        Sets the sufficient probability threshold after which",0);
//        log("                                                     exploration of the probability space can be terminated.",0);   
//        log("                                                     Default: 1 (meaning 100% of the probability space)",0);
        log("    -t time_limit                Sets the time limit (in milliseconds) for the simulation.",0);
        log("                                 A value of 0 means no time limit.",0);   
        log("                                 Default: 0 (meaning no time limit)",0);
        log("    --no-cfg                     Prevents HyLeak from drawing the CFG of the program.",0);
        log("",0);

        

        
    }
    
     private static void printVerbLocations() {
        log("HyLeak 3.0 (Hybrid Abstract Analysis for Information Leakage)",0);
        log("",0);
        log("Verbosity locations for command --verb:",0);
        log("    PREPROCESSOR        The preprocessing steps.",0);
        log("    CODEREADER          Acquisition of the preprocessed code.",0);
        log("    CFG                 Control Flow Graph computation.",0);
        log("    PRIORDIST           Prior distribution computation.",0);
        log("    EXPLORER            Precise trace analysis phase.",0);
        log("    SIMULATOR           Simulation phase.",0);
        log("    POSTERIORDIST       Posterior distribution computation.",0);
        log("    POSTERIORENTROPY    Posterior entropy computation.",0);
        log("    BIASEDLEAKAGE       Leakage computation (with bias).",0);
        log("    CORRECTEDLEAKAGE    Leakage computation (bias corrected).",0);
        log("    CONFIDENCEINTERVAL  Confidence interval computation.",0);
        log("    TIME                Total execution time.",0);
        log("",0);

        

        
    }

    public static void main(String args[]) {
        
        // verbose mode between 0 and 4
        int verbose = 2;
        
        // precision
        int precision = 10;
        
        // number of threads
        int threads = max(1,Runtime.getRuntime().availableProcessors());
        
        // iterations for the simulator
        int iterations = 50000;
        
        // treshold for the simulator to terminate
        BigDecimal threshold = new BigDecimal("1");
        
        // time limit for the simulator
        long timelimit = 0;
        // Markov chains to print
        
        // tasks to perform
        boolean run_preprocessor = true;
       
        boolean run_matrix = false;
        //boolean run_explorer = false;
        boolean run_hybrid = true;
        
        boolean add_free = false;
        boolean clean = false;
        boolean draw_cfg = true;
        
        // files
        
        String filename = "";
        
        
        // constants
        HashMap<String,BigDecimal> constants = new HashMap();
        
        // verbosity levels
        HashMap<String,Integer> verbosities = new HashMap();
         
        // parse command line options
        String arg;
        for( int i=0; i<args.length; i++) {
            arg = args[i];
            
            // help
            if (arg.equals("-h")) {
                printHelp();
                System.exit(1);
            }
            else if (arg.equals("-?")) {
                printHelp();
                System.exit(1);
            }
            else if (arg.equals("--help")) {
                printHelp();
                System.exit(1);
            }
            else if (arg.equals("--verb-locations")) {
                printVerbLocations();
                System.exit(1);
            }
            // verbose
            else if (arg.equals("-v")) {
                i++;
                verbose = Integer.parseInt(args[i]);
                if ((verbose<0)||(verbose>4)){
                    log("ERROR: verbosity must be between 0 and 4",1);
                    System.exit(1);
                }
            }
            // number of threads
            else if (arg.equals("-th")) {
                i++;
                threads = Integer.parseInt(args[i]);
                if ((threads<1)){
                    log("ERROR: number of threads must be at least 1",1);
                    System.exit(1);
                }
            }
            
            // iterations
            else if (arg.equals("-i")) {
                i++;
                iterations = Integer.parseInt(args[i]);
                if (iterations<0){
                    log("ERROR: at least 0 iteration needed",1);
                    System.exit(1);
                }
            }
            // precision
            else if (arg.equals("-p")) {
                i++;
                precision = Integer.parseInt(args[i]);
                if (precision<1){
                    log("ERROR: at least 1 digit precision needed",1);
                    System.exit(1);
                }
                if (precision<5){
                    log("WARNING: at least 5 digit precision recommended",2);
                    
                }
            }
            // probability threshold for the simulation
            else if (arg.equals("-l")) {
                i++;
                try{
                threshold = new BigDecimal(args[i]);
                } catch (NumberFormatException e){
                    log("ERROR: unable to convert '"+args[i]+"' in a BigDecimal threshold",1);
                    System.exit(1);
                }
                if (threshold.compareTo(new BigDecimal(0))<=0){
                    log("ERROR: threshold must be larger than zero",1);
                    System.exit(1);
                }
                if (threshold.compareTo(new BigDecimal(1))>0){
                    log("ERROR: threshold must be smaller or equal than one",1);
                    System.exit(1);
                }
            }
            // time limit for the simulation
           else if (arg.equals("-t")) {
                i++;
                timelimit = Long.parseLong(args[i]);
                if (timelimit<0){
                    log("ERROR: time limit should be a positive number of milliseconds, or 0 for no limit",1);
                    System.exit(1);
                }
            }
            
            // tasks
           else if (arg.equals("--no-cfg")) {
                draw_cfg = false;
            }
            else if (arg.equals("--preprocess")) {
                run_hybrid = false;
            }
            else if (arg.equals("--no-preprocess")) {
                run_preprocessor = false;
            }
//            else if (arg.equals("--hybrid")) {
//                run_explorer = false;
//                run_hybrid = true;
//            }
            else if (arg.equals("--matrix")) {
                run_preprocessor = false;
                //run_explorer = false;
                run_hybrid = false;
                run_matrix = true;
            }
            // constants            
            else if (arg.startsWith("--const:")) {
                String[] constant = arg.substring(8).split(":=");
                constants.put(constant[0],new BigDecimal(constant[1]));
            }
            
            // verbosities            
            else if (arg.startsWith("--verb:")) {
                String[] verb = arg.substring(7).split(":=");
                verbosities.put(verb[0],Integer.parseInt(verb[1]));
            }
            
            // clean
            else if (arg.equals("-c") || arg.equals("--clean")) {
                clean = true;
            }
            else {
                filename = arg;
            }
        }
        
               
        // read file
        File f = new File(filename);
        if (!f.exists()) {
            System.err.println("Error: invalid filename " + filename);
            System.exit(1);
        }
        
        
        
              
        logger.setVerbosity(verbose);
        logger.setDefaultVerbosity(verbose);
        logger.setExceptionMap(verbosities);
        log("Setting default verbosity to "+verbose,4);
        
        MathContext context = new MathContext(precision,RoundingMode.HALF_EVEN);
        Contexter contexter = Contexter.getInstance();
        contexter.setContext(context);
        log("Setting precision to "+precision,4);
        //BigDecimal deczero = contexter.getDecimalZero();
        //log("Decimal zero is "+deczero.toPlainString()+" with precision "+deczero.precision()+" and scale "+deczero.scale(),4);

        // preprocessing
        if (run_preprocessor) {
	    logger.changeVerbosity("PREPROCESSOR");
            try {
                Preprocessor.preprocess(filename,constants,add_free);
            }
            catch(Exception e) {
                System.err.println("Error while preprocessing program. Aborting. ");
                e.printStackTrace(System.err);
                System.exit(1);
            }
            filename += ".4.pp";
            f = new File(filename);
        }else{
            if (!filename.endsWith(".pp")){
                filename = filename + ".4.pp";
            }
        }

        

       /* if (run_analyzer) {
            Analyzer anal = new Analyzer();
            anal.analyze(filename,printMc);
        }*/


        if (run_matrix){
            MatrixComputator mc = new MatrixComputator(iterations);
            mc.computeMatrices(filename);
        }
        
//        if (run_explorer){
//                ConcurrentExplorer exp = new ConcurrentExplorer(threads);
//                exp.explore(filename);
//            
//            
//        }
        
        if (run_hybrid){
            ConcurrentHybridExplorer exp = new ConcurrentHybridExplorer(iterations,threshold,timelimit,threads,draw_cfg);
            exp.explore(filename);
        }
        
        
        
        // clean
        if (run_preprocessor && clean) {
            f.delete();
        }
    }   
}
