/*  HyLeakL is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2014-2017 - 

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: 
*/

package hyleak;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Set;
import support.Logger;
import theory.BigIntegerCartesianIterator;

/**
 *
 * @author  
 */
public class IntervalAssignments extends HashMap<String, Interval> implements Comparable<IntervalAssignments>{

	public IntervalAssignments(VariableList secrets) {
        for (Variable secretvar : secrets){
                Interval interv = ((PrivateVariable)secretvar).getInterval();
                this.put(secretvar.getName(), interv);
            }
    }
    
    public IntervalAssignments(HashMap<String, Interval> map) {
        this.putAll(map);
    }

    public String printNice() {
        String toreturn = "";
        Set<String> keysSet = this.keySet();
        ArrayList<String> keys = new ArrayList<>();
        keys.addAll(keysSet);
        Collections.sort(keys);
        for (String key : keys){
            toreturn +=key+"="+this.get(key).toString()+',';
        }
        return toreturn.substring(0, toreturn.length()-1); // to remove the last comma
    }
    
    public boolean equals(IntervalAssignments other){
        //log("Checking equivalence of interval assignments "+this.printNice()+" and "+other.printNice(),4);
        for (String myvar:this.keySet()){
            //log("Checking variable "+myvar,4);
            if (!other.containsKey(myvar)){
                //log("Different: Other interval assignment does not have variable "+myvar,4);
                return false;
            }else if (!this.get(myvar).equals(other.get(myvar))){
                //log("Different: My interval has assignment "+this.get(myvar)+" while other interval has assignment "+other.get(myvar),4);
                return false;
            }
        }
        //log("Interval assignments are equivalent",4);
        return true;
    }
    
    public BigInteger getSize(){
        BigInteger toreturn = BigInteger.ONE;
        for (Interval interv : this.values()){
            BigInteger thissize = interv.getSize();
            toreturn = toreturn.multiply(thissize);
        }
        return toreturn;
    }
    
    public boolean hasEmptyInterval() {
        for (Interval interv : this.values()){
            if (interv.isEmpty()){
                return true;
            }
        }
        return false;
    }
    
    public ArrayList<IntervalAssignments>[] splitIntervals(IntervalAssignments newassignment) {
        // [0] is intersection
        // [1] is restofold
        // [2] is restofnew
        
        ArrayList<IntervalAssignments>[] toreturn = new ArrayList[3];
        toreturn[0] = new ArrayList<>();
        toreturn[1] = new ArrayList<>();
        toreturn[2] = new ArrayList<>();
        
        
        HashMap<String, Interval> mapintersection = new HashMap();
        //HashMap<String, Interval> maprestofold = new HashMap();
        //HashMap<String, Interval> maprestofnew = new HashMap();
        
        for (String varname : this.keySet()){
            log("Now splitting variable "+varname,3);
            Interval myinterv = this.get(varname);
            Interval hisinterv = newassignment.get(varname);
            Interval intersection = myinterv.intersect(hisinterv);
            log ("Intersection of "+varname+" is "+intersection.toString(),3);
            mapintersection.put(varname, intersection);
            //maprestofold.put(varname, myinterv.subtract(intersection));
            //maprestofnew.put(varname,hisinterv.subtract(intersection));
        }
        
        // if the intersection has empty intervals, there are no shared assignments
        IntervalAssignments intersection = new IntervalAssignments(mapintersection);
        if (intersection.hasEmptyInterval()){
            log("Intersection has size zero, no assignments in common",3);
            toreturn[1].add(this);
            toreturn[2].add(newassignment);
            return toreturn;
        }
        // if there is an intersection, we need to unroll all the assignments and compute them appropriately
        log("Intersection has size "+intersection.getSize().toString()+", computing rest of old and rest of new",3);
        toreturn[0].addAll(intersection.unroll());
        toreturn[1].addAll(this.unroll());
        toreturn[2].addAll(newassignment.unroll());
        if (Logger.getInstance().getVerbosity()>2){
            log("After unrolling:",3);
            log("intersection: ",3);
            for (IntervalAssignments ia : toreturn[0]) log(ia.printNice(),3);

            log("restofold: ",3);
            for (IntervalAssignments ia : toreturn[1]) log(ia.printNice(),3);

            log("restofnew: ",3);
            for (IntervalAssignments ia : toreturn[2]) log(ia.printNice(),3);
       
        }
        log("Removing duplicates",3);
        for (IntervalAssignments int_ia : toreturn[0]){
            log("Removing "+int_ia.printNice(),3);
            ArrayList<IntervalAssignments> removefromold = new ArrayList();
            for (IntervalAssignments old_ia : toreturn[1]){
                if (int_ia.equals(old_ia)) {
                    removefromold.add(old_ia);
                    log("Removed duplicate "+old_ia.printNice()+" from rest of old",3);
                }
                
            }
            toreturn[1].removeAll(removefromold);
            
            ArrayList<IntervalAssignments> removefromnew = new ArrayList();
            for (IntervalAssignments new_ia : toreturn[2]){
                if (int_ia.equals(new_ia)) {
                    log("Removed duplicate "+new_ia.printNice()+" from rest of new",3);
                    removefromnew.add(new_ia);
                }
                
            }
            toreturn[2].removeAll(removefromnew);
        }
        if (Logger.getInstance().getVerbosity()>2){
            log("I will return the following:",3);
            log("intersection: ",3);
            for (IntervalAssignments ia : toreturn[0]) log(ia.printNice(),3);

            log("restofold: ",3);
            for (IntervalAssignments ia : toreturn[1]) log(ia.printNice(),3);

            log("restofnew: ",3);
            for (IntervalAssignments ia : toreturn[2]) log(ia.printNice(),3);
       
        }
        return toreturn;
    }
    
    String[] getAllValues(){
        String[] vallabels = new String[this.getSize().intValue()];
        ArrayList<String> secretvars = new ArrayList();
        secretvars.addAll(this.keySet());
        Collections.sort(secretvars);
        
        
        Iterable[] iters = new Iterable[secretvars.size()];
        
        
        for (int i=0; i<secretvars.size(); i++){
            iters[i] = this.get(secretvars.get(i));
        }
        //CartesianIterator cartiter = new CartesianIterator(iters);
        int pos = 0;
       
        for (BigIntegerCartesianIterator cartiter = new BigIntegerCartesianIterator(iters); cartiter.hasNext();){
            Object[] values = cartiter.next();
            vallabels[pos]="";
            for (int i=0; i<values.length; i++){
                BigInteger value =(BigInteger) values[i];
                vallabels[pos] += secretvars.get(i)+'='+value.intValue()+';';
            }
            pos++;
        }
        return vallabels;
    }
    
    ArrayList<IntervalAssignments> unroll(){
        ArrayList<IntervalAssignments> toreturn = new ArrayList<>();
        //String[] vallabels = new String[this.getSize().intValue()];
        ArrayList<String> secretvars = new ArrayList();
        secretvars.addAll(this.keySet());
        Collections.sort(secretvars);
        
        
        Iterable[] iters = new Iterable[secretvars.size()];
        
        
        for (int i=0; i<secretvars.size(); i++){
            iters[i] = this.get(secretvars.get(i));
        }
        //CartesianIterator cartiter = new CartesianIterator(iters);
        //int pos = 0;
       
        for (BigIntegerCartesianIterator cartiter = new BigIntegerCartesianIterator(iters); cartiter.hasNext();){
            Object[] values = cartiter.next();
            HashMap<String,Interval> onevalue = new HashMap<>();
            for (int i=0; i<values.length; i++){
                BigInteger value =(BigInteger) values[i];
                if (value == null) {
                    log("WARNING: null value from BigIntegerCartesianIterator?",1);
                    System.exit(1);
                }
                onevalue.put(secretvars.get(i), new Interval(value));
                //vallabels[pos] += secretvars.get(i)+'='+value.intValue()+';';
            }
            toreturn.add(new IntervalAssignments(onevalue));
            //pos++;
        }
        return toreturn;
    }
        
    private static void log(Object aObject, int level){
        Logger loggy = Logger.getInstance();
        loggy.log(String.valueOf(aObject),level);
    }

    @Override
    public int compareTo(IntervalAssignments other) {
        if (this.size()!=other.size()) return Integer.compare(this.size(), other.size());
        if (this.equals(other)) return 0;
        // where are they different?
        
        // smaller first
        BigInteger mySize = this.getSize();
        BigInteger otherSize = other.getSize();
        if (mySize.compareTo(otherSize)!=0) {
            if (mySize.compareTo(otherSize)<0) return -1;
            else return 1;
        }
        
        // same size. same variables?
        ArrayList<String> myVars = new ArrayList<>();
        myVars.addAll(this.keySet());
        Collections.sort(myVars);
        ArrayList<String> otherVars = new ArrayList<>();
        otherVars.addAll(this.keySet());
        Collections.sort(otherVars);
        for (int i=0; i<myVars.size(); i++){
            String myVar = myVars.get(i);
            String otherVar = otherVars.get(i);
            int comp = myVar.compareTo(otherVar);
            if (comp!=0) return comp;
        }
        
        // ok they also have the same variables. now we compare the intervals for each variable
        for (int i=0; i<myVars.size(); i++){
            Interval myInterv = this.get(myVars.get(i));
            Interval otherInterv = other.get(myVars.get(i));
            int comp = myInterv.compareTo(otherInterv);
            if (comp!=0) return comp;
        }
        //this should never happen
        log("WARNING: IntervalAssignments comparison failed",1);
        System.exit(1);
        return 0;
    }


}
