/*  HyLeak is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2014-2017 - 

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: 
*/

package hyleak;

import java.math.BigDecimal;
import java.util.HashMap;

import distributions.MultiIntervalProbabilityDistribution;
import support.Contexter;
import support.Logger;

/**
 *
 * @author  
 */
public class PosteriorDistributionManager{

    
    private static final Contexter contexter = Contexter.getInstance();
    private static final Logger logger = Logger.getInstance();
    
    private final HashMap<String,MultiIntervalProbabilityDistribution> outputs = new HashMap();
    private final WeightedProbabilityDistribution outputdist = new WeightedProbabilityDistribution();
    IntervalAssignments initialsecretintervals;
  

    PosteriorDistributionManager(HashMap<String, Interval> initsecint) {
       if (initsecint == null){
           initialsecretintervals = null;
       }else{
           initialsecretintervals = new IntervalAssignments(initsecint);
       }
    }
    
    public BigDecimal computePosteriorShannonEntropy() {
        BigDecimal toreturn = contexter.getDecimalZero();
        //outputdist.normalize();
        BigDecimal totobsprob = outputdist.getTotalProbability();
        boolean willpad = false;
        log("I observed "+outputdist.size()+" outputs for a total observed probability of "+totobsprob.toPlainString(),2);
        if (totobsprob.compareTo(contexter.getDecimalOne())<0 && initialsecretintervals!=null){
            willpad = true;
            log("Padding will be applied.",2);
        }
       
        
        for (String aout : outputdist.keySet()){
            BigDecimal prob = outputdist.getProbability(aout);
            log("Output "+aout+" has probability "+prob+".\nIts conditional probability distribution is",3);
            BigDecimal entropy = contexter.getDecimalZero();
            MultiIntervalProbabilityDistribution dist = outputs.get(aout);
            if (willpad){
                log("Posterior distribution on the secrets before padding and normalizing:",3);
            }else{
                log("Posterior distribution on the secrets before normalizing:",3);
            }
            log(dist.printNice(),3);
            
            if (willpad){
                //pad
                dist.pad(initialsecretintervals);
            }
            //normalize
            dist.normalize();
            
            if (willpad){
                log("Posterior distribution on the secrets after padding and normalizing:",3);
            }else{
                log("Posterior distribution on the secrets after normalizing:",3);
            }
            log(dist.printNice(),3);
            
            //get entropy
            log(dist.printNice(),3);
            entropy = entropy.add(dist.getEntropy(),contexter.getContext());
            log("Its conditional entropy is "+entropy,3);
                    
                    
            BigDecimal contr = prob.multiply(entropy, contexter.getContext());
            log("Output "+aout+" contributes "+contr+" bits to the posterior entropy estimate\n",3);
            toreturn = toreturn.add(contr);
        }
        // assume that the unobserved part of the system has the same leakage as the observed part
        // totobsprob = totobsprob.multiply(new BigDecimal(100));
        log("Conditional entropy after observing "+(totobsprob.multiply(new BigDecimal(100))).toEngineeringString()+"% of the probability space is "+toreturn.toEngineeringString(),2);
        toreturn = toreturn.divide(totobsprob, contexter.getContext());
        
        return toreturn;
    }
    
    private static void log(Object aObject, int level){
        Logger loggy = Logger.getInstance();
        loggy.log(String.valueOf(aObject),level);
    }
    
    public void inferInformation(MarkovChainState finalstate) {
        //BigDecimal entropy = new BigDecimal(0);
        //BigInteger newsecretsize = new BigInteger("0");
        VariableList secrets = new VariableList();
        VariableList observables = new VariableList();
        VariableList vars = finalstate.getVariables();
        
      
        // TEMPORARY - TESTING ONLY
       // logger.setVerbosity(3);
        
        for (Variable var : vars){
            if (var.isPrivate()){
                
                PrivateVariable privvar =(PrivateVariable)var;
                if (privvar.isSecret()){
                    secrets.add(privvar);            
                }
                
                //newsecretsize = newsecretsize.add(asecret.getSize());
            }else{
                PublicVariable pubvar =(PublicVariable)var;
                if (pubvar.isObservable()){
                    observables.add(pubvar);            
                }
            }
        }
        log("",3);
        log("Observation: "+observables.getIdentifier()+"\nSecret: "+secrets.getIdentifier(),3);
        
        // use the name and value of the observables as key. the identifier is a string like "o=17;"
        String key = observables.getIdentifier();
        BigDecimal stateprob = finalstate.getProbability();
        String varidentifier = vars.getIdentifier();
        
        log("Updating probability distribution after observing "+key+" with variables "+varidentifier+" and probability "+stateprob.toPlainString(),3);
        
        // outputdist is the probability distributions over all observed outputs
        outputdist.update(key, varidentifier, stateprob);
        log("The new distribution on the observables is:\n"+outputdist.printNice(),3);
        log("with total probability "+outputdist.getTotalProbability(),3);
        
        log("Updating conditional posterior distribution for "+key,3);
        
// outputs maps each possible output to the corresponding posterior distributions over the secrets
        if (outputs.containsKey(key)){ //the output has already appeared; update the corresponding prob dists with the new information about the secrets
            
            MultiIntervalProbabilityDistribution mapsecrets =(MultiIntervalProbabilityDistribution) outputs.get(key);
            log("Old distribution on the secret if the observable is "+key+":\n"+mapsecrets.printNice(),3); 
            mapsecrets.update(secrets,varidentifier,stateprob);
            /*for (MarkovStateVariable secretvar : secrets){
                IntervalProbabilityDistribution dist = mapsecrets.get(secretvar.getName());
                Interval interv = ((MarkovStatePrivateVariable)secretvar).getInterval();
                dist.update(interv,varidentifier,stateprob);
            }*/
            
            log("New distribution on the secret if the observable is "+key+":\n"+mapsecrets.printNice(),3); 

        }else{ // the output has been observed for the first time; create a new prob dist
            MultiIntervalProbabilityDistribution newmap = new MultiIntervalProbabilityDistribution();
            newmap.initialize(secrets,varidentifier,stateprob);
            outputs.put(key, newmap); // maps this output to the new prob dist 
            log("Created new distributions for output "+key,3);
        }

        //TEMPORARY - TESTING ONLY
        //logger.setVerbosity(2);
    }

    BigDecimal getTotalProbability() {
        return outputdist.getTotalProbability();
    }

    WeightedProbabilityDistribution getDistributionOnOutputs() {
        return outputdist;
    }
}
