/*  HyLeak is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2014-2017 - 

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: 
*/

package hyleak;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.math.MathContext;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;
import support.Contexter;
import support.Logger;
import theory.BigFunctions;

/**
 *
 * @author  
 */
class SimpleMultiIntervalProbabilityDistribution extends ConcurrentHashMap< IntervalAssignments , BigDecimal >{
    protected static final MathContext context = Contexter.getInstance().getContext();
    protected static final Logger logger = Logger.getInstance();
  
    public String printNice() {
        String toreturn = "";
        Set<IntervalAssignments> keysSet = this.keySet();
        ArrayList<IntervalAssignments> keys = new ArrayList<>();
        keys.addAll(keysSet);
        Collections.sort(keys);
        for (IntervalAssignments key : keys){
            
            toreturn +="assignment "+key.printNice()+" with probability "+this.get(key).toEngineeringString()+"\n";
        }
        return toreturn;
    }

    private static void log(Object aObject, int level){
        if (logger.getVerbosity() >= level) {
            Logger loggy = Logger.getInstance();
            loggy.log(String.valueOf(aObject),level);
        }
    }

    void update(VariableList secrets, BigDecimal stateprob) {
        this.update(new IntervalAssignments(secrets),stateprob);
    }
    
    synchronized void update(IntervalAssignments bignewassignment, BigDecimal prob) {
        log("Updating with assignment "+bignewassignment.printNice()+" and probability "+prob.toEngineeringString(),3);
        
        // if the new assignment already exists, just update its value
        for (IntervalAssignments oldassignment : this.keySet()){
            if (oldassignment.equals(bignewassignment)){
                log("The assignment already exists; updating its probability",4);
                this.addProbability(oldassignment,prob);
                return;
            }
        }
        
        /* 
        if the assignment does not exist, it could intersect with any of the 
        others. Also, we may have to split it, so we need to keep some 
        collection. We will put everything in the new map newdist, and in the end
        we replace the content of the current distribution with newdist.
        */
        HashMap< IntervalAssignments , BigDecimal > newdist = new HashMap();
        
        // unroll all the current assignments and put them in with the appropriate probability
        for (IntervalAssignments ia : this.keySet()){
            BigDecimal myprob = this.get(ia);
            ArrayList<IntervalAssignments> unrolled = ia.unroll();
            if (unrolled.size()>1){
                myprob = myprob.divide(new BigDecimal(unrolled.size()), context);
            }
            for (IntervalAssignments unria : unrolled){
                newdist.put(unria, myprob);
            }
        }
        
        // unroll the new assignment
        ArrayList<IntervalAssignments> newunrolled = bignewassignment.unroll();
        if (newunrolled.size()>1){
            prob = prob.divide(new BigDecimal(newunrolled.size()), context);
        }
        
        for (IntervalAssignments newia : newunrolled){
            boolean found = false;
            for (IntervalAssignments oldassignment : newdist.keySet()){
                if (oldassignment.equals(newia)){
                    log("The assignment already exists; updating its probability",4);
                    newdist.put(oldassignment,prob.add(newdist.get(oldassignment),context));
                    found = true;
                    break;
                }
            }
            if (found == false){
                newdist.put(newia, prob);
            }
        }
        
        
        
        // everything is done, we replace the current distribution with newdist
        this.clear();
        this.putAll(newdist);
    }
    
    synchronized void OLDupdate(IntervalAssignments bignewassignment, BigDecimal prob) {
        log("Updating with assignment "+bignewassignment.printNice()+" and probability "+prob.toEngineeringString(),3);
        Set<IntervalAssignments> keysSet = this.keySet(); 
        ArrayList<IntervalAssignments> keys = new ArrayList<>();
        
        keys.addAll(keysSet);
        Collections.sort(keys);
        // if the new assignment already exists, just update its value
        for (IntervalAssignments oldassignment : keys){
            if (oldassignment.equals(bignewassignment)){
                log("The assignment already exists; updating its probability",4);
                this.addProbability(oldassignment,prob);
                return;
            }
        }
        
        ArrayList<IntervalAssignments> newassignmentset = new ArrayList<>();
        newassignmentset.add(bignewassignment);
        HashMap< IntervalAssignments , BigDecimal > newdist = new HashMap();
        
        
        for (IntervalAssignments oldassignment : keys){
            log("entering loop",4);
            if (newassignmentset.isEmpty()){
                newdist.put(oldassignment, this.get(oldassignment));
                log("new assigment set is empty",4);
                continue;
            }
            IntervalAssignments newassignment = newassignmentset.iterator().next();
            if (oldassignment.equals(newassignment)){
                newdist.put(oldassignment, this.get(oldassignment).add(prob,context));
                log("new and old assignments coincide, updating probability",4);
                newassignmentset.remove(newassignment);
                continue;
            }
            //log("Old HFP: "+cloneoldHFP.printNice(),3);
            // we have the old assignment and the new assignment; we obtain their intersection, the rest of the old assignment without the intersection, and the rest of the new assignment without the interseciton
            BigInteger oldassignmentsize = oldassignment.getSize();
            BigInteger newassignmentsize = newassignment.getSize();
            BigInteger oldTotalSize = oldassignmentsize.add(newassignmentsize);
            log("Splitting "+oldassignment.printNice()+" with "+newassignment.printNice(),3);
            ArrayList<IntervalAssignments>[] splitted = oldassignment.splitIntervals(newassignment);
            log("\nSplitting current assignment "+oldassignment.printNice()+" with probability "+this.get(oldassignment).toEngineeringString()+" with new assignment "+newassignment.printNice()+" with probability "+prob.toEngineeringString(),4);
            ArrayList<IntervalAssignments> intersection = splitted[0];
            ArrayList<IntervalAssignments> restofold    = splitted[1];
            ArrayList<IntervalAssignments> restofnew    = splitted[2];
            if (logger.getVerbosity()>1){
                log("Result of split:",4);
                log("intersection: ",4);
                for (IntervalAssignments ia : intersection) log(ia.printNice(),4);
                
                log("restofold: ",4);
                for (IntervalAssignments ia : restofold) log(ia.printNice(),4);
                
                log("restofnew: ",4);
                for (IntervalAssignments ia : restofnew) log(ia.printNice(),4);
            }
            
            
            BigInteger intersectionsize = BigInteger.ZERO;
            for (IntervalAssignments ia : intersection) intersectionsize = intersectionsize.add(ia.getSize());
            
            BigInteger restofoldsize = BigInteger.ZERO;
            for (IntervalAssignments ia : restofold) restofoldsize = restofoldsize.add(ia.getSize());
            
            BigInteger restofnewsize = BigInteger.ZERO;
            for (IntervalAssignments ia : restofnew) restofnewsize = restofnewsize.add(ia.getSize());
            
            BigInteger newTotalSize = intersectionsize.add(intersectionsize).add(restofoldsize).add(restofnewsize);
            if (oldTotalSize.compareTo(newTotalSize)!=0){
                // intersection failed. This should never happen.
                
                log("\nERROR: size mismatch; secrets lost in updating. Please file a bug report.",1);
                log("Old assignment:"+oldassignment.printNice()+", size = "+oldassignment.getSize(),1);
                log("New assignment:"+newassignment.printNice()+", size = "+newassignment.getSize(),1);
                log("",1);
                log("Intersection: ",1);
                for (IntervalAssignments ia : intersection) log(ia.printNice(),1);
                log("Intersection size = "+intersectionsize,1);
                log("Rest of old: ",1);
                for (IntervalAssignments ia : restofold) log(ia.printNice(),1);
                log("Rest of old size = "+restofoldsize,1);
                log("Rest of new: ",1);
                for (IntervalAssignments ia : restofnew) log(ia.printNice(),1);
                log("Rest of new size = "+restofnewsize,1);
                System.exit(1);
            }

            
            // old interval: change probability to account for reduced size and put in
            if (restofoldsize.compareTo(BigInteger.ZERO)>0){
                BigDecimal newprob = this.get(oldassignment);
                //BigInteger restofoldsize = restofoldsize;
                if (oldassignmentsize.compareTo(restofoldsize)!=0){
                    BigDecimal tomult = new BigDecimal(restofoldsize,context);
                    tomult = tomult.divide(new BigDecimal(oldassignmentsize,context), context);
                    newprob = newprob.multiply(tomult, context);
                    // probability has to be divided over all elements of rest of old
                    if (restofold.size()>1){
                        newprob = newprob.divide(new BigDecimal(restofold.size()), context);
                    }
                    
                }
                log("Putting the "+restofold.size()+" elements of rest of old back in with probability for each of "+newprob,4);
                for (IntervalAssignments ia : restofold) newdist.put(ia, newprob);
                //newdist.put(restofold, newprob);
            }else{
                log("The rest of old has an empty interval",4);
            }
            
            // intersection: take probability by size from old interval, then update with prob by size from new interval (freq is freq of old plus one)
            if (intersectionsize.compareTo(BigInteger.ZERO)>0){
                
                BigDecimal intersprob =  this.get(oldassignment);
                if (oldassignmentsize.compareTo(intersectionsize)!=0){
                    BigDecimal tomult = new BigDecimal(intersectionsize).divide(new BigDecimal(oldassignmentsize), context);
                    intersprob = intersprob.multiply(tomult,context);
                }
                BigDecimal toadd = prob;
                toadd = toadd.multiply(new BigDecimal(intersectionsize,context), context);
                toadd = toadd.divide(new BigDecimal(newassignmentsize,context), context);
                
                intersprob = intersprob.add(toadd, context);
                if (intersection.size()>1){
                        intersprob = intersprob.divide(new BigDecimal(intersection.size()), context);
                    }
                log("The intersection has "+intersection.size()+" elements, putting them in in with probability "+intersprob.toEngineeringString(),4);
                for (IntervalAssignments ia : intersection) newdist.put(ia, intersprob);
                //newdist.put(intersection, intersprob);
                
                // change remaining probability to reflect the fact that some of it has gone to the intersection
                prob = prob.subtract(toadd, context);
                
            }else{
                log("The intersection has an empty interval",4);
            }
            
            
            // new interval: change probability to account for reduced size, wait for next iteration
            newassignmentset = restofnew;
            log("The remainder is :",4);
            for (IntervalAssignments ia : newassignmentset) log(ia.printNice(),4);
            log("with probability "+prob.toEngineeringString(),4);
            log("exiting loop, reminder size is "+newassignmentset.size(),4);
        }
        // if new interval is now not empty, put it with the remaining probability and frequency 1
        log("Exited the loop, reminder set size is "+newassignmentset.size(),4);
        
        if (!newassignmentset.isEmpty()){
            if (newassignmentset.size()>1){
                prob = prob.divide(new BigDecimal(newassignmentset.size()), context);
            }
            log("Inserting the remainder of the new assignment with probability "+prob.toPlainString(),4);
            for (IntervalAssignments ia : newassignmentset) newdist.put(ia, prob);
            //newdist.put(newassignment, prob);
        }
        this.clear();
        this.putAll(newdist);
    }

    void initialize(VariableList secrets, BigDecimal stateprob) {
        IntervalAssignments newkey = new IntervalAssignments(secrets);
        
       
        
        this.put(newkey, stateprob);
        
    }


    

    public BigDecimal getEntropy() {
        BigDecimal entropy = Contexter.getInstance().getDecimalZero();
        for (IntervalAssignments inter : this.keySet()){
            BigInteger intervsize = inter.getSize();
            if (intervsize.compareTo(BigInteger.ZERO)!=0){
                BigDecimal prob = this.get(inter);
                BigDecimal localent = prob.multiply(log2(prob.divide(new BigDecimal(intervsize), context)),context);
            
                log("Local entropy: "+localent,4);
                entropy = entropy.subtract(localent, context);
            }
            
            

        }
        return entropy;

    }
    
    protected BigDecimal log2(BigDecimal value) {
        if (value.compareTo(Contexter.getInstance().getDecimalOne())==0){
            return Contexter.getInstance().getDecimalZero();
        }
        log("Computing the logarithm in base 2 of "+value.toPlainString()+" with scale "+value.scale(),4);
        BigDecimal toreturn = BigFunctions.ln(value,context.getPrecision());
        log("The natural logarithm is "+toreturn,4);
        toreturn = toreturn.divide(BigFunctions.ln(new BigDecimal("2",context),context.getPrecision()),context);
        log("The logarithm in base 2 is "+toreturn,4);
        
        return toreturn;
    }

    public void normalizeTo(BigDecimal prob) {
        BigDecimal currprob = this.getTotalProbability();
        if (currprob.compareTo(prob)!=0){
            BigDecimal multiplier = prob.divide(currprob, context);
            for (IntervalAssignments interv : this.keySet()){
                this.put(interv, this.get(interv).multiply(multiplier, context));
            }
        }
    }
    
    public void normalize() {
        this.normalizeTo(Contexter.getInstance().getDecimalOne());

    }

    private BigDecimal getTotalProbability() {
        BigDecimal toreturn = Contexter.getInstance().getDecimalZero();
        for (BigDecimal prob: this.values()){
            toreturn = toreturn.add(prob, context);
        }
        return toreturn;
    }

    private void addProbability(IntervalAssignments newassignment, BigDecimal prob) {
        BigDecimal currprob = this.get(newassignment);
        this.put(newassignment, currprob.add(prob, context));
    }

    BigDecimal getHighestProbability() {
        BigDecimal max = Contexter.getInstance().getDecimalZero();
        for (BigDecimal prob: this.values()){
            if (prob.compareTo(max)>0){
                max = prob;
            }
        }
        return max;
    }
}
