/*  HyLeak is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2014-2017 - 

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: 
*/

grammar Preprocessor1;

@parser::header {
    package preprocessor;
    
    import java.math.BigInteger;
    import java.math.BigDecimal;
    import java.util.Iterator;
    import java.util.LinkedList;
    import support.Logger;
}

@lexer::header {
    package preprocessor; 
}

@members {
    Programme prog;
    Logger logger = Logger.getInstance();
    
}

prog_grammar[Programme p] returns [String str] 			@init{ 	$str = "";
									prog=p; 
								}:
		( decl						{ $str += $decl.str; }
		)*
		st_list 					{ $str += $st_list.str; }
		EOF
		;

decl returns [String str]:	
		const_decl					{ $str = ""; }
		| public_decl					{ $str = $public_decl.str; }
		| private_decl 					{ $str = $private_decl.str; }
		;

const_decl:	'const' IDENTIFIER ':=' 
		( val=assign_expr[false] 			{ try {
								      BigDecimal int_val = new BigDecimal($val.str);
								      if (!prog.addConstant($IDENTIFIER.text,int_val)) {
									  logger.log("WARNING: on line " + $IDENTIFIER.getLine() + ", constant " + $IDENTIFIER.text + " is already declared.",1);
								      }
								  }
								  catch(NumberFormatException e) {
								      logger.log("ERROR: on line " + $IDENTIFIER.getLine() + ", " + $val.str + " cannot be converted to integer.",1);
								      System.exit(1);
								  }
								}
		| fval=float_number				{ if (!prog.addConstant($IDENTIFIER.text,$fval.f)) {
								      logger.log("WARNING: on line " + $IDENTIFIER.getLine() + ", constant " + $IDENTIFIER.text + " is already declared.",1);
								  }
								}
		)
		';'
		;
	  
public_decl returns [String str]: 	public_type var_type IDENTIFIER init_val ';'
					{   PublicVariable v = new PublicVariable(1);
					    v.setPublicity($public_type.b);
					    if (!prog.addPublicVariable($IDENTIFIER.text,v)) {
						logger.log("WARNING: on line " + $IDENTIFIER.getLine() + ", variable " + $IDENTIFIER.text + " is already declared.",1);
					    }
					    else{
						v.setName($IDENTIFIER.text);
					    }
					    $str = $public_type.text + " " + $var_type.str + " " + $IDENTIFIER.text + " " + $init_val.str + ";\n"; 
					}
	   ;

public_type returns [boolean b]:
	   'public' 		{ $b = true; }
	   | 'observable' 	{ $b = false; }
	   ;
	   
private_decl returns [String str]:	private_type var_type IDENTIFIER init_intervals ';'
					{   PrivateVariable v = new PrivateVariable(1);
					    v.setPublicity($private_type.b);
					    if (!prog.addPrivateVariable($IDENTIFIER.text,v)) {
						logger.log("WARNING: on line " + $IDENTIFIER.getLine() + ", variable " + $IDENTIFIER.text + " is already declared.",1);
					    }
					    else {
						v.setName($IDENTIFIER.text);
					    }
					    $str = $private_type.text + " " + $var_type.str + " " + $IDENTIFIER.text + " " + $init_intervals.str + ";\n"; 
					}
					;
	      
private_type returns [boolean b]:
	    'private'		{ $b = true; }
	    | 	'secret' 	{ $b = false; }
	    ;

/* var_type must have defined integer values. 
 * Expressions are replaced by this preprocessor.
*/
var_type returns [String str] :
	INT			
	{   String sizeString = $INT.text.substring(3);
	    int sizeInt = 0;
	    try {
		sizeInt = Integer.parseInt(sizeString);
	    }
	    catch (NumberFormatException e) {
		BigDecimal val = prog.getConstantValue(sizeString);
		if (val != null)
		    sizeInt = val.intValue();
		else {
		    logger.log("ERROR: on line " + $INT.getLine() + ", unknown constant: " + sizeString,1);
		    System.exit(1);
		}
	    }
	    $str = "int" + sizeInt;
	}
	| a='array' '[' e1=assign_expr[false] ('..' e2=assign_expr[false])? ']' 'of' w=var_type
	{   $str = "array [";
	    String smin = "0";
	    String smax = $e1.str;
	    if (e2 != null) {
		smin = smax;
		smax = $e2.str;
	    }
	    try {
		int min = Integer.parseInt(smin);
		int max = Integer.parseInt(smax);
		if (e2 == null)
		    max -= 1;
		if (max < min) {
		    logger.log("ERROR: on line " + a.getLine() + ", array size is zero.",1);
		    System.exit(1);
		}
		$str += min + ".." + max + "] of " + $w.str;
	    }
	    catch(NumberFormatException e) {
		logger.log("ERROR: on line " + a.getLine() + ", array size " + smin + " or " + smax + " cannot be converted to integer.",1);
		System.exit(1);
	    }
	}
	;

/* init_val must have defined integer values. 
 * Expressions are replaced by this preprocessor.
 */
init_val returns [String str] :	
									{ $str = ""; }	// empty
		| a=':='  						{ $str = ":="; }
			  ( val=assign_expr[false]			{ try {
									      BigInteger int_val = new BigInteger($val.str);
									      $str += int_val;
									  }
									  catch(NumberFormatException e) {
									      logger.log("ERROR: on line " + a.getLine() + ", " + $val.str + " cannot be converted to integer.",1);
									      System.exit(1);
									  }
									}
			  | number_list					{ $str += $number_list.str; }
			  )
		;
		
/* number_list must have defined integer values. 
 * Expressions are replaced by this preprocessor.
 */
number_list returns [String str] :
		 a='{' val1=assign_expr[false] 				{ try {
									      BigInteger int_val = new BigInteger($val1.str);
									      $str = "{" + int_val;
									  }
									  catch(NumberFormatException e) {
									      logger.log("ERROR: on line " + a.getLine() + ", " + $val1.str + " cannot be converted to integer.",1);
									      System.exit(1);
									  }
									}
		(',' val2=assign_expr[false] 				{ try {
									      BigInteger int_val = new BigInteger($val2.str);
									      $str += "," + int_val;
									  }
									  catch(NumberFormatException e) {
									      logger.log("ERROR: on line " + a.getLine() + ", " + $val2.str + " cannot be converted to integer.",1);
									      System.exit(1);
									  }
									}
		)*
		'}'							{ $str += "}"; }
		;
		
init_intervals returns [String str] @init { $str = ""; }:		
									// empty
		| ':=' 							{ $str = ":= "; }
		  (interval						{ $str += $interval.str; }
		  )+				  		        
		;

/* interval must have defined integer values. 
 * Expressions are replaced by this preprocessor.
 */
interval returns [String str]: 	
		a='[' min=assign_expr[false] ',' max=assign_expr[false] ']'
		{   try {
			BigInteger minVal = new BigInteger($min.str);
			BigInteger maxVal = new BigInteger($max.str);
			$str = "[" + minVal + "," + maxVal + "]"; 
		    }
		    catch(NumberFormatException e) {
			logger.log("ERROR: on line " + $a.getLine() + ", " + $min.str + " or " + $max.str + " cannot be converted to integer.",1);
			System.exit(1);
		    }
		}
		;

st_list returns [String str] @init { str = ""; } :
				( st 					{ $str += $st.str; }
				)*  				
				;
	
st returns [String str]:	assignement_st				{ $str = $assignement_st.str; }
				| print_st 				{ $str = $print_st.str; }
				| return_st				{ $str = $return_st.str; }
				| simulate_st				{ $str = $simulate_st.str; }
				| random_st				{ $str = $random_st.str; }
				| while_st				{ $str = $while_st.str; }
				| for_st				{ $str = $for_st.str; }
				| if_st					{ $str = $if_st.str; }
				;
  
assignement_st returns [String str]:	'assign' var=variable[false] ':=' e=assign_expr[false] ';' 	{ $str = "assign " + $var.str + " := " + $e.str + ";\n"; };

print_st returns [String str]:		'print' ';' 						{ $str = "print;\n"; };

return_st returns [String str]:		
		'return' ';'									{ $str = "return;\n"; }
		| 'return' 'after' 'simulate' 
		  var1=variable[false] ':=' filename '(' var2=variable[true] ')' ';'		{ $str = "return after simulate " + $var1.str + " := " + $filename.text + " (" + $var2.str + ");\n"; }
		;

simulate_st returns [String str]:		
		'simulate' ';' 					{ $str = "simulate;\n"; }
		| 'simulate-abs' ';' 				{ $str = "simulate-abs;\n"; }
		;

random_st returns [String str]:	
		ran='random' var=variable[false] ':='		{ $str = "random " + $var.str + " := "; }
		( 'randombit' '(' 				{ $str += "randombit("; }
				  ( float_number 		{ if ($float_number.f.signum() < 0 || $float_number.f.compareTo(BigDecimal.ONE) == 1) {
								    logger.log("ERROR: on line " + $ran.getLine() + ", float value " +  $float_number.f.toPlainString() + " must be in [0,1].",1);
								    System.exit(1);
								  }
								  $str += $float_number.f;
								}
				  | IDENTIFIER 			{ BigDecimal val = prog.getConstantValue($IDENTIFIER.text);
								  if (val != null) {
								    if (val.signum() < 0 || val.compareTo(BigDecimal.ONE) == 1) {
								      logger.log("ERROR: on line " + $ran.getLine() + ", float value " +  val.toPlainString() + " must be in [0,1].",1);
								      System.exit(1);
								    }
								  }
								  else {
								      logger.log("ERROR: on line " + $ran.getLine() + ", unknown constant: " + $IDENTIFIER.text,1);
								  }
				  				  $str += val; 
				  				}
				  ) ')'								{ $str += ")"; }
		| 'random' '(' e=assign_expr[false] ',' f=assign_expr[false] ')'		{ $str += "random(" + $e.str + "," + $f.str + ")"; }
		) ';'										{ $str += ";\n"; }
		;

while_st returns [String str]:	
		'while' '(' assign_expr[true] ')' 'do' st_list 'od' 				{ $str = "while(" + $assign_expr.str + ") do\n" + $st_list.str + "od\n"; }
		;

for_st returns [String str] @init { Variable baseTabVar; }:	
		'for' '(' var=IDENTIFIER 'in'			{ PublicVariable v = new PublicVariable(1);
								  if (!prog.addPublicVariable($var.text,v))
								    logger.log("WARNING: on line " + $var.getLine() + ", variable " +  $var.text + " is already declared.",1);
								  $str = "for(" + $var.text + " in ";
								}
		( tab=IDENTIFIER indices 		 	{ baseTabVar = prog.getVariable($tab.text);
								  if (baseTabVar == null) {
								    logger.log("ERROR: on line " + $tab.getLine() + ", variable " +  $tab.text + " is not declared.",1);
								    System.exit(1);
								  }
								  $str +=  $tab.text;
								  for ( String i : $indices.l) {
								    $str += i;
								  }
								}
		| interval					{ $str += $interval.str; }
		)
		')' 'do' st_list 'od'				{ str += ") do\n" + $st_list.str + "od\n"; }
		;

if_st returns [String str]:	
		if_case 									{ $str = $if_case.str; }
		( else_cases 									{ $str += $else_cases.str; }
		)? 'fi'										{ $str += "fi\n"; }
		;

if_case returns [String str]:
		'if' '(' assign_expr[true] ')' 'then' st_list					{ $str = "if (" + $assign_expr.str + ") then\n" + $st_list.str; }
		;
      

else_cases returns [String str]:	
		elif_case 									{ $str = $elif_case.str; }
		( e=else_cases 									{ $str += $e.str; }
		)?
		| 'else' st_list								{ $str = "else\n" + $st_list.str; }
		;

elif_case returns [String str]:
		'elif' '(' assign_expr[true] ')' 'then' st_list 				{ $str = "elif (" + $assign_expr.str + ") then\n" + $st_list.str; }
		;

assign_expr [boolean guard] returns [String str]:
    a=and_expr[guard]			{ $str = $a.str; }
    ( '||' b=and_expr[guard]		{ try {
					    BigInteger val_str = new BigInteger(str);
					    BigInteger val_b = new BigInteger($b.str);
					    if (val_str.signum()!=0 || val_b.signum()!=0)
						$str = new String("1");
					    else
						$str = new String("0");
					  }
					  catch(NumberFormatException e) {
					    $str = new String("(" + str + ")||(" + $b.str + ")");
					  } 
					}
    )*
    ;
	   	   
and_expr [boolean guard] returns [String str]:	
    a=xor_expr[guard]			{ $str = $a.str; }
    ( '&&' b=xor_expr[guard]		{ try {
					      BigInteger val_str = new BigInteger(str);
					      BigInteger val_b = new BigInteger($b.str);
					      if (val_str.signum()!=0 && val_b.signum()!=0)
						  $str = new String("1");
					      else
						  $str = new String("0");
					  }
					  catch(NumberFormatException e) {
					      $str = new String("(" + str + ")&&(" + $b.str + ")");
					  }
					}
    )*
    ;

xor_expr [boolean guard] returns [String str]:	
    a=comp_expr[guard]			{ $str = $a.str; }
    ( '^' b=comp_expr[guard]		{ try {
					      BigInteger val_str = new BigInteger(str);
					      BigInteger val_b = new BigInteger($b.str);
					      if ((val_str.signum()==0 && val_b.signum()!=0) || (val_str.signum()!=0 && val_b.signum()==0))
						  $str = new String("1");					      
					      else
						  $str = new String("0");					     
					  }
					  catch(NumberFormatException e) {
					      $str = new String("(" + str + ")^(" + $b.str + ")");
					  }
				        }
    )*
    ;

        
comp_expr[boolean guard] returns [String str]:
    a=sum_expr[guard]			{ $str = $a.str; }
    ( '<=' b=sum_expr[false]		{ $str = new String("(" + $a.str + ")<=(" + $b.str + ")"); }
    | '>=' b=sum_expr[false]		{ $str = new String("(" + $a.str + ")>=(" + $b.str + ")"); }
    | '<'  b=sum_expr[false]		{ $str = new String("(" + $a.str + ")<(" + $b.str + ")"); }
    | '>'  b=sum_expr[false]		{ $str = new String("(" + $a.str + ")>(" + $b.str + ")"); }
    | '==' b=sum_expr[false]		{ $str = new String("(" + $a.str + ")==(" + $b.str + ")"); }
    | '!=' b=sum_expr[false]		{ $str = new String("(" + $a.str + ")!=(" + $b.str + ")"); }
    )?
    ;

sum_expr[boolean guard] returns [String str]:
    a=prod_expr[guard]					{   $str = $a.str; }
    ( sym='+' b=prod_expr[guard]			{   if(guard) {
								logger.log("ERROR: on line " + sym.getLine() + ", sums are not allowed in left members of a guard.",1);
								System.exit(1);
							    }
							    try {
								BigInteger val_str = new BigInteger(str);
								BigInteger val_b = new BigInteger($b.str);
								$str = val_str.add(val_b).toString();
							    }
							    catch(NumberFormatException e) {
								$str = new String("(" + str + ")+(" + $b.str + ")");
							    }
							}
     | sym='-' b=prod_expr[guard]			{   if(guard) {
								logger.log("ERROR: on line " + sym.getLine() + ", differences are not allowed in left members of a guard.",1);
								System.exit(1);
							    }
							    try {
								BigInteger val_str = new BigInteger(str);
								BigInteger val_b = new BigInteger($b.str);
								$str = val_str.subtract(val_b).toString();
							    }
							    catch(NumberFormatException e) {
								$str = new String("(" + str + ")-(" + $b.str + ")");
							    }
							}
     )*	 						
     ;

prod_expr[boolean guard] returns [String str]:
    a=unar_expr[guard]					{   $str = $a.str; }
    ( sym='*' b=unar_expr[guard]			{   if(guard) {
								logger.log("ERROR: on line " + sym.getLine() + ", products are not allowed in left members of a guard.",1);
								System.exit(1);
							    }
							    try {
								BigInteger val_str = new BigInteger(str);
								BigInteger val_b = new BigInteger($b.str);
								$str = val_str.multiply(val_b).toString();
							    }
							    catch(NumberFormatException e) {
								$str = new String("(" + str + ")*(" + $b.str + ")");
							    }
							}
    | sym='/' b=unar_expr[guard]			{   if(guard) {
								logger.log("ERROR: on line " + sym.getLine() + ", quotients are not allowed in left members of a guard.",1);
								System.exit(1);
							    }
							    try {
								BigInteger val_str = new BigInteger(str);
								BigInteger val_b = new BigInteger($b.str);
								$str = val_str.divide(val_b).toString();
							    }
							    catch(NumberFormatException e) {
								$str = new String("(" + str + ")/(" + $b.str + ")");
							    }
							}
    | sym='\%' b=unar_expr[guard]				{   if(guard) {
								logger.log("ERROR: on line " + sym.getLine() + ", modulos are not allowed in left members of a guard.",1);
								System.exit(1);
							    }
							    try {
								BigInteger val_str = new BigInteger(str);
								BigInteger val_b = new BigInteger($b.str);
								$str = val_str.mod(val_b).toString();
							    }
							    catch(NumberFormatException e) {
								$str = new String("(" + str + ")\%(" + $b.str + ")");
							    }
							}
    )*
    ;

unar_expr[boolean guard] returns [String str]:
    term[guard]				{   $str = $term.str; }
    | sym='!' a=term[guard]	    	{   if(guard) {
						logger.log("ERROR: on line " + sym.getLine() + ", negations are not allowed in left members of a guard.",1);
						System.exit(1);
					    }
					    try {
						BigInteger val = new BigInteger($term.str);
						if(val.signum()==0) {
						    $str = new String("1");
						}
						else {
						    $str = new String("0");
						} 
					    }
					    catch(NumberFormatException e) {
						$str = new String("!(" + $term.str + ")");
					    }
					}
    | sym='-' term[guard]		{   if(guard) {
						logger.log("ERROR: on line " + sym.getLine() + ", negations are not allowed in left members of a guard.",1);
						System.exit(1);
					    }
					    try {
						BigInteger val = new BigInteger($term.str);
						$str = val.negate().toString();
					    }
					    catch(NumberFormatException e) {
						$str = new String("-(" + $term.str + ")");
					    } 
					}
    ;

term [boolean guard] returns [String str]:
    sym='(' a=assign_expr[false] ')'	{   if(guard) {
						logger.log("ERROR: on line " + sym.getLine() + ", expressions are not allowed in left members of a guard.",1);
						System.exit(1);
					    }
					    $str = $assign_expr.str; 
					}
    | leaf_term[guard]			{   $str = $leaf_term.str; }
    ;

leaf_term [boolean guard] returns [String str]:
    NUMBER 				{   $str = $NUMBER.text; }
    | variable_term[guard]		{   $str = $variable_term.str; }
    ;

variable_term[boolean guard] returns [String str]:
    IDENTIFIER indices			{   BigDecimal val = prog.getConstantValue($IDENTIFIER.text);
					    if (val != null) {
						if ($indices.l.isEmpty()) {
						    $str = val.toBigInteger().toString();
						}
						else {
						    logger.log("ERROR: on line " + $IDENTIFIER.getLine()  + ", constant " + $IDENTIFIER.text + " cannot be an array.",1);
						    System.exit(1);
						}
					    }
					    else {
						Variable v = prog.getVariable($IDENTIFIER.text);
						if(v != null) {
						    if (!guard && v.getPublicity() < 0) {
							logger.log("ERROR: on line " + $IDENTIFIER.getLine() + ", private variables are not allowed in this expression.",1);
							System.exit(1);
						    }
						    $str = $IDENTIFIER.text;
						    for (Iterator<String> it = $indices.l.iterator(); it.hasNext();) {
							String s = it.next();
							$str = $str.concat(s);
						    }
						}
						else {
						    logger.log("ERROR: on line " + $IDENTIFIER.getLine() + ", variable " + $IDENTIFIER.text + " is not declared.",1);
						    System.exit(1);
						}
					    }
					}
					;

variable[boolean vprivate] returns [String str]:
    IDENTIFIER indices			{   Variable v = prog.getVariable($IDENTIFIER.text);
					    if(v != null) {
					      if (!vprivate && v.getPublicity() < 0) {
						  logger.log("ERROR: on line " + $IDENTIFIER.getLine() + ", private variables are not allowed in this expression.",1);
						  System.exit(1);
					      }
					      $str = $IDENTIFIER.text;
					      for (String i : $indices.l) {
						$str += i;
					      }
					    }
					    else {
					      logger.log("ERROR: on line " + $IDENTIFIER.getLine() + ", variable " + $IDENTIFIER.text + " is not declared.",1);
					      System.exit(1);
					    }
					}
    ;

indices returns [LinkedList<String> l] @init{ $l = new LinkedList<String>(); }:
    ('[' assign_expr[false] ']' 	{ $l.add(new String("[(" + $assign_expr.str + ")]")); }
    )*
    ;
    
float_number returns [BigDecimal f]:
    n1=NUMBER '.' n2=NUMBER 		{ $f = new BigDecimal($n1.text + "." + $n2.text); }
    ;
    
filename:	('.' | IDENTIFIER | NUMBER)+ ; 

    	 
INT :		'int'('a'..'z'|'A'..'Z'|'0'..'9'|'_')+ ;
IDENTIFIER  :   ('a'..'z'|'A'..'Z'|'_')('a'..'z'|'A'..'Z'|'0'..'9'|'_')* ;
NUMBER :	'0'..'9'+ ;
//FILENAME :      ('a'..'z'|'A'..'Z'|'0'..'9'|'_'|'.')+ ;
// FLOAT :		'0''.'('0'..'9')+ ;
NEWLINE :	'\n' { skip(); } ;
SPACES :	(' '|'\t'|'\r')+ { skip(); } ;
COMMENT : 	'//' ~('\n'|'\r')* '\r'? '\n' { $channel=HIDDEN; } ;
