/*  HyLeak is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2014-2017 - 

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: 
*/

grammar Preprocessor2;

@parser::header {
    package preprocessor;
    
    import java.math.BigInteger;
    import java.math.BigDecimal;
    import java.util.LinkedList;
    import support.Logger;
}

@lexer::header {
    package preprocessor; 
}

@members {
    Programme prog;
    boolean simulate;
    Logger logger = Logger.getInstance();
}

prog_grammar[Programme p] returns [String str] @init{ prog=p; simulate=false; }:
		(decl)* st_list				{ $str = $st_list.str; }
		EOF
		;

decl:	public_decl | private_decl;
	  
public_decl:
		public_type var_type[true] IDENTIFIER init_val ';'
		{   $var_type.v.setPublicity($public_type.b);
		    prog.addPublicVariable($IDENTIFIER.text,((PublicVariable)$var_type.v));
		    $var_type.v.setName($IDENTIFIER.text);
		    prog.addDeclaration($IDENTIFIER.text);
		    if (!$init_val.l.isEmpty()) {
			int r = ((PublicVariable)$var_type.v).updateInitialValues($init_val.l);
			if (r>0) {
			    logger.log(" on line " + $IDENTIFIER.getLine() + ".", 1);
			    if (r==2)
				System.exit(1);
			}
		    }
		}
	        ;

public_type returns [boolean b]:
		'public' 		{ $b = true; }
		| 'observable' 		{ $b = false; }
		;
	   
private_decl:	
		private_type var_type[false] IDENTIFIER init_intervals ';'
		{   $var_type.v.setPublicity($private_type.b);
		    prog.addPrivateVariable($IDENTIFIER.text,((PrivateVariable)$var_type.v));
		    $var_type.v.setName($IDENTIFIER.text);
		    prog.addDeclaration($IDENTIFIER.text);
		    if (!$init_intervals.l.isEmpty()) {
			int r = ((PrivateVariable)$var_type.v).updateIntRange($init_intervals.l);
			if (r>0) {
			    logger.log(" on line " + $IDENTIFIER.getLine() + ".", 1);
			    if (r==2)
				System.exit(1);
			}
		    }
		}
		;
	      
private_type returns [boolean b]:
	    'private'		{ $b = true; }
	    | 	'secret' 	{ $b = false; }
	    ;
	    
var_type[boolean pub] returns [Variable v] :
	INT
	{   String sizeString = $INT.text.substring(3);
	    int sizeInt = 0;
	    try {
		sizeInt = Integer.parseInt(sizeString);
	    }
	    catch (NumberFormatException e) {
		logger.log("ERROR: on line " + $INT.getLine() + ", value of " + $INT.text + " cannot be converted to integer.",1);
		System.exit(1);
	    }    
	    if (pub) {
		$v = new PublicVariable(sizeInt);
	    }
	    else { 
		$v = new PrivateVariable(sizeInt);
	    }
	}
	| a='array' '[' e1=NUMBER '..' e2=NUMBER ']' 'of' w=var_type[$pub]
	{   String smin = $e1.text;
	    String smax = smax = $e2.text;
	    $v = $w.v;
	    try {
		int min = Integer.parseInt(smin);
		int max = Integer.parseInt(smax);
		$v.addDimensionFirst(new ArrayDimension(min,max));
	    }
	    catch(NumberFormatException e) {
		logger.log("ERROR: on line " + a.getLine() + ", array size " + smin + " or " + smax + " cannot be converted to integer.",1);
		System.exit(1);
	    }
	}
	;
   
init_val returns [LinkedList<BigInteger> l] @init{ l = new LinkedList<BigInteger>(); } :
								// empty
	| a=':='( val=NUMBER					{   try {
									BigInteger int_val = new BigInteger($val.text);
									$l.add(int_val);
								    }
								    catch(NumberFormatException e) {
									logger.log("ERROR: on line " + a.getLine() + ", " + $val.text + " cannot be converted to integer.",1);
									System.exit(1);
								    }
								}
		| '{' number_list '}'				{   $l = $number_list.l; }
		)
	;

number_list returns [LinkedList<BigInteger> l]	@init{ l = new LinkedList<BigInteger>(); } :
	val1=NUMBER 						{   try {
									BigInteger int_val = new BigInteger($val1.text);
									$l.add(int_val);
								    }
								    catch(NumberFormatException e) {
									logger.log("ERROR: on line " + a.getLine() + ", " + $val1.text + " cannot be converted to integer.",1);
									System.exit(1);
								    }
								}
	(a=',' val2=NUMBER					{   try {
									BigInteger int_val = new BigInteger($val2.text);
									$l.add(int_val);
								    }
								    catch(NumberFormatException e) {
									logger.log("ERROR: on line " + a.getLine() + ", " + $val2.text + " cannot be converted to integer.",1);
									System.exit(1);
								    }
								}
	)*
	;

init_intervals returns [LinkedList<BigInteger> l] @init{ l = new LinkedList<BigInteger>(); } :
	// empty
	| a=':=' (interval		  			{ if ($l.isEmpty()) {
								      $l.addAll($interval.l);
								  }
								  else {
								      int comp = $l.getLast().compareTo($interval.l.getFirst().subtract(BigInteger.ONE));
								      if (comp < 0) {
									  $l.add($interval.l.getFirst());
									  $l.add($interval.l.getLast());
								      }
								      else {
									  comp = $l.getLast().compareTo($interval.l.getFirst());
									  if (comp <= 0) {
									      $l.removeLast();
									      $l.add($interval.l.getLast());
									  }
									  else {
									      logger.log("ERROR: on line " + $a.getLine() + ", overlapping intervals.",1);
									      System.exit(1);
									  }
								      }	    
								  }
								}
		 )+
	;
	
interval returns [LinkedList<BigInteger> l]:
	a='[' min=NUMBER ',' max=NUMBER ']'
	{  try {
		BigInteger minVal = new BigInteger($min.text);
		BigInteger maxVal = new BigInteger($max.text);
		$l = new LinkedList<BigInteger>();
		if (minVal.compareTo(maxVal) > 0) 
		    logger.log("WARNING: on line " + $a.getLine() + ", interval [" + minVal.toString() + "," + maxVal.toString() + "] is empty.",1);
		else {
		    $l.add(minVal);
    		    $l.add(maxVal);
		}
            }
	    catch(NumberFormatException e) {
		logger.log("ERROR: on line " + $a.getLine() + ", " + $min.text + " or " + $max.text + " cannot be converted to integer.",1);
		System.exit(1);
	    }
	}
	;
	
st_list returns [String str] @init { str = ""; } :
				( st 					{ $str += $st.str; }
				)*  				
				;
	
st returns [String str]:	assignement_st				{ $str = $assignement_st.str; }
				| print_st 				{ $str = $print_st.str; }
				| return_st				{ $str = $return_st.str; }
				| simulate_st				{ $str = $simulate_st.str; }
				| random_st				{ $str = $random_st.str; }
				| while_st				{ $str = $while_st.str; }
				| forarray_st				{ $str = $forarray_st.str; }
				| forloop_st				{ $str = $forloop_st.str; }
				| if_st					{ $str = $if_st.str; }
				;
  
assignement_st returns [String str]:	'assign' var=variable ':=' e=assign_expr ';'	 	{ $str = "assign " + $var.text + " := " + $e.text + ";\n"; };

print_st returns [String str]:		'print' ';' 						{ $str = "print;\n"; };

return_st returns [String str]:		
		'return' ';'									{ $str = "return;\n"; }
		| 'return' 'after' 'simulate' 
		  var1=variable ':=' filename '(' var2=variable ')' ';'				{ $str = "return after simulate " + $var1.text + " := " + $filename.text + " (" + $var2.text + ");\n"; }
		;
		
simulate_st returns [String str] @init{ simulate = true; } :
		'simulate' ';' 					{ $str = "simulate;\n"; }
		| 'simulate-abs' ';' 				{ $str = "simulate-abs;\n"; }
		;

random_st returns [String str]:	
		ran='random' var=variable ':='							{ $str = "random " + $var.text + " := "; }
		( 'randombit' '(' float_number ')'						{ $str += "randombit( " + $float_number.f + ")"; }
		| 'random' '(' e=assign_expr ',' f=assign_expr ')'				{ $str += "random( " + $e.text + "," + $f.text + ")"; }
		) ';'										{ $str += ";\n"; }
		;

while_st returns [String str]:	
		'while' '(' assign_expr ')' 'do' st_list 'od' 					{ $str = "while( " + $assign_expr.text + ") do\n" + $st_list.str + "od\n"; }
		;

forarray_st returns [String str] @init { Variable baseTabVar; }:	
		'for' '(' var=IDENTIFIER 'in' tab=IDENTIFIER indices ')'
										{ baseTabVar = prog.getVariable($tab.text);
										  ArrayDimension tabVar = baseTabVar.getArrayDimension($indices.l.size());
										  if (tabVar == null) {
										      logger.log("ERROR: on line " + $tab.getLine() + ", variable " +  $tab.text + " has not enough dimensions.",1);
										      System.exit(1);
										  }
										  PublicVariable v = new PublicVariable(baseTabVar.getIntType());
										  prog.addPublicVariable($var.text,v);
										  for (int i = 1+$indices.l.size(); i < baseTabVar.getDimensions(); i++) {
										      v.addDimensionFirst(baseTabVar.getArrayDimension(i));
										  }
										}
		'do' st_list 'od'				
										{ $str = "";
										  for (int i = tabVar.getMinIndex(); i <= tabVar.getMaxIndex(); i++) {
										      String newTabIndex = new String($tab.text);
										      for (String s : $indices.l) {
											  newTabIndex += s;
										      }
										      newTabIndex += "["+i+"]";
										      $str += $st_list.str.replace($var.text,newTabIndex);
										  }
										  prog.removeVariable($var.text);
										}
		;
			
forloop_st returns [String str] :	
		'for' '(' var=IDENTIFIER 'in' interval ')'
										{ int size = (int) Math.ceil(Math.log($interval.l.getLast().intValue() + 1)/Math.log(2));
										  PublicVariable v = new PublicVariable(size);
										  prog.addPublicVariable($var.text,v);
										}
		'do' st_list 'od'
										{ $str = "";
										  for (int i = $interval.l.getFirst().intValue(); i<= $interval.l.getLast().intValue(); i++) {
										      String newstr = $st_list.str.replace(" " + $var.text, " "+Integer.toString(i));
										      newstr = newstr.replace("(" + $var.text, "("+Integer.toString(i));
										      $str += newstr;
										  }
										  prog.removeVariable($var.text);
										}
		;	


if_st returns [String str]:	
		if_case 									{ $str = $if_case.str; }
		( else_cases 									{ $str += $else_cases.str; }
		)? 'fi'										{ $str += "fi\n"; }
		;

if_case returns [String str]:
		'if' '(' assign_expr ')' 'then' st_list						{ $str = "if ( " + $assign_expr.text + ") then\n" + $st_list.str; }
		;
      

else_cases returns [String str]:	
		elif_case 									{ $str = $elif_case.str; }
		( e=else_cases 									{ $str += $e.str; }
		)?
		| 'else' st_list								{ $str = "else\n" + $st_list.str; }
		;

elif_case returns [String str]:
		'elif' '(' assign_expr ')' 'then' st_list 					{ $str = "elif ( " + $assign_expr.text + ") then\n" + $st_list.str; }
		;

assign_expr:	a=and_expr ( '||' b=and_expr )* ;
	   	   
and_expr:	a=xor_expr ( '&&' b=xor_expr )* ;

xor_expr:	a=comp_expr ( '^' b=comp_expr )* ;

        
comp_expr:	a=sum_expr
		( '<=' b=sum_expr
		| '>=' b=sum_expr
		| '<'  b=sum_expr
		| '>'  b=sum_expr
		| '==' b=sum_expr
		| '!=' b=sum_expr
		)?
		;

sum_expr:	a=prod_expr ( ( '+' | '-') b=prod_expr )* ;

prod_expr:	a=unar_expr ( ('*' | '/' | '\%') b=unar_expr )* ;

unar_expr:	('!' | '-')? term ;

term: 		'(' assign_expr ')' | leaf_term ;

leaf_term:	NUMBER | variable ;
		
variable:  IDENTIFIER indices;

indices returns [LinkedList<String> l] @init{ $l = new LinkedList<String>(); }:
    ('[' assign_expr ']' 		{ $l.add("[" + $assign_expr.text + "]"); }
    )*
    ;
    
float_number returns [BigDecimal f]:
    n1=NUMBER '.' n2=NUMBER 		{ $f = new BigDecimal($n1.text + "." + $n2.text); }
    ;
    
filename:	('.' | IDENTIFIER | NUMBER)+ ; 

    	 
INT :		'int'('a'..'z'|'A'..'Z'|'0'..'9'|'_')+ ;
IDENTIFIER  :   ('a'..'z'|'A'..'Z'|'_')('a'..'z'|'A'..'Z'|'0'..'9'|'_')* ;
NUMBER :	'0'..'9'+ ;
//FILENAME :      ('a'..'z'|'A'..'Z'|'0'..'9'|'_'|'.')+ ;
// FLOAT :		'0''.'('0'..'9')+ ;
NEWLINE :	'\n' { skip(); } ;
SPACES :	(' '|'\t'|'\r')+ { skip(); } ;
COMMENT : 	'//' ~('\n'|'\r')* '\r'? '\n' { $channel=HIDDEN; } ;
