/*  HyLeak is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2014-2017 - 

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: 
*/

grammar Preprocessor4;

@parser::header {
    package preprocessor;
    
    import java.math.BigInteger;
    import java.math.BigDecimal;
    import java.util.LinkedList;
}

@lexer::header {
    package preprocessor; 
}


@members {
    Programme prog;
}

prog_grammar[Programme p]: { prog=p; }  decl* st_list[null] EOF  { prog.addStatements($st_list.l); } ;

decl:	public_decl 
    |	private_decl
    ;

public_decl:	public_type INT IDENTIFIER num_indices init_val ';'
		{   int sizeInt = Integer.parseInt($INT.text.substring(3)); 
		    PublicVariable v = new PublicVariable(sizeInt);
		    v.setPublicity($public_type.b);
		    v.setName($IDENTIFIER.text + $num_indices.text);
		    prog.addPublicVariable(v.getName(),v);
		    prog.addDeclaration(v.getName());
		    if (!$init_val.l.isEmpty())
			v.updateInitialValues($init_val.l);		    
		}
	        ;

public_type returns [boolean b]:
		'public' 		{ $b = true; }
		| 'observable' 		{ $b = false; }
		;
	   
private_decl:	private_type INT IDENTIFIER num_indices init_intervals ';'
		{   int sizeInt = Integer.parseInt($INT.text.substring(3));
		    PrivateVariable v = new PrivateVariable(sizeInt);
	    	    v.setPublicity($private_type.b);
		    v.setName($IDENTIFIER.text + $num_indices.text);
	    	    prog.addPrivateVariable(v.getName(),v);
		    prog.addDeclaration(v.getName());
		    if (!$init_intervals.l.isEmpty())
			v.updateIntRange($init_intervals.l);
		}
		;
	      
private_type returns [boolean b]:
	    'private'		{ $b = true; }
	    | 	'secret' 	{ $b = false; }
	    ;
	    
num_indices returns [LinkedList<Integer> l] @init{ $l = new LinkedList<Integer>(); }:
    ('[' NUMBER ']' 		{ $l.add(new Integer(Integer.parseInt($NUMBER.text))); }
    )*
    ;
	    
init_val returns [LinkedList<BigInteger> l] @init{ l = new LinkedList<BigInteger>(); } :
								// empty
	| a=':=' val=NUMBER					{   BigInteger int_val = new BigInteger($val.text);
								    $l.add(int_val);
								}
	;

init_intervals returns [LinkedList<BigInteger> l] @init{ l = new LinkedList<BigInteger>(); } :
	// empty
	| a=':=' (interval		  			{ $l.addAll($interval.l); }
		 )+
	;
	
interval returns [LinkedList<BigInteger> l]:
	a='[' min=NUMBER ',' max=NUMBER ']'			{   $l = new LinkedList<BigInteger>();
								    $l.add(new BigInteger($min.text));
								    $l.add(new BigInteger($max.text));
								}
	;

st_list [Statement parent] returns [LinkedList<Statement> l]
@init { 
  $l = new LinkedList<Statement>(); 
} :
( st[parent] 			{ $l.addAll($st.l); } 
)*
;
	
st [Statement parent] returns [LinkedList<Statement> l]:
    assignement_st[parent]	{ $l = new LinkedList<Statement>(); $l.add($assignement_st.s); }
    | simulate_st               { $l = new LinkedList<Statement>(); $l.add($simulate_st.s); }
    | print_st               	{ $l = new LinkedList<Statement>(); $l.add($print_st.s); }
    | return_st[parent]		{ $l = new LinkedList<Statement>(); $l.add($return_st.s); }
    | random_st[parent]		{ $l = new LinkedList<Statement>(); $l.add($random_st.s); }
    | while_st[parent] 		{ $l = new LinkedList<Statement>(); $l.add($while_st.ifst); }
    | if_st[parent] 		{ $l = new LinkedList<Statement>(); $l.add($if_st.ifst); }
    ;
  
assignement_st [Statement parent] returns [AssignStatement s]
@init {
  $s = new AssignStatement();
  if (parent == null)
    parent = $s;
} :
'assign' var=variable[parent] ':=' e=assign_expr[parent] ';' 
{
    $s.setVariable(new String("(" + $var.text + ")"));
    $s.setExpression(new String("(" + $e.text + ")"));
}
;

simulate_st returns [SimulateStatement s]:
( 'simulate'    		{ $s = new SimulateStatement(false); }
| 'simulate-abs'        	{ $s = new SimulateStatement(true); }
)
';'
;

print_st returns [PrintStatement s]:
'print' ';'          { $s = new PrintStatement(); }
;

return_st [Statement parent] returns [ReturnStatement s]
@init {
  $s = new ReturnStatementSimulate();
  if (parent == null)
    parent = $s;
} :
'return' ';' 			{ $s = new ReturnStatement(); }
| 'return' 'after' 'simulate' var1=variable[parent] ':=' filename '(' var2=variable[parent] ')' ';'
{ 
    ReturnStatementSimulate ss = (ReturnStatementSimulate)$s;
    ss.setObsVariable(new String("(" + $var1.text + ")")); 
    ss.setFilename($filename.text); 
    ss.setPrivVariable(new String("(" + $var2.text + ")"));
}
;

random_st [Statement parent] returns [RandomStatement s]
@init {
  $s = new RandomStatement();
  if (parent == null)
    parent = $s;
} :
ran='random' var=variable[parent] ':=' 			{ $s.setVariable(new String("(" + $var.text + ")")); }
( 'randombit' '(' float_number ')'      		{ $s.setProbability($float_number.f); } 
| 'random' '(' e=assign_expr[parent] ',' f=assign_expr[parent] ')'
							{ $s.setExpression(new String("(" + $e.text + "),(" + $f.text +')')); }
)
';'
;

while_st [Statement parent] returns [IfStatement ifst]
@init {
  $ifst = new IfStatement();
  if (parent == null)
    parent = $ifst;
} :
'while' '(' assign_expr[parent] ')' 'do' st_list[parent] 'od' 
{
    $ifst.setGuard($assign_expr.text);
    $ifst.setThenGoto(new GotoStatement(2));
    for (Statement st : $st_list.l) {
	$ifst.addThenStatement(st);
    }
    $ifst.setElseGoto(new GotoStatement($ifst.getNbThenStatements()+2));
    $ifst.setExitGoto(new GotoStatement((-1)*$ifst.getNbThenStatements()-3));
}
;

if_st [Statement parent] returns [IfStatement ifst]
@init {
    ifst = new IfStatement();
    if (parent == null)
        parent = ifst;
} :
if_case[parent,ifst] (else_cases[parent,ifst])? 'fi'
{
    $ifst.setExitGoto(new GotoStatement($ifst.getNbElseStatements()+1));
} 
;

if_case [Statement parent, IfStatement ifst] :
    'if' '(' assign_expr[parent] ')' 'then' st_list[parent]
    {
    	ifst.setGuard($assign_expr.text);
	$ifst.setThenGoto(new GotoStatement(2));
	for (Statement st : $st_list.l) {
	    $ifst.addThenStatement(st);
	}
	$ifst.setElseGoto(new GotoStatement($ifst.getNbThenStatements()+2));
    }
    ;

else_cases [Statement parent, IfStatement ifst] :
    elif_case[parent] (e=else_cases[parent,$elif_case.ifst])?
    {
	$ifst.addElseStatement($elif_case.ifst);
	$elif_case.ifst.setExitGoto(new GotoStatement($elif_case.ifst.getNbElseStatements()+1));
    }
    | 'else' st_list[parent]
    {
    	for (Statement st : $st_list.l) {
	    $ifst.addElseStatement(st);
	}
    }
    ;

elif_case [Statement parent] returns [IfStatement ifst]
@init {
    ifst = new IfStatement();
    // parent is never null
} :
'elif' '(' assign_expr[parent] ')' 'then' st_list[parent]
{
    $ifst.setGuard($assign_expr.text);
    $ifst.setThenGoto(new GotoStatement(2));
    for (Statement st : $st_list.l) {
	$ifst.addThenStatement(st);
    }
    $ifst.setElseGoto(new GotoStatement($ifst.getNbThenStatements()+2));
}
;

assign_expr [Statement st]:
    and_expr[st] ( '||' assign_expr[st] )? ;
	   	   
and_expr [Statement st]:
    xor_expr[st] ( '&&' and_expr[st] )? ;

xor_expr [Statement st]:	
    comp_expr[st] ( '^' xor_expr[st] )? ;

        
comp_expr [Statement st]:
sum_expr[st]
( '<=' sum_expr[st]
| '>=' sum_expr[st]
| '<'  sum_expr[st]
| '>'  sum_expr[st]
| '==' sum_expr[st]
| '!=' sum_expr[st]
)?
;

sum_expr [Statement st]:
    prod_expr[st] ( '+' sum_expr[st] | '-' sum_expr[st] )? ;

prod_expr [Statement st]:
    unar_expr[st] ( '*' prod_expr[st] | '/' prod_expr[st] | '\%' prod_expr[st] )? ;

unar_expr [Statement st]:
    term[st] | '!' term[st] | '-' term[st] ;

term [Statement st]:
    '(' assign_expr[st] ')' | leaf_term[st] ;

leaf_term [Statement st]: NUMBER | variable[st] ;	 

variable [Statement st]:
    IDENTIFIER indices[st]		{   Variable v = prog.getVariable($IDENTIFIER.text);
					    if(v != null) {
						if (v.getPublicity() == 1 && $st != null) {
						    v.setLastUsed($st);
						    $st.addUsedVariable(v);
						}
					    }
					}
    ;

indices [Statement st]:
    ('[' assign_expr[st] ']' )*
    ;
    
    
float_number returns [BigDecimal f]:
n1=NUMBER '.' n2=NUMBER 	{ $f = new BigDecimal($n1.text + "." + $n2.text); }
;

filename:
('.' | IDENTIFIER | NUMBER)+
; 

    	 
INT :		'int'('a'..'z'|'A'..'Z'|'0'..'9'|'_')+ ;
IDENTIFIER  :   ('a'..'z'|'A'..'Z'|'_')('a'..'z'|'A'..'Z'|'0'..'9'|'_')* ;
NUMBER :	'0'..'9'+ ;
//FILENAME :      ('a'..'z'|'A'..'Z'|'0'..'9'|'_'|'.')+ ;
// FLOAT :		'0''.'('0'..'9')+ ;
NEWLINE :	'\n' { skip(); } ;
SPACES :	(' '|'\t'|'\r')+ { skip(); } ;
COMMENT : 	'//' ~('\n'|'\r')* '\r'? '\n' { $channel=HIDDEN; } ;
