/*  HyLeak is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2014-2017 - 

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: 
*/

package preprocessor;

import java.io.IOException;
import java.io.PrintWriter;
import java.math.BigDecimal;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.ListIterator;

public class Programme {
    private HashMap<String,BigDecimal> cmdlineConstants;
    private HashMap<String,BigDecimal> vconst;
    private HashMap<String,PublicVariable> vpublic;
    private HashMap<String,PrivateVariable> vprivate;
    
    /** Used to mainain an ordered list of the declarations. */
    private LinkedList<String> declarations;
    private LinkedList<Statement> statements;
    
    public Programme(HashMap<String, BigDecimal> constants) {
	cmdlineConstants = constants;
        vconst = new HashMap<String, BigDecimal>();
        vpublic = new HashMap<String, PublicVariable>();
        vprivate = new HashMap<String, PrivateVariable>();
        statements = new LinkedList<Statement>();
        declarations = new LinkedList<String>();
    }
    
    public boolean addConstant(String name, BigDecimal val) {
        if (this.hasVariable(name)) {
            return false;
        }
        else {
	    if (cmdlineConstants != null && cmdlineConstants.containsKey(name))
		vconst.put(name, cmdlineConstants.get(name));
	    else
		vconst.put(name, val);
            return true;
        }
    }
    
    public boolean addPublicVariable(String name, PublicVariable val) {
        if (this.hasVariable(name)) {
            return false;
        }
        else {
            vpublic.put(name, val);
            return true;
        }
    }
    
    public boolean addPrivateVariable(String name, PrivateVariable val) {
        if (this.hasVariable(name)) {
            return false;
        }
        else {
            vprivate.put(name, val);
            return true;
        }
    }
    
    public void addStatements(LinkedList<Statement> l) {
        statements.addAll(l);
    }
    
    public void addDeclaration(String s) {
        declarations.add(s);
    }
    
    public boolean hasVariable(String name) {
        return hasConstant(name) || hasPublicVariable(name) || hasPrivateVariable(name);
    }
    
        
    public boolean hasConstant(String name) {
        return vconst.containsKey(name);
    }
            
    public boolean hasPublicVariable(String name) {
        return vpublic.containsKey(name);
    }
    
    public boolean hasPrivateVariable(String name) {
        return vprivate.containsKey(name);
    }
    
    public BigDecimal getConstantValue(String name) {
        return vconst.get(name);
    }
    
    public PublicVariable getPublicVariable(String name) {
        return vpublic.get(name);
    }
    
    public PrivateVariable getPrivateVariable(String name) {
        return vprivate.get(name);
    }
        
    public Variable getVariable(String name) {
        Variable var = vpublic.get(name);
        if (var == null ) {
            var = vprivate.get(name);
        }
        return var;
    }
    
    public Variable removeVariable(String name) {
        Variable var = vpublic.remove(name);
        if (var == null ) {
            var = vprivate.remove(name);
        }
        return var;
    }
    
    public void freeVariables() {
        for (ListIterator<Statement> it = statements.listIterator(0); it.hasNext();) {
            Statement statement = it.next();
            for(Variable v : statement.getUsedVariables()) {
            	if (v.getLastUsed() == statement && v.name != null) {
            		LinkedList<FreeStatement> freed = v.freeVariable();
            		for (FreeStatement f : freed) {
            			it.add(f);
            		}
            	}
            }
        }
    }
    
    public void checkConstants() throws Exception {
	// check that all the constants given from the command line have been declared in the model
	if (cmdlineConstants != null) {
	    for (String c : cmdlineConstants.keySet()) {
		if (!vconst.containsKey(c)) {
		    throw new Exception("ERROR: constant " + c + " is not declared in the program.");
		}
	    }
	}
    }
    
    public void print(PrintWriter out) throws IOException {
        int line = 1;
        for (String varname : declarations) {
            Variable var = getVariable(varname);
            line += var.print(out);
        }        
        for (Statement statement : statements) {
            out.print(statement.toString(line));
            line += statement.getNbStatements();
        }
    }
}
