/*  HyLeak is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2014-2017 - 

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: 
*/
package preprocessor;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.HashMap;
import java.util.Map.Entry;
import java.util.Set;
import java.util.HashSet;

public class Ranges {
    
    /** Range of the variables. */
    protected Map<Variable, BigIntegerInterval > variables_range;
    
    protected BigInteger totalObservable;
    
    protected BigInteger totalInternal;
    
    public Ranges() {
	variables_range = new HashMap<Variable, BigIntegerInterval>();
	totalObservable = BigInteger.ONE;
	totalInternal = BigInteger.ONE;
    }
    
    public Ranges(Ranges r) {
	variables_range = new HashMap<Variable, BigIntegerInterval>(r.variables_range);
	totalObservable = r.totalObservable;
	totalInternal = r.totalInternal;
    }
    
    private void recomputeTotalRanges() {
    	totalObservable = BigInteger.ONE;
	totalInternal = BigInteger.ONE;
	for (Map.Entry<Variable, BigIntegerInterval> entry : variables_range.entrySet()) {
	    if (entry.getKey().getPublicity() == 2) // observable
		totalObservable = totalObservable.multiply(entry.getValue().getSize());
	    else if (entry.getKey().getPublicity() == 1) // internal
		totalInternal = totalInternal.multiply(entry.getValue().getSize());
	}
    }
    
    /** Compute a union of the ranges in which appears only the variables that are in both ranges. */
    // not used
    public Ranges union(Ranges r) {
	Ranges ret = new Ranges();
	for (Map.Entry<Variable, BigIntegerInterval> entry : variables_range.entrySet()) {
	    BigIntegerInterval ithis = entry.getValue();
	    BigIntegerInterval ir = r.getVariableRange(entry.getKey());
	    if (ir != null) {
		ir = new BigIntegerInterval(ir.getMin().min(ithis.getMin()),ir.getMax().max(ithis.getMax()));
		ret.addVariableRange(entry.getKey(),ir);
	    }
	}
	recomputeTotalRanges();
	return ret;
    }
    
    /** Compute a union of the ranges in which appears any variable that is in one range. */
    public Ranges union_large(Ranges r) {
	Ranges ret = new Ranges(r);
	for (Map.Entry<Variable, BigIntegerInterval> entry : variables_range.entrySet()) {
	    BigIntegerInterval ithis = entry.getValue();
	    BigIntegerInterval ir = r.getVariableRange(entry.getKey());
	    if (ir != null) {
		ir = new BigIntegerInterval(ir.getMin().min(ithis.getMin()),ir.getMax().max(ithis.getMax()));
		ret.addVariableRange(entry.getKey(),ir);
	    }
	    else {
   		ret.addVariableRange(entry.getKey(),ithis);
	    }
	}
	recomputeTotalRanges();
	return ret;
    }
    
    public void addVariableRange(Variable v, BigIntegerInterval r) {
	BigIntegerInterval ret = variables_range.put(v,r);
	if (ret == null) {
	    if (v.getPublicity() == 2) // observable
		totalObservable = totalObservable.multiply(r.getSize());
	    else if (v.getPublicity() == 1) // internal
		totalInternal = totalInternal.multiply(r.getSize());
	}
    }
    
    public void modifyVariableRange(Variable v, BigIntegerInterval r) {
	BigIntegerInterval ret = variables_range.put(v, new BigIntegerInterval(r.getMin(), r.getMax().min(v.getMaxInt())));
	if (ret != null) { //variable was already initialised with a range
	  if (v.getPublicity() == 2) { // observable
	      totalObservable = totalObservable.divide(ret.getSize());
	      totalObservable = totalObservable.multiply(r.getSize());
	  } else if (v.getPublicity() == 1) { // internal
	      totalInternal = totalInternal.divide(ret.getSize());
	      totalInternal = totalInternal.multiply(r.getSize());
	  }
	}
	else {
	  if (v.getPublicity() == 2) // observable
	      totalObservable = totalObservable.multiply(r.getSize());
	  else if (v.getPublicity() == 1) // internal
	      totalInternal = totalInternal.multiply(r.getSize());
	}
    }
    
    public void updateVariableRange(Ranges global_ranges) {
	// update the ranges of non initialized variables
	for(Map.Entry<Variable,BigIntegerInterval> entry : global_ranges.getVariablesRange().entrySet()) {
	    if (getVariableRange(entry.getKey()) == null)
		addVariableRange(entry.getKey(),entry.getValue());
	}
	recomputeTotalRanges();
    }
    
    public BigIntegerInterval getVariableRange(Variable v) {
	return variables_range.get(v);
    }
    
    public Map<Variable, BigIntegerInterval> getVariablesRange() {
	return variables_range;
    }
    
    public BigInteger getTotalObservable() {   
	return totalObservable;
    }
    
    public BigInteger getTotalInternal() {   
	return totalInternal;
    }
    
    public int getNbVariables() {
	return variables_range.size();
    }
    
    public boolean simulateOk() {
	return totalObservable.compareTo(totalInternal) < 0;
    }
    
   
    
    public String toString() {
	String ret = "";
	for (Map.Entry<Variable, BigIntegerInterval> entry : variables_range.entrySet()) {
	    ret += entry.getKey().getName();
	    ret += ": [" + entry.getValue().getMin() + "," + entry.getValue().getMax() + "]; ";
	}
	ret += "TOTAL_OBS = " + totalObservable;
	ret += " TOTAL_INT = " + totalInternal;
	return ret;
    }
    
    
    
  
}
