package support;

//import java.math.BigInteger;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.TreeSet;

import hyleak.JointPosteriorDistributionManager;
import hyleak.Signature;
import distributions.SingleDistribution;
import distributions.JointDistribution;
import distributions.JointFrequencyDistribution;
import bham.leakiest.ARFFFile;
import bham.leakiest.Observations;
import bham.leakiest.ReadFile;
import bham.leakiest.State;
import bham.leakiest.ProbDist;
import bham.leakiest.Channel;

/**
*
* @author  
*/
public class ConvertLeakiEstObjects {
    private static final Logger logger = Logger.getInstance();

    /////////////////////////////////////////////////////////////
	// Methods for constructing prior distributions
	/**
	 * Computes the single distribution from the given probability
	 * distribution of the tool leakiEst's format.
	 * @param pd probability distribution of leakiEst's format
	 * @return the single distribution
	 */
    public static SingleDistribution computeSingleDistribution(ProbDist pd) {
    	if(pd == null) {
            log("ERROR: The given probability distribution (in leakiEst's format) is null.",1);
    		return null;
    	}
    	// Get information on the sample
    	int sizeOfDomain = pd.sizeSampleSpace();
    	String[] labels = new String[sizeOfDomain];
    	State[] sts = pd.getStatesArray();
    	// Compute the single distribution map
    	Map<String,BigDecimal> probMap = new LinkedHashMap<String,BigDecimal>();
    	for(int i = 0; i < sizeOfDomain; i++) {
    		labels[i] = sts[i].stringState();
    		BigDecimal prob = BigDecimal.valueOf(pd.getProb(sts[i]));
    		probMap.put(labels[i], prob);
    	}
    	// Construct the single distribution
    	SingleDistribution dist = new SingleDistribution(labels, probMap);
    	return dist;
    }
    
    /**
	 * Computes the joint distribution from a given prior file.
     * 
     * @param fileName name of a given prior file (lieakiEst format)
	 * @return the prior distribution of leakiEst's format
     */
    public static ProbDist computeProbDistFromPriorFile(String fileName) {
		// Read the prior (input) distribution array using the ReadFile Class
		ReadFile priorFileReader = new ReadFile(fileName, 0, "discrete prior");
		priorFileReader.readDistribution();
    	ProbDist pd = priorFileReader.getDistribution();
    	return pd;
    }
	
    /**
	 * Computes the joint distribution from a given prior file.
     * 
     * @param fileName name of a given prior file (lieakiEst format)
	 * @return the prior distribution
     */
    public static SingleDistribution computeSingleDistributionFromPriorFile(String fileName) {
    	ProbDist pd = computeProbDistFromPriorFile(fileName);
    	SingleDistribution dist = computeSingleDistribution(pd);
    	return dist;
    }
	
    
    /////////////////////////////////////////////////////////////
	// Methods for constructing joint distributions
    /**
     * Computes the (posterior) joint distribution from the tool
     * leakiEst's Channel object.
     * 
     * @return the (posterior) joint distribution
     */
    public static JointDistribution computeDistribution(Channel channel) {
    	if(channel == null) {
            log("ERROR: The given observations (in leakiEst's format) is null.",1);
            return null;
    	}
    	// Get information on the sample
		String[] secretLabels = channel.getInputNames();
		String[] observableLabels = channel.getOutputNames();
		double[][] matrix = channel.getMatrix();
    	
		Map<String,Map<String,BigDecimal>> jointProbMap = new LinkedHashMap<String,Map<String,BigDecimal>>();
		for(int row = 0; row < secretLabels.length; row++) {
			String secret = secretLabels[row];
			// Setup a map from observables to probabilities
			Map<String,BigDecimal> observableProbMap;
			if(!jointProbMap.containsKey(secret)) {
	    	    observableProbMap = new LinkedHashMap<String,BigDecimal>();
	    	    jointProbMap.put(secret, observableProbMap);
			} else {
				observableProbMap = jointProbMap.get(secret);
			}
			// Calculate the joint probability
			for(int col = 0; col < observableLabels.length; col++) {
				String observable = observableLabels[col];
				BigDecimal probBD = BigDecimal.valueOf(matrix[row][col]);
				observableProbMap.put(observable, probBD);
			}
		}
    	
    	// Construct the joint distribution
    	JointDistribution dist = new JointDistribution(secretLabels, observableLabels, jointProbMap);
    	return dist;
    }

    
    /**
     * Computes the (posterior) joint distribution from the tool
     * leakiEst's Observations object.
     * 
     * @return the (posterior) joint distribution
     */
    public static JointFrequencyDistribution computeJointFrequencyDistribution(Observations obs) {
    	if(obs == null) {
            log("ERROR: The given observations (in leakiEst's format) is null.",1);
            return null;
    	}
    	// Get information on the sample
		String[] secretLabels = obs.getInputNames();
		String[] observableLabels = obs.getOutputNames();
		LinkedHashMap<String, LinkedHashMap<String, Integer>> sampleMatrixMap = obs.getObservationsMatrixMap();
		Map<String, Map<String, Integer>> jointFreqMap = new LinkedHashMap<String, Map<String, Integer>>();
		jointFreqMap.putAll(sampleMatrixMap);
		
		JointFrequencyDistribution JDF = new JointFrequencyDistribution(secretLabels, observableLabels, jointFreqMap);
		for(String sec : secretLabels) {
			log("sec = "+sec,3);
		}
		for(String obser : JDF.getObservableLabels()) {
			log("obs = "+obser,3);
		}
		return JDF;
    }

    
    /**
     * Computes the (posterior) joint distribution from the tool
     * leakiEst's Observations object.
     * 
     * @return the (posterior) joint distribution
     */
    public static JointDistribution computeJointDistribution(Observations obs) {
    	if(obs == null) {
            log("ERROR: The given observations (in leakiEst's format) is null.",1);
            return null;
    	}
    	// Get information on the sample
    	BigDecimal sampleSize = BigDecimal.valueOf(obs.getSampleCount());
		String[] secretLabels = obs.getInputNames();
		String[] observableLabels = obs.getOutputNames();
		LinkedHashMap<String, LinkedHashMap<String, Integer>> sampleMatrixMap = obs.getObservationsMatrixMap();
		
		Map<String,Map<String,BigDecimal>> jointProbMap = new LinkedHashMap<String,Map<String,BigDecimal>>();
		for(String secret : secretLabels) {
			LinkedHashMap<String, Integer> observablesCountsMap = sampleMatrixMap.get(secret);
			if(observablesCountsMap != null) {
				// Setup a map from observables to probabilities
				Map<String,BigDecimal> observableProbMap;
				if(!jointProbMap.containsKey(secret)) {
		    	    observableProbMap = new LinkedHashMap<String,BigDecimal>();
		    	    jointProbMap.put(secret, observableProbMap);
				} else {
					observableProbMap = jointProbMap.get(secret);
				}
				// Calculate the joint probability
				for(String observable : observableLabels) {
					Integer count = observablesCountsMap.get(observable);
					if(count != null && count > 0) {
						BigDecimal countBD = BigDecimal.valueOf(count);
						BigDecimal jointProb = countBD.divide(sampleSize);
						observableProbMap.put(observable, jointProb);
					}
				}
			}
		}
    	// Construct the joint distribution
    	JointDistribution dist = new JointDistribution(secretLabels, observableLabels, jointProbMap);
    	return dist;
    }

    /**
	 * Computes the joint distribution from a given channel file (leakiEst's format).
     * 
     * @param fileName name of a given channel file (leakiEst's format)
     * @return conditional distribution representing the channel (in leakiEst's format)
     */
    public static Channel computeChannelFromChannelFile(String fileName) {
		//Calculate the channel of leakiEst's format
		ReadFile channelFileReader = new ReadFile(fileName, 0, "discrete channel");
		channelFileReader.readChannel();
		Channel channel = channelFileReader.getChannel();
		return channel;
    }
    
    /**
	 * Computes the joint distribution from a given channel file (leakiEst's format).
     * 
     * @param fileName name of a given channel file (leakiEst's format)
     * @return conditional distribution representing the channel
     */
    public static JointDistribution computeChannelDistFromChannelFile(String fileName) {
    	Channel channel = computeChannelFromChannelFile(fileName);
		//Calculate the joint distributions
    	JointDistribution channelDist = computeDistribution(channel);
    	return channelDist;
    }
    
    /**
	 * Computes the joint distribution from a given channel file (leakiEst's format).
     * 
     * @param priorFileName name of a given prior file (leakiEst's format)
     * @param channelFileName name of a given channel file (leakiEst's format)
     * @return joint distribution (in leakiEst's format)
     */
    public static Channel computeJointDistributionFromDistributionFiles(String priorFileName, String channelFileName) {
		//Read the prior (input) distribution array using the ReadFile Class
		ReadFile priorFileReader = new ReadFile(priorFileName, 0, "discrete prior");
		priorFileReader.readDistribution();
    	ProbDist pd = priorFileReader.getDistribution();

		//Calculate the channel of leakiEst's format
		ReadFile channelFileReader = new ReadFile(channelFileName, 0, "discrete channel");
		channelFileReader.readChannel();
		Channel channel = channelFileReader.getChannel();
		
		//Compute the joint distribution from the prior and channel
		Channel joint = channel.getJointDist(pd);
		return joint;
    }
    
    /**
	 * Computes the joint distribution from a given channel file (leakiEst's format).
     * 
     * @param priorFileName name of a given prior file (leakiEst's format)
     * @param channelFileName name of a given channel file (leakiEst's format)
     * @return joint distribution
     */
    public static JointDistribution computeJointDistFromDistributionFiles(String priorFileName, String channelFileName) {
		//Compute the joint distribution from the prior and channel
    	Channel joint = computeJointDistributionFromDistributionFiles(priorFileName, channelFileName);
		
		//Calculate the joint distributions
    	JointDistribution jointDist = computeDistribution(joint);
    	return jointDist;
    }
    
    /**
	 * Computes the joint distribution from a given observations file (leakiEst's format).
     * 
     * @param fileName name of a given observations file (leakiEst's format)
     * @return joint distribution
     */
    public static JointDistribution computeJointDistributionFromObsFile(String fileName) {
		//Calculate the observations of leakiEst's format
		ReadFile obsReader = new ReadFile(fileName, 0, "discrete observations");
		obsReader.readObservations();
		Observations obs = obsReader.getObservations();
		//Calculate the joint distributions
    	JointDistribution jointDist = computeJointDistribution(obs);
    	return jointDist;
    }
    
    /**
	 * Computes the joint distribution from a given ARFF file and given features.
     * 
     * @param fileName name of a given ARFF file
	 * @param highFeaturesSet set of strings representing high features
	 * @param lowFeaturesSet set of strings representing low features
     * @return joint distribution
     */
    public static JointDistribution computeJointDistributionFromARFFFile(String fileName, TreeSet<String> highFeaturesSet, TreeSet<String> lowFeaturesSet) {
		//Calculate the observations of leakiEst's format
    	Observations obs = ARFFFile.computeObservationsFromARFF(fileName, highFeaturesSet, lowFeaturesSet);
		//Calculate the joint distributions
    	JointDistribution jointDist = computeJointDistribution(obs);
    	return jointDist;
    }

    /////////////////////////////////////////////////////////////
    // extenson of JointPosteriorDistributionManager with leakiEst's objects
    /*
     * Computes the (initial) sample sizes for the staistical simulations of
     * components.
     *
     * @param mapSubProbDist hash map that maps a componentID to its sub-prior (of leakiEst's format)
     * @param roundsLeft the number of rounds left
     * @return the initial sample sizes for the staistical simulations of components
     */
    /*
    public Map<Signature, BigInteger> computeSampleSizesInitially(JointPosteriorDistributionManager postDistMan, HashMap<Signature, ProbDist> mapSubProbDist, int roundsLeft) {
    	HashMap<Signature, SingleDistribution> mapSubPriors = new HashMap<Signature, SingleDistribution>();
    	for(Signature sig : mapSubProbDist.keySet()) {
    		ProbDist pd = mapSubProbDist.get(sig);
    		SingleDistribution dist = ConvertLeakiEstObjects.computeSingleDistribution(pd);
    		mapSubPriors.put(sig, dist);
    	}
    	Map<Signature, BigInteger> mapSampleSizes = postDistMan.computeSampleSizesInitially(mapSubPriors, roundsLeft);
    	return mapSampleSizes;
    }
    */
    
    
    /////////////////////////////////////////////////////////////
    // log
    private static void log(Object aObject, int level){
        if (logger.getVerbosity() >= level) {
            Logger loggy = Logger.getInstance();
            loggy.log(String.valueOf(aObject),level);
        }
    }
}
