/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package theory;

import java.math.BigInteger;
import java.math.BigDecimal;
import java.util.HashMap;
import java.util.Set;

import distributions.JointDistribution;
import distributions.SingleDistribution;
import support.Contexter;
import support.Logger;

/**
*
* @author 
*/
public class EstimateMI {
    protected static final Contexter contexter = Contexter.getInstance();
    private static final Logger logger = Logger.getInstance();
	
    /*
     * Notations in our paper
     * 
     * P        joint distribution of the whole system S
     * pi       prior distribution of the whole system S
     * n_i      samplze size for component S_i
     * R_i      joint sub-distribution of component S_i
     *          P = sum_i R_i
     * D_i      normalized joint sub-distribution of component S_i
     *          D_i = R_i / prob(S_i)
     * pi_i     sub-prior distribution of component S_i
     *          pi = sum_i pi_i
     * xi_i     coefficient (pi_i/lambda_i)
     */

	/**
     * Calculates the bias of an estimated non-zero mutual information estimate
     * without using the abstraction-then-sampling technique.
     * 
	 * @param totalJoint joint distribution of the whole system
	 * @param distSecrets distribution of secrets of the whole system
	 * @param distObservables distribution of observables of the whole system
	 * @param listSampleSizes array of the sample sizes of all components
	 * @param listSubJoints array of the joint sub-distributions of all components
	 * @param listExecProbs array of the probability of executing each component
	 * @param listSubPriors array of the sub-prior distribution of each component
	 * @param priorIsKnown whetther the prior is given to the analyst 
     * @return the bias of an estimated non-zero mutual information estimate
     */
	public static BigDecimal getBiasOfEstimatedMI(JointDistribution totalJoint, SingleDistribution distSecrets, SingleDistribution distObservables, BigInteger[] listSampleSizes, JointDistribution[] listSubJoints, BigDecimal[] listExecProbs, SingleDistribution[] listSubPriors, boolean priorIsKnown) {
		boolean[] listAbstractables = new boolean[listSubJoints.length];
		for(int i = 0; i < listAbstractables.length; i++) {
			listAbstractables[i] = false;
		}
		return getBiasOfEstimatedMI(totalJoint, distSecrets, distObservables, listSampleSizes, listSubJoints, listExecProbs, listSubPriors, listAbstractables, priorIsKnown);
	}
    
	/**
     * Calculates the bias of an estimated non-zero mutual information estimate
     * possibly by using the abstraction-then-sampling technique.
     * 
	 * @param totalJoint joint distribution of the whole system
	 * @param distSecrets distribution of secrets of the whole system
	 * @param distObservables distribution of observables of the whole system
	 * @param listSampleSizes array of the sample sizes of all components
	 * @param listSubJoints array of the joint sub-distributions of all components
	 * @param listExecProbs array of the probability of executing each component
	 * @param listSubPriors array of the sub-prior distribution of each component
	 * @param listAbstractables array of whether the abstraction-then-sampling
	 *        can be applied to each component
	 * @param priorIsKnown whetther the prior is given to the analyst 
     * @return the bias of an estimated non-zero mutual information estimate
     */
	public static BigDecimal getBiasOfEstimatedMI(JointDistribution totalJoint, SingleDistribution distSecrets, SingleDistribution distObservables, BigInteger[] listSampleSizes, JointDistribution[] listSubJoints, BigDecimal[] listExecProbs, SingleDistribution[] listSubPriors, boolean[] listAbstractables, boolean priorIsKnown) {
        logger.changeVerbosity("MI_BIAS");
		log("=========================================",3);
		//totalJoint.print("totalJoint",2);
		//distSecrets.print("distSecrets",2);
		//distObservables.print("distObservables",2);
		log(listSampleSizes,"listSampleSizes",3);
		for(int i = 0; i < listSubJoints.length; i++) {
			if(listSubJoints[i] != null) {
				listSubJoints[i].print("listSubJoints["+i+"]",3);
			}
			if(listExecProbs[i] != null) {
				log("listExecProbs["+i+"] = "+listExecProbs[i],3);
			}
		}
		
		BigDecimal biasWithBaseE = BigDecimal.ZERO;
		for(int i = 0; i < listSubJoints.length; i++) {
			if(listSampleSizes[i] == null) {
	            log("CAUTION(ESE68): listSampleSizes+["+i+"] is null in getBiasOfEstimatedShannonEntropy.",1);
			}
			if(listSampleSizes[i] != null && listSampleSizes[i].compareTo(BigInteger.ZERO) != 0) {
				BigDecimal sum = BigDecimal.ZERO;
				if(listAbstractables[i]) {
					// calculate ratio
					BigDecimal ratio = listExecProbs[i].multiply(listExecProbs[i]).divide(BigDecimal.valueOf(2).multiply(new BigDecimal(listSampleSizes[i])), contexter.getContext().getPrecision(), BigDecimal.ROUND_HALF_UP);
					// calculate sumPsiXY
					BigDecimal sumPsiXY = computeSumPsi(listSubJoints[i], totalJoint, listExecProbs[i], listSubPriors[i]);
					// calculate sumPhiY
					BigDecimal sumPhiY = EstimateShannonEntropy.computeSumPhi(listSubJoints[i].getProjectionObservables(), distObservables, listExecProbs[i]);
					// sum
					sum = ratio.multiply(sumPsiXY.subtract(sumPhiY));
				} else if(priorIsKnown) {
					// TODO: to be verified
					Set<String> domainSecrets = listSubJoints[i].getSecretLabels();
					for(String input : domainSecrets) {
						// calculate ratio
						BigDecimal execProb = listSubPriors[i].getProbability(input);
						//log("  execProb[" + i + "](" + input + ") = " + execProb,1);
						if(listExecProbs[i] != null && listExecProbs[i].compareTo(BigDecimal.ZERO) != 0 && execProb.compareTo(BigDecimal.ZERO) != 0) {
							//BigDecimal sampleSize = (new BigDecimal(listSampleSizes[i])).multiply(execProb).divide(listExecProbs[i], contexter.getContext().getPrecision(), BigDecimal.ROUND_HALF_UP).setScale(0,BigDecimal.ROUND_DOWN);
							//log("  sampleSize = " + sampleSize + " from listSampleSizes["+i+"] = "+listSampleSizes[i],1);
							BigDecimal ratio = execProb.multiply(execProb).divide(BigDecimal.valueOf(2).multiply(new BigDecimal(listSampleSizes[i])), contexter.getContext().getPrecision(), BigDecimal.ROUND_HALF_UP);
							//log("  ratio      = " + ratio,1);
							// calculate sumPhiXY(input)
							BigDecimal sumPhiXY = computeSumPhi(listSubJoints[i], totalJoint, execProb, input);
							//log("  sumPhiXY   = " + sumPhiXY,1);
							// calculate sumPhiY(input)
							BigDecimal sumPhiY = EstimateShannonEntropy.computeSumPhi(listSubJoints[i].getRow(input), distObservables, execProb);
							//log("  sumPhiY    = " + sumPhiY,1);
							// sum
							sum = sum.add(ratio.multiply(sumPhiXY.subtract(sumPhiY)));
						}
					}
				} else {
					// calculate ratio
					BigDecimal ratio = listExecProbs[i].multiply(listExecProbs[i]).divide(BigDecimal.valueOf(2).multiply(new BigDecimal(listSampleSizes[i])), contexter.getContext().getPrecision(), BigDecimal.ROUND_HALF_UP);
					// calculate sumPhiXY
					BigDecimal sumPhiXY = computeSumPhi(listSubJoints[i], totalJoint, listExecProbs[i]);
					// calculate sumPhiX
					BigDecimal sumPhiX = EstimateShannonEntropy.computeSumPhi(listSubJoints[i].getProjectionSecrets(), distSecrets, listExecProbs[i]);
					// calculate sumPhiY
					BigDecimal sumPhiY = EstimateShannonEntropy.computeSumPhi(listSubJoints[i].getProjectionObservables(), distObservables, listExecProbs[i]);
					// sum
					sum = ratio.multiply(sumPhiXY.subtract(sumPhiX).subtract(sumPhiY));
				}

				// bias
				biasWithBaseE = biasWithBaseE.add(sum);
				//System.out.println("  biasWithBaseE = " + biasWithBaseE);
			}
		}
		
		BigDecimal biasWithBase2 = biasWithBaseE.divide( BigFunctions.ln(BigDecimal.valueOf(2), contexter.getContext().getPrecision()), contexter.getContext().getPrecision(), BigDecimal.ROUND_HALF_UP );
		//BigDecimal biasWithBase2 = biasWithBaseE.multiply( BigFunctions.log2(BigDecimal.valueOf(Math.E), contexter.getContext().getPrecision()), contexter.getContext());
		//System.out.println("  biasWithBase2 = " + biasWithBase2);
        logger.restoreVerbosity();
		return biasWithBase2;
	}
	
	/**
	 * Computes sum_{x,y} Di[x,y] (1 - Di[x,y]) / P[x,y].
	 * 
	 * @param Ri the joint sub-distribution of the component
	 * @param P the joint total distribution of the whole system
	 * @param execProb the probability of executing the component
	 * @return sum_{x,y} Di[x,y] (1 - Di[x,y]) / P[x,y]
	 */
	protected static BigDecimal computeSumPhi(JointDistribution Ri, JointDistribution P, BigDecimal execProb) {
		if(execProb.compareTo(BigDecimal.ZERO) == 0) {
			return BigDecimal.ZERO;
		} else {
			BigDecimal sumPhi = BigDecimal.ZERO;
			Set<String> domainSecrets = Ri.getSecretLabels();
			Set<String> domainObservables = Ri.getObservableLabels();
			for(String input : domainSecrets) {
				for(String output : domainObservables) {
					BigDecimal pxy = P.getProbability(input, output);
					BigDecimal dixy = Ri.getProbability(input, output).divide(execProb, contexter.getContext().getPrecision(), BigDecimal.ROUND_HALF_UP );
					if(pxy != null && pxy.compareTo(BigDecimal.ZERO) != 0 && dixy != null) {
						BigDecimal phi = EstimateShannonEntropy.computePhi(dixy, pxy);
						sumPhi = sumPhi.add(phi);
					}
				}
			}
			return sumPhi;
		}
	}

	/**
	 * Computes sum_{y} Di[x,y] (1 - Di[x,y]) / P[x,y] for a given x.
	 * 
	 * @param Ri the joint sub-distribution of the component
	 * @param P the joint total distribution of the whole system
	 * @param execProb the probability of executing the component
	 * @param input input x w.r.t which we compute sum_{y} Di[x,y] (1 - Di[x,y]) / P[x,y]
	 * @return sum_{y} Di[x,y] (1 - Di[x,y]) / P[x,y]
	 */
	private static BigDecimal computeSumPhi(JointDistribution Ri, JointDistribution P, BigDecimal execProb, String input) {
		if(execProb.compareTo(BigDecimal.ZERO) == 0) {
			return BigDecimal.ZERO;
		} else {
			BigDecimal sumPhi = BigDecimal.ZERO;
			Set<String> domainObservables = Ri.getObservableLabels();
			for(String output : domainObservables) {
				BigDecimal pxy = P.getProbability(input, output);
				BigDecimal dixy = Ri.getProbability(input, output).divide(execProb, contexter.getContext().getPrecision(), BigDecimal.ROUND_HALF_UP );
				if(pxy != null && pxy.compareTo(BigDecimal.ZERO) != 0 && dixy != null) {
					BigDecimal phi = EstimateShannonEntropy.computePhi(dixy, pxy);
					sumPhi = sumPhi.add(phi);
				}
			}
			return sumPhi;
		}
	}
	
	
	/**
	 * Computes sum_{x,y} Di[x,y] (pi[x] - Di[x,y]) / P[x,y]
	 * where pi[x] := subPrior[x] / execProb.
	 * 
	 * @param Ri the joint sub-distribution of the component
	 * @param P the joint total distribution of the whole system
	 * @param execProb the probability of executing the component
	 * @param subPrior 
	 * @return sum_{x,y} Di[x,y] (pi[x] - Di[x,y]) / P[x,y]
	 */
	protected static BigDecimal computeSumPsi(JointDistribution Ri, JointDistribution P, BigDecimal execProb, SingleDistribution subPrior) {
		if(execProb.compareTo(BigDecimal.ZERO) == 0) {
			return BigDecimal.ZERO;
		} else {
			BigDecimal sumPsi = BigDecimal.ZERO;
			Set<String> domainSecrets = Ri.getSecretLabels();
			Set<String> domainObservables = Ri.getObservableLabels();
			for(String input : domainSecrets) {
				BigDecimal pix = subPrior.getProbability(input).divide(execProb, contexter.getContext().getPrecision(), BigDecimal.ROUND_HALF_UP );
				for(String output : domainObservables) {
					BigDecimal pxy = P.getProbability(input, output);
					BigDecimal dixy = Ri.getProbability(input, output).divide(execProb, contexter.getContext().getPrecision(), BigDecimal.ROUND_HALF_UP );
					if(pxy != null && pxy.compareTo(BigDecimal.ZERO) != 0 && dixy != null) {
						BigDecimal psi = EstimateShannonEntropy.computePsi(dixy, pxy, pix);
						sumPsi = sumPsi.add(psi);
					}
				}
			}
			return sumPsi;
		}
	}
	
	
	/**
     * Calculates the variance of an estimated non-zero mutual information estimate.
     *         
	 * @param totalJoint joint distribution of the whole system
	 * @param distSecrets distribution of secrets of the whole system
	 * @param distObservables distribution of observables of the whole system
	 * @param listSampleSizes array of the sample sizes of all components
	 * @param listSubJoints array of the joint sub-distributions of all components
	 * @param listExecProbs array of the probability of executing each component
	 * @param listSubPriors array of the sub-prior distribution of each component
	 * @param priorIsKnown whetther the prior is given to the analyst 
	 * 
	 * @return the variance for an estimated non-zero mutual information estimate
	 */
	public static BigDecimal getVarianceOfEstimatedMI(JointDistribution totalJoint, SingleDistribution distSecrets, SingleDistribution distObservables, BigInteger[] listSampleSizes, JointDistribution[] listSubJoints, BigDecimal[] listExecProbs, SingleDistribution[] listSubPriors, boolean priorIsKnown) {
		boolean[] listAbstractables = new boolean[listSubJoints.length];
		for(int i = 0; i < listAbstractables.length; i++) {
			listAbstractables[i] = false;
		}
		return getVarianceOfEstimatedMI(totalJoint, distSecrets, distObservables, listSampleSizes, listSubJoints, listExecProbs, listSubPriors, listAbstractables, priorIsKnown);
	}
	
	/**
     * Calculates the variance of an estimated non-zero mutual information estimate.
     *         
	 * @param totalJoint joint distribution of the whole system
	 * @param distSecrets distribution of secrets of the whole system
	 * @param distObservables distribution of observables of the whole system
	 * @param listSampleSizes array of the sample sizes of all components
	 * @param listSubJoints array of the joint sub-distributions of all components
	 * @param listExecProbs array of the probability of executing each component
	 * @param listSubPriors array of the sub-prior distribution of each component
	 * @param listAbstractables array of whether the abstraction-then-sampling
	 *        can be applied to each component
	 * @param priorIsKnown whetther the prior is given to the analyst 
	 * 
	 * @return the variance for an estimated non-zero mutual information estimate
	 */
	public static BigDecimal getVarianceOfEstimatedMI(JointDistribution totalJoint, SingleDistribution distSecrets, SingleDistribution distObservables, BigInteger[] listSampleSizes, JointDistribution[] listSubJoints, BigDecimal[] listExecProbs, SingleDistribution[] listSubPriors, boolean[] listAbstractables, boolean priorIsKnown) {
        logger.changeVerbosity("MI_VARIANCE");
		BigDecimal varianceWithBaseE = BigDecimal.ZERO;	// Talyor expansion is done with base e
		for(int i = 0; i < listSubJoints.length; i++) {
			if(listSampleSizes[i] != null && listSampleSizes[i].compareTo(BigInteger.ZERO) != 0) {
				// domains
				Set<String> domainSecrets = listSubJoints[i].getSecretLabels();
				Set<String> domainObservables = listSubJoints[i].getObservableLabels();
				//totalJoint.print("totalJoint:::");
				//distObservables.print("distObservables:::");
				
				// calculate the variance
				BigDecimal var = BigDecimal.ZERO;
				if(listAbstractables[i]) {
					BigDecimal ratio = listExecProbs[i].multiply(listExecProbs[i]).divide(new BigDecimal(listSampleSizes[i]), contexter.getContext().getPrecision(), BigDecimal.ROUND_HALF_UP);
					// For each y, compute D_{Yi}[y] and log(P_Y[y]) - pi_i[x] log P_{XY}[x, y]
				    HashMap<String, BigDecimal> mapDyiy = new HashMap<>();
				    HashMap<String, BigDecimal> mapForm = new HashMap<>();
					for(String output : domainObservables) {
						BigDecimal outputProb = distObservables.getProbability(output);
						if(outputProb != null && outputProb.compareTo(BigDecimal.ZERO) != 0) {
							// Compute pi_i[x] log P_{XY}[x, y]
							BigDecimal sumPiLog = BigDecimal.ZERO;
							for(String input : domainSecrets) {
								BigDecimal probInput = listSubPriors[i].getProbability(input).divide(listExecProbs[i], contexter.getContext().getPrecision(), BigDecimal.ROUND_HALF_UP );
								BigDecimal piLog = probInput.multiply(BigFunctions.ln(listSubJoints[i].getProbability(input,output), contexter.getContext().getPrecision()));
								sumPiLog = sumPiLog.add(piLog);
							}
							// Compute the subtraction log(P_Y[y]) - pi_i[x] log P_{XY}[x, y]
							BigDecimal formula = BigFunctions.ln(outputProb, contexter.getContext().getPrecision()).subtract(sumPiLog);
							mapForm.put(output, formula);
							// Compute D_{Yi}[y]
							BigDecimal Dyiy = BigDecimal.ZERO;
							for(String input : domainSecrets) {
								BigDecimal Dixy = listSubJoints[i].getProbability(input, output).divide(listSubPriors[i].getProbability(input), contexter.getContext().getPrecision(), BigDecimal.ROUND_HALF_UP );
								Dyiy = Dyiy.add(Dixy);
							}
							mapDyiy.put(output, Dyiy);
						}
					}
					// Compute v1 and v2
					BigDecimal v1 = BigDecimal.ZERO;
					BigDecimal v2 = BigDecimal.ZERO;
					for(String output : domainObservables) {
						BigDecimal Dyiy = mapDyiy.get(output);
						BigDecimal tmp = mapForm.get(output);
						BigDecimal v2y = Dyiy.multiply(tmp);
						BigDecimal v1y = v2y.multiply(tmp);
						v1 = v1.add(v1y);
						v2 = v2.add(v2y);
					}
					var = (v1.subtract( v2.multiply(v2) )).multiply( ratio );
				} else if(priorIsKnown) {
					for(String input : domainSecrets) {
						// TODO: to be verified
						BigDecimal execProb = listSubPriors[i].getProbability(input);
						//log("  execProb[" + i + "](" + input + ") = " + execProb,1);
						if(listExecProbs[i] != null && listExecProbs[i].compareTo(BigDecimal.ZERO) != 0) {
							// calculate ratio
							//BigDecimal sampleSize = (new BigDecimal(listSampleSizes[i])).multiply(execProb).divide(listExecProbs[i], contexter.getContext().getPrecision(), BigDecimal.ROUND_HALF_UP).setScale(0,BigDecimal.ROUND_DOWN);
							//log("  sampleSize = " + sampleSize + " from listSampleSizes["+i+"] = "+listSampleSizes[i],1);
							BigDecimal ratio = execProb.multiply(execProb).divide(BigDecimal.valueOf(2).multiply(new BigDecimal(listSampleSizes[i])), contexter.getContext().getPrecision(), BigDecimal.ROUND_HALF_UP);
							//log("  ratio      = " + ratio,1);
							// calculate v1 and v2
							BigDecimal v1 = BigDecimal.ZERO;
							BigDecimal v2 = BigDecimal.ZERO;
							BigDecimal px = distSecrets.getProbability(input);
							if(px != null && px.compareTo(BigDecimal.ZERO) != 0) {
								for(String output : domainObservables) {
									BigDecimal pxy = totalJoint.getProbability(input,output);
									BigDecimal dixy = listSubJoints[i].getProbability(input,output).divide(execProb, contexter.getContext().getPrecision(), BigDecimal.ROUND_HALF_UP );
									BigDecimal py = distObservables.getProbability(output);
									if(pxy != null && pxy.compareTo(BigDecimal.ZERO) != 0 && dixy != null && py != null) {
										BigDecimal frac = px.multiply(py).divide(pxy, contexter.getContext().getPrecision(), BigDecimal.ROUND_HALF_UP);
										BigDecimal log = BigDecimal.ONE.add( BigFunctions.ln(frac, contexter.getContext().getPrecision()) );
										v1 = v1.add( dixy.multiply(log).multiply(log) );
										v2 = v2.add( dixy.multiply(log) );
									}
								}
							}
							// calculate sum
							var = var.add((v1.subtract( v2.multiply(v2) )).multiply( ratio ));
						}
					}
				} else {
					BigDecimal ratio = listExecProbs[i].multiply(listExecProbs[i]).divide(new BigDecimal(listSampleSizes[i]), contexter.getContext().getPrecision(), BigDecimal.ROUND_HALF_UP);
					BigDecimal v1 = BigDecimal.ZERO;
					BigDecimal v2 = BigDecimal.ZERO;
					for(String input : domainSecrets) {
						BigDecimal px = distSecrets.getProbability(input);
						if(px != null && px.compareTo(BigDecimal.ZERO) != 0) {
							for(String output : domainObservables) {
								BigDecimal pxy = totalJoint.getProbability(input,output);
								BigDecimal dixy = listSubJoints[i].getProbability(input,output).divide(listExecProbs[i], contexter.getContext().getPrecision(), BigDecimal.ROUND_HALF_UP );
								BigDecimal py = distObservables.getProbability(output);
								if(pxy != null && pxy.compareTo(BigDecimal.ZERO) != 0 && dixy != null && py != null) {
									BigDecimal frac = px.multiply(py).divide(pxy, contexter.getContext().getPrecision(), BigDecimal.ROUND_HALF_UP);
									BigDecimal log = BigDecimal.ONE.add( BigFunctions.ln(frac, contexter.getContext().getPrecision()) );
									v1 = v1.add( dixy.multiply(log).multiply(log) );
									v2 = v2.add( dixy.multiply(log) );
								}
							}
						}
					}
					var = (v1.subtract( v2.multiply(v2) )).multiply( ratio );
				}
				/*
				System.out.println("   sampleSize_" + i + " = " + listSampleSizes[i]);
				System.out.println("   v1      = " + v1);
				System.out.println("   v2      = " + v2);
				System.out.println("   var     = " + var);
				*/
				varianceWithBaseE = varianceWithBaseE.add( var );
			}
		}
		
		BigDecimal logE2 = BigFunctions.ln(BigDecimal.valueOf(2), contexter.getContext().getPrecision());
		BigDecimal varianceWithBase2 = varianceWithBaseE.divide( logE2.multiply(logE2), contexter.getContext().getPrecision(), BigDecimal.ROUND_HALF_UP ); // Here we use base 2
		log(" varianceWithBase2 = "+varianceWithBase2,4);
        logger.restoreVerbosity();

		return varianceWithBase2;
	}
	
	/**
	 * Calculates the confidence interval for corrected mutual information.
	 * 
	 * @param MICorrectedPointEstimate point estimate of the mutual information after correction of bias
	 * @param totalJoint joint distribution of the whole system
	 * @param distSecrets distribution of secrets of the whole system
	 * @param distObservables distribution of observables of the whole system
	 * @param listSampleSizes array of the sample sizes of all components
	 * @param listSubJoints array of the joint sub-distributions of all components
	 * @param listExecProbs array of the probability of executing each component
	 * @param listSubPriors array of the sub-prior distribution of each component
	 * @param priorIsKnown whetther the prior is given to the analyst 
	 * 
	 * @return Confidence interval
	 */
	public static BigDecimal[] getCorrectedMIConfidenceInterval(BigDecimal MICorrectedPointEstimate, JointDistribution totalJoint, SingleDistribution distSecrets, SingleDistribution distObservables, BigInteger[] listSampleSizes, JointDistribution[] listSubJoints, BigDecimal[] listExecProbs, SingleDistribution[] listSubPriors, boolean priorIsKnown) {
		//BigDecimal bias = getBiasOfEstimatedMIUnderKnownPrior(totalJoint, distObservables, listSampleSizes, listSubJoints, listExecProbs, listImpPriors);
		//BigDecimal mean = MIPointEstimate.subtract(bias).max(BigDecimal.ZERO);
		BigDecimal variance = getVarianceOfEstimatedMI(totalJoint, distSecrets, distObservables, listSampleSizes, listSubJoints, listExecProbs, listSubPriors, priorIsKnown);
		BigDecimal[] interval = new BigDecimal[2];
		int scale = contexter.getContext().getPrecision();
		interval[0] = Statistics.lowerBoundNormal95(MICorrectedPointEstimate, variance, scale).max(BigDecimal.ZERO); // lower bound
		interval[1] = Statistics.upperBoundNormal95(MICorrectedPointEstimate, variance, scale).max(BigDecimal.ZERO); // upper bound
		
		/*
		System.out.println("--------------------------------------------");
		//System.out.println("point.est = " + MIPointEstimate);
		//System.out.println("bias      = " + bias);
		System.out.println("mean      = " + MICorrectedPointEstimate);
		System.out.println("variance  = " + variance);
		System.out.println("lower     = " + interval[0]);
		System.out.println("upper     = " + interval[1]);
		System.out.println("--------------------------------------------");
		*/
		return interval;
	}

	/**
	 * Initializes the sample sizes for statistical simulations.
	 * 
	 * @param totalSampleSize the total number of simulations
	 *        (i.e. the sume of all sample sizes)
	 * @param listExecProbs array of the probability of executing each component
	 * @return the initial sample sizes for statistical simulation
	 */
	public static BigInteger[] initializeSampleSizes(BigInteger totalSampleSize, BigDecimal[] listExecProbs) {
		int numOfComponents = listExecProbs.length;
		BigDecimal[] ratios = new BigDecimal[numOfComponents];
		ratios[0] = BigDecimal.ONE;  // by trace analysis
		for(int i = 1; i < numOfComponents; i++) {
			ratios[i] = listExecProbs[i];
		}
		return computeInitialSampleSizes(totalSampleSize, ratios);
	}

	/*
	 * Initializes the sample sizes for statistical simulations.
	 * 
	 * @param totalSampleSize the total number of simulations
	 *        (i.e. the sume of all sample sizes)
	 * @param listExecProbs array of the probability of executing each component
	 * @param listSizesOfSupports array of the sizes of supports
	 * @return the initial sample sizes for statistical simulation
	 */
	/*
	private static BigInteger[] initializeSampleSizes(BigInteger totalSampleSize, BigDecimal[] listExecProbs, int[] listSizesOfSupports) {
		int numOfComponents = listExecProbs.length;
		BigDecimal[] ratios = new BigDecimal[numOfComponents];
		ratios[0] = BigDecimal.ONE;  // by trace analysis
		for(int i = 1; i < numOfComponents; i++) {
			ratios[i] = listExecProbs[i].divide(BigDecimal.valueOf(listSizesOfSupports[i]), contexter.getContext().getPrecision(), BigDecimal.ROUND_HALF_UP);
		}
		return computeInitialSampleSizes(totalSampleSize, ratios);
	}
	*/
	
	/**
	 * TODO: This method is just temporal and should be modified after
	 * supporting the estimated matrix size supplied by the preprocessor.
	 * 
	 * Initializes the sample sizes for statistical simulations.
	 * 
	 * @param totalSampleSize the total number of simulations
	 *        (i.e. the sume of all sample sizes)
	 * @param listExecProbs array of the probability of executing each component
	 * @param listEstimatedSizesOfObservables array of the estimated sizes of observables
	 * @param listSizesOfSupports array of the sizes of supports
	 * @return the initial sample sizes for statistical simulation
	 */
	public static BigInteger[] initializeSampleSizes(BigInteger totalSampleSize, BigDecimal[] listExecProbs, int[] listEstimatedSizesOfObservables, int[] listSizesOfSupports) {
		int numOfComponents = listExecProbs.length;
		BigDecimal[] ratios = new BigDecimal[numOfComponents];
		ratios[0] = BigDecimal.ONE;  // by trace analysis
		for(int i = 1; i < numOfComponents; i++) {
	        //log("EMI504: listSizesOfObservables[i] = "+listEstimatedSizesOfObservables[i],1);
			//BigDecimal fraction = BigDecimal.valueOf(0.9); //TODO: This is just a temporal and should be replaced later
			//BigDecimal fraction = BigDecimal.ONE.subtract(BigDecimal.ONE.divide(BigDecimal.valueOf(listEstimatedSizesOfObservables[i]), contexter.getContext().getPrecision(), BigDecimal.ROUND_HALF_UP));
			BigDecimal fraction = BigDecimal.ONE.divide(BigDecimal.valueOf(listSizesOfSupports[i]).max(BigDecimal.ONE), contexter.getContext().getPrecision(), BigDecimal.ROUND_HALF_UP);
	        //log("EMI507: fraction = "+fraction,1);
			ratios[i] = listExecProbs[i].multiply(fraction);
		}
		return computeInitialSampleSizes(totalSampleSize, ratios);
	}
	
	/*
	 * Computes the initial sample sizes for statistical simulations.
	 * 
	 * @param totalSampleSize the total number of simulations
	 *        (i.e. the sume of all sample sizes)
	 * @param ratios array of the ratios of assigning sample sizes
	 * @return the initial sample sizes for statistical simulation
	 */
	private static BigInteger[] computeInitialSampleSizes(BigInteger totalSampleSize, BigDecimal[] ratios) {
		int numOfComponents = ratios.length;
		BigDecimal sumProbs = BigDecimal.ZERO;
		for(int i = 1; i < numOfComponents; i++) {
			sumProbs = sumProbs.add(ratios[i]);
	        //log("EMI227: sumProbs = "+sumProbs,1);
		}
		BigInteger[] listSampleSizes = new BigInteger[numOfComponents];
		listSampleSizes[0] = BigInteger.ZERO;  // by trace analysis
		for(int i = 1; i < numOfComponents; i++) {
			if(sumProbs.compareTo(BigDecimal.ZERO) == 0) {
				listSampleSizes[i] = BigInteger.ZERO;
			} else {
				listSampleSizes[i] = (new BigDecimal(totalSampleSize)).multiply(ratios[i]).divide(sumProbs, contexter.getContext().getPrecision(), BigDecimal.ROUND_HALF_UP).add(BigDecimal.valueOf(0.5)).toBigInteger();
			}
		}
		return listSampleSizes;
	}

	/*
     * Computes the information for comparing variances of sub-MI's of components
     * in order to update the sample sizes in statistical simulations.
     *         
	 * @param totalJoint joint distribution of the whole system
	 * @param distSecrets distribution of secrets of the whole system
	 * @param distObservables distribution of observables of the whole system
	 * @param listSubJoints array of the joint sub-distributions of all components
	 * @param listExecProbs array of the probability of executing each component
	 * @return the information for updating the sample sizes in statistical simulations.
	 */
	private static BigDecimal[] getInfoForUpdatingSampleSizes(JointDistribution totalJoint, SingleDistribution distSecrets, SingleDistribution distObservables, JointDistribution[] listSubJoints, BigDecimal[] listExecProbs) {
    	if(totalJoint == null) {
            log("ERROR(EMI255): totalJoint is null in getInfoForUpdatingSampleSizes.",1);
            System.exit(1);
    	}
    	if(distSecrets == null) {
            log("ERROR(EMI259): distSecrets is null in getInfoForUpdatingSampleSizes.",1);
            System.exit(1);
    	}
    	if(distObservables == null) {
            log("ERROR(EMI263): distObservables is null in getInfoForUpdatingSampleSizes.",1);
            System.exit(1);
    	}
    	
		Set<String> domainSecrets = totalJoint.getSecretLabels();
		Set<String> domainObservables = totalJoint.getObservableLabels();
		
		BigDecimal[] scaleOfVariances = new BigDecimal[listSubJoints.length];
		scaleOfVariances[0] = BigDecimal.ZERO;  // by trace analysis
		for(int i = 1; i < listSubJoints.length; i++) {
			// calculate the variance coefficient
			BigDecimal square = listExecProbs[i].multiply(listExecProbs[i]);
			
			BigDecimal v1 = BigDecimal.ZERO;
			BigDecimal v2 = BigDecimal.ZERO;
			// calculate the variance
			for(String input : domainSecrets) {
				BigDecimal px = distSecrets.getProbability(input);
				if(px != null && px.compareTo(BigDecimal.ZERO) != 0) {
					for(String output : domainObservables) {
						BigDecimal pxy = totalJoint.getProbability(input,output);
						BigDecimal dixy = listSubJoints[i].getProbability(input,output).divide(listExecProbs[i], contexter.getContext().getPrecision(), BigDecimal.ROUND_HALF_UP );
						BigDecimal py = distObservables.getProbability(output);
						if(pxy != null && pxy.compareTo(BigDecimal.ZERO) != 0 && dixy != null && py != null) {
							BigDecimal frac = px.multiply(py).divide(pxy, contexter.getContext().getPrecision(), BigDecimal.ROUND_HALF_UP);
							//log("px   = "+px,1);
							//log("py   = "+py,1);
							//log("pxy  = "+pxy,1);
							//log("frac = "+frac,1);
							BigDecimal log = BigDecimal.ONE.add( BigFunctions.ln(frac, contexter.getContext().getPrecision()) );
							v1 = v1.add( dixy.multiply(log).multiply(log) );
							v2 = v2.add( dixy.multiply(log) );
						}
					}
				}
			}
			scaleOfVariances[i] = (v1.subtract( v2.multiply(v2) )).multiply( square );
			/*
			log("   v1      = " + v1,4);
			log("   v2      = " + v2,4);
			log("   square  = " + square,4);
			log("   var     = " + scaleOfVariances[i],4);
			*/
		}
		return scaleOfVariances;
	}
	
	/**
	 * Updates the sample sizes for statistical simulations
	 * by using trial sample.
	 * 
	 * @param totalSampleSize the total number of simulations
	 *        (i.e. the sume of all sample sizes)
	 * @param usedSampleSizes the number of simulations used for a trial sample
	 * @param totalJoint joint distribution of the whole system
	 * @param distSecrets distribution of secrets of the whole system
	 * @param distObservables distribution of observables of the whole system
	 * @param listSubJoints array of the joint sub-distributions of all components
	 * @param listExecProbs array of the probability of executing each component
	 * @return the updated sample sizes for statistical simulations
	 */
	public static BigInteger[] updateSampleSizes(BigInteger totalSampleSize, BigInteger[] usedSampleSizes, JointDistribution totalJoint, SingleDistribution distSecrets, SingleDistribution distObservables, JointDistribution[] listSubJoints, BigDecimal[] listExecProbs) {
		int numOfComponents = listExecProbs.length;
		log("   numOfComponents = "+numOfComponents,3);
		BigDecimal totalSampleSizeBD = new BigDecimal(totalSampleSize);
		
		// get scales of variances of all components
		BigDecimal[] scaleOfVariances = getInfoForUpdatingSampleSizes(totalJoint, distSecrets, distObservables, listSubJoints, listExecProbs);
		BigDecimal[] sqrtScales = new BigDecimal[numOfComponents];
		sqrtScales[0] = BigDecimal.ZERO;  // by trace analysis
		BigDecimal sumSqrtScales = BigDecimal.ZERO;
		for(int i = 1; i < numOfComponents; i++) {
			log("   scaleOfVariances["+i+"] = "+scaleOfVariances[i],3);
            sqrtScales[i] = BigFunctions.sqrt(scaleOfVariances[i], contexter.getContext().getPrecision());
			sumSqrtScales = sumSqrtScales.add(sqrtScales[i]);
		}

		// (temporal) new sample sizes of components
		BigDecimal[] listSampleSizesBD = new BigDecimal[numOfComponents];
		listSampleSizesBD[0] = BigDecimal.ZERO;  // by trace analysis
		BigDecimal sumSampleSizes = BigDecimal.ZERO;
		for(int i = 1; i < numOfComponents; i++) {
			if(sumSqrtScales.compareTo(BigDecimal.ZERO) == 0) {
				listSampleSizesBD[i] = BigDecimal.ZERO;
			} else {
				listSampleSizesBD[i] = sqrtScales[i].multiply(totalSampleSizeBD).divide(sumSqrtScales, contexter.getContext().getPrecision(), BigDecimal.ROUND_HALF_UP);
				listSampleSizesBD[i] = listSampleSizesBD[i].max(new BigDecimal(usedSampleSizes[i]));
				sumSampleSizes = sumSampleSizes.add(listSampleSizesBD[i]);
			}
		}
		
		// in case sumSampleSizes is larger than the totalSampleSize
		BigDecimal totalSampleSizeDB = new BigDecimal(totalSampleSize);
		BigDecimal diff = sumSampleSizes.subtract(totalSampleSizeDB);
		if(diff.compareTo(BigDecimal.ZERO) > 0) {
			totalSampleSizeBD = totalSampleSizeDB.subtract(diff);
			for(int i = 1; i < numOfComponents; i++) {
				if(sumSqrtScales.compareTo(BigDecimal.ZERO) == 0) {
					listSampleSizesBD[i] = BigDecimal.ZERO;
				} else {
					listSampleSizesBD[i] = sqrtScales[i].multiply(totalSampleSizeBD).divide(sumSqrtScales, contexter.getContext().getPrecision(), BigDecimal.ROUND_HALF_UP);
					listSampleSizesBD[i] = listSampleSizesBD[i].max(new BigDecimal(usedSampleSizes[i]));
					sumSampleSizes = sumSampleSizes.add(listSampleSizesBD[i]);
				}
			}
		}

		// convert BigDecimal array to BigInteger array
		// and then subtract the numers of used sample sizes
		BigInteger[] listSampleSizes = new BigInteger[numOfComponents];
		for(int i = 0; i < numOfComponents; i++) {
			listSampleSizes[i] = listSampleSizesBD[i].toBigInteger().subtract(usedSampleSizes[i]);
		}
		return listSampleSizes;
	}

    private static void log(Object aObject, int level){
        if (logger.getVerbosity() >= level) {
        	Logger loggy = Logger.getInstance();
        	loggy.log(String.valueOf(aObject),level);
        }
    }
    private static void log(Object[] array, String name, int level){
        if (logger.getVerbosity() >= level) {
        	Logger loggy = Logger.getInstance();
        	loggy.log(array,name,level);
        }
    }
	
}
