/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package theory;

import java.math.BigInteger;
import java.math.BigDecimal;
import java.util.Set;

import distributions.SingleDistribution;
import support.Contexter;
import support.Logger;

/**
*
* @author  
*/
public class EstimateShannonEntropy {
    protected static final Contexter contexter = Contexter.getInstance();
    private static final Logger logger = Logger.getInstance();
	protected static int verbose = 5;
	
    /*
     * Notations in our paper
     * 
     * P        single distribution of the whole system S
     * n_i      samplze size for component S_i
     * D_i      sub-distribution of component S_i
     *          P = sum_i D_i
     */

	/**
     * Calculates the bias of an estimated non-zero Shannon entropy estimate.
     * 
	 * @param singleDist single distribution
	 * @param listSubDists array of the sub-distributions of all components
	 * @param listSampleSizes array of the sample sizes of all components
	 * @param listExecProbs array of the probability of executing each component
	 * 
     * @return the bias of an estimated non-zero Shannon entropy estimate
     */
	public static BigDecimal getBiasOfEstimatedShannonEntropy(SingleDistribution totalSingleDist, SingleDistribution[] listSubDists, BigInteger[] listSampleSizes, BigDecimal[] listExecProbs) {
		log("=========================================",1);
		totalSingleDist.print("singleDist",3);
		log(listSampleSizes,"listSampleSizes",3);
		log(listExecProbs,"listExecProbs",3);
		for(int i = 0; i < listSubDists.length; i++) {
			if(listSubDists[i] != null) {
				listSubDists[i].print("listSubDists["+i+"]",3);
			}
		}
		
		BigDecimal biasWithBaseE = BigDecimal.ZERO;
		for(int i = 0; i < listSubDists.length; i++) {
			if(listSampleSizes[i] == null) {
	            log("CAUTION(ESE59): listSampleSizes+["+i+"] is null in getBiasOfEstimatedShannonEntropy.",1);
			}
			if(listSampleSizes[i] != null && listSampleSizes[i].compareTo(BigInteger.ZERO) != 0) {
				BigDecimal ratio = listExecProbs[i].multiply(listExecProbs[i]).divide(BigDecimal.valueOf(2).multiply(new BigDecimal(listSampleSizes[i])), contexter.getContext().getPrecision(), BigDecimal.ROUND_HALF_UP);
				BigDecimal sumPhi = computeSumPhi(listSubDists[i], totalSingleDist, listExecProbs[i]);
				biasWithBaseE = biasWithBaseE.add(ratio.multiply(sumPhi));
			}
			//System.out.println("  biasWithBaseE = " + biasWithBaseE);
		}
		
		BigDecimal biasWithBase2 = biasWithBaseE.divide( BigFunctions.ln(BigDecimal.valueOf(2), contexter.getContext().getPrecision()), contexter.getContext().getPrecision(), BigDecimal.ROUND_HALF_UP );
		//BigDecimal biasWithBase2 = biasWithBaseE.multiply( BigFunctions.log2(BigDecimal.valueOf(Math.E), contexter.getContext().getPrecision()), contexter.getContext());
		//System.out.println("  biasWithBase2 = " + biasWithBase2);
		return biasWithBase2;
	}

	
	/**
	 * Computes phi = d * (1-d) / p.
	 * 
	 * @param d normalized probability in the component
	 * @param p probability in the whole system
	 * @return  d * (1-d) / p
	 */
	protected static BigDecimal computePhi(BigDecimal d, BigDecimal p) {
		BigDecimal phi = d.multiply(BigDecimal.ONE.subtract(d)).divide(p, contexter.getContext().getPrecision(), BigDecimal.ROUND_HALF_UP);
		if(phi == null) {
			log("ERROR(ESE86): phi is null for d = "+d+" p = "+p, 1);
		}
		return phi;
	}

	/**
	 * Computes sum_x Di[x] (1 - Di[x]) / P[x].
	 * 
	 * @param Ri the sub-distribution of the component
	 * @param P the total distribution of the whole system
	 * @param execProb the probability of executing the component
	 * @return sum_x Di[x] (1 - Di[x]) / P[x]
	 */
	protected static BigDecimal computeSumPhi(SingleDistribution Ri, SingleDistribution P, BigDecimal execProb) {
		if(execProb.compareTo(BigDecimal.ZERO) == 0) {
			return BigDecimal.ZERO;
		} else {
			BigDecimal sumPhi = BigDecimal.ZERO;
			for(String label : P.getLabels()) {
				BigDecimal px = P.getProbability(label);
				BigDecimal dix = Ri.getProbability(label).divide(execProb, contexter.getContext().getPrecision(), BigDecimal.ROUND_HALF_UP );
				if(px != null && px.compareTo(BigDecimal.ZERO) != 0 && dix != null) {
					BigDecimal phi = computePhi(dix, px);
					sumPhi = sumPhi.add(phi);
				}
			}
			return sumPhi;
		}
	}

	/**
	 * Computes psi = d * (pr - d) / p.
	 * This methods is for ATS (abstraction-then-sampling).
	 * 
	 * @param d normalized probability in the component
	 * @param p probability in the whole system
	 * @param pr prior probability for the component
	 * @return  d * (pr - d) / p
	 */
	protected static BigDecimal computePsi(BigDecimal d, BigDecimal p, BigDecimal pr) {
		BigDecimal psi = d.multiply(pr.subtract(d)).divide(p, contexter.getContext().getPrecision(), BigDecimal.ROUND_HALF_UP);
		if(psi == null) {
			log("ERROR(ESE124): phi is null for d = "+d+" p = "+p+" p = "+pr, 1);
		}
		return psi;
	}

	/**
	 * Computes sum_x Di[x] (pi_i[x] - Di[x]) / P[x].
	 * This methods is for ATS (abstraction-then-sampling).
	 * 
	 * @param Ri the sub-distribution of the component
	 * @param P the total distribution of the whole system
	 * @param execProb the probability of executing the component
	 * @param pii prior probability distribution for the component
	 * @return sum_x Di[x] (pi_i[x] - Di[x]) / P[x]
	 */
	protected static BigDecimal computeSumPsi(SingleDistribution Ri, SingleDistribution P, BigDecimal execProb, SingleDistribution pii) {
		if(execProb.compareTo(BigDecimal.ZERO) == 0) {
			return BigDecimal.ZERO;
		} else {
			BigDecimal sumPsi = BigDecimal.ZERO;
			for(String label : P.getLabels()) {
				BigDecimal px = P.getProbability(label);
				BigDecimal dix = Ri.getProbability(label).divide(execProb, contexter.getContext().getPrecision(), BigDecimal.ROUND_HALF_UP );
				BigDecimal piix = pii.getProbability(label);
				if(px != null && px.compareTo(BigDecimal.ZERO) != 0 && dix != null) {
					BigDecimal psi = computePsi(dix, px, piix);
					sumPsi = sumPsi.add(psi);
				}
			}
			return sumPsi;
		}
	}

	/**
     * Calculates the variance of an estimated non-zero Shannon entropy estimate
     *         
	 * @param totalSingleDist total distribution of the whole system
	 * @param listSubDists array of the sub-distributions of all components
	 * @param listSampleSizes array of the sample sizes of all components
	 * @param listExecProbs array of the probability of executing each component
	 * 
	 * @return the variance for an estimated non-zero Shannon entropy estimate
	 */
	public static BigDecimal getVarianceOfEstimatedShannonEntropy(SingleDistribution totalSingleDist, SingleDistribution[] listSubDists, BigInteger[] listSampleSizes, BigDecimal[] listExecProbs) {
		Set<String> domain = totalSingleDist.getLabels();
		
		BigDecimal varianceWithBaseE = BigDecimal.ZERO;	// Talyor expansion is done with base e
		for(int i = 0; i < listSubDists.length; i++) {
			if(listSampleSizes[i] != null && listSampleSizes[i].compareTo(BigInteger.ZERO) != 0) {
				BigDecimal ratio = listExecProbs[i].multiply(listExecProbs[i]).divide(new BigDecimal(listSampleSizes[i]), contexter.getContext().getPrecision(), BigDecimal.ROUND_HALF_UP);
				BigDecimal v1 = BigDecimal.ZERO;
				BigDecimal v2 = BigDecimal.ZERO;
				// calculate the variance
				for(String label : domain) {
					BigDecimal px = totalSingleDist.getProbability(label);
					BigDecimal dix = listSubDists[i].getProbability(label);
					if(px != null && px.compareTo(BigDecimal.ZERO) != 0 && dix != null) {
						BigDecimal log = BigDecimal.ONE.add(BigFunctions.ln(px, contexter.getContext().getPrecision()));
						v1 = v1.add( dix.multiply(log).multiply(log) );
						v2 = v2.add( dix.multiply(log) );
					}
				}
				BigDecimal var = (v1.subtract( v2.multiply(v2) )).multiply( ratio );
				/*
				System.out.println("   sampleSize_" + i + " = " + listSampleSizes[i]);
				System.out.println("   v1      = " + v1);
				System.out.println("   v2      = " + v2);
				System.out.println("   var     = " + var);
				*/
				varianceWithBaseE = varianceWithBaseE.add( var );
			}
		}
		
		BigDecimal logE2 = BigFunctions.ln(BigDecimal.valueOf(2), contexter.getContext().getPrecision()); // TODO: +10?
		BigDecimal varianceWithBase2 = varianceWithBaseE.divide( logE2.multiply(logE2), contexter.getContext().getPrecision(), BigDecimal.ROUND_HALF_UP ); // Here we use base 2
		log(" varianceWithBase2 = "+varianceWithBase2,5);

		return varianceWithBase2;
	}
	
	/**
	 * Calculates the confidence interval for corrected Shannon entropy.
	 * 
	 * @param MICorrectedPointEstimate point estimate of the mutual information after correction of bias
	 * @param totalSingleDist total distribution of the whole system
	 * @param listSubDists array of the sub-distributions of all components
	 * @param listSampleSizes array of the sample sizes of all components
	 * @param listExecProbs array of the probability of executing each component
	 * 
	 * @return Confidence interval
	 */
	public static BigDecimal[] getCorrectedShannonEntropyConfidenceInterval(BigDecimal MICorrectedPointEstimate, SingleDistribution totalSingleDist, SingleDistribution[] listSubDists, BigInteger[] listSampleSizes, BigDecimal[] listExecProbs) {
		//BigDecimal bias = getBiasOfEstimatedMIUnderKnownPrior(totalSingleDist, distObservables, listSampleSizes, listSubDists, listSubPriors, listImpPriors);
		//BigDecimal mean = MIPointEstimate.subtract(bias).max(BigDecimal.ZERO);
		BigDecimal variance = getVarianceOfEstimatedShannonEntropy(totalSingleDist, listSubDists, listSampleSizes, listExecProbs);
		BigDecimal[] interval = new BigDecimal[2];
		int scale = contexter.getContext().getPrecision(); // TODO: +10?
		interval[0] = Statistics.lowerBoundNormal95(MICorrectedPointEstimate, variance, scale).max(BigDecimal.ZERO); // lower bound
		interval[1] = Statistics.upperBoundNormal95(MICorrectedPointEstimate, variance, scale); // upper bound
		
		/*
		System.out.println("--------------------------------------------");
		//System.out.println("point.est = " + MIPointEstimate);
		//System.out.println("bias      = " + bias);
		System.out.println("mean      = " + MICorrectedPointEstimate);
		System.out.println("variance  = " + variance);
		System.out.println("lower     = " + interval[0]);
		System.out.println("upper     = " + interval[1]);
		System.out.println("--------------------------------------------");
		*/
		return interval;
	}

	/**
	 * Initializes the sample sizes for statistical simulations.
	 * 
	 * @param totalSampleSize the total number of simulations
	 *        (i.e. the sume of all sample sizes)
	 * @param listSubDists array of the sub-distributions of all components
	 * 
	 * @return the (initial) importance prior distribution for statistical simulation
	 */
	public static BigInteger[] initializeSampleSizes(BigInteger totalSampleSize, SingleDistribution[] listSubDists) {
		int numOfComponents = listSubDists.length;
		BigDecimal[] ratios = new BigDecimal[numOfComponents];
		ratios[0] = BigDecimal.ONE;  // by trace analysis
		BigDecimal sumProbs = BigDecimal.ZERO;
		for(int i = 1; i < numOfComponents; i++) {
			ratios[i] = listSubDists[i].getSumOfProbabilities();
			sumProbs = sumProbs.add(ratios[i]);
	        //log("ESE212: sumProbs = "+sumProbs,1);
		}
		BigInteger[] listSampleSizes = new BigInteger[numOfComponents];
		listSampleSizes[0] = BigInteger.ZERO;  // by trace analysis
		for(int i = 1; i < numOfComponents; i++) {
			if(sumProbs.compareTo(BigDecimal.ZERO) == 0) {
				listSampleSizes[i] = BigInteger.ZERO;
			} else {
				listSampleSizes[i] = (new BigDecimal(totalSampleSize)).multiply(ratios[i]).divide(sumProbs, contexter.getContext().getPrecision(), BigDecimal.ROUND_HALF_UP).add(BigDecimal.valueOf(0.5)).toBigInteger();
			}
		}
		return listSampleSizes;
	}
	
	/*
     * Computes the information for comparing variances of sub-MI's of components
     * in order to update the sample sizes in statistical simulations.
     * 
     * 
	 * @param totalSingleDist total distribution of the whole system
	 * @param listSubDists array of the sub-distributions of all components
	 * @param listExecProbs array of the probability of executing each component
	 * 
	 * @return the information for updating the sample sizes in statistical simulations.
	 */
	private static BigDecimal[] getInfoForUpdatingSampleSizes(SingleDistribution totalSingleDist, SingleDistribution[] listSubDists, BigDecimal[] listExecProbs) {
    	if(totalSingleDist == null) {
            log("ERROR(ESE239): totalSingleDist is null in getInfoForUpdatingSampleSizes.",1);
            System.exit(1);
    	}
    	
		Set<String> domain = totalSingleDist.getLabels();
		
		BigDecimal[] scaleOfVariances = new BigDecimal[listSubDists.length];
		scaleOfVariances[0] = BigDecimal.ZERO;  // by trace analysis
		for(int i = 1; i < listSubDists.length; i++) {
			// calculate the variance coefficient
			BigDecimal square = listExecProbs[i].multiply(listExecProbs[i]);
			BigDecimal v1 = BigDecimal.ZERO;
			BigDecimal v2 = BigDecimal.ZERO;
			// calculate the variance
			for(String label : domain) {
				BigDecimal px = totalSingleDist.getProbability(label);
				BigDecimal dix = listSubDists[i].getProbability(label);
				if(px != null && px.compareTo(BigDecimal.ZERO) != 0 && dix != null) {
					BigDecimal log = BigDecimal.ONE.add(BigFunctions.ln(px, contexter.getContext().getPrecision()));
					v1 = v1.add( dix.multiply(log).multiply(log) );
					v2 = v2.add( dix.multiply(log) );
				}
			}
			scaleOfVariances[i] = (v1.subtract( v2.multiply(v2) )).multiply( square );
			/*
			System.out.println("   sampleSize_" + i + " = " + listSampleSizes[i]);
			System.out.println("   v1      = " + v1);
			System.out.println("   v2      = " + v2);
			System.out.println("   var     = " + scaleOfVariances[i]);
			*/
		}
		return scaleOfVariances;
	}
	
	/**
	 * Updates the sample sizes for statistical simulations
	 * by using trial sample.
	 * 
	 * @param totalSampleSize the total number of simulations
	 *        (i.e. the sume of all sample sizes)
	 * @param usedSampleSizes the number of simulations used for a trial sample
	 * @param totalSingleDist total distribution of the whole system
	 * @param listSubDists array of the sub-distributions of all components
	 * @param listExecProbs array of the probability of executing each component
	 * 
	 * @return the updated sample sizes for statistical simulations
	 */
	public static BigInteger[] updateSampleSizes(BigInteger totalSampleSize, BigInteger[] usedSampleSizes, SingleDistribution totalSingleDist, SingleDistribution[] listSubDists, BigDecimal[] listExecProbs) {
		int numOfComponents = listSubDists.length;
		log("   numOfComponents = "+numOfComponents,3);
		BigDecimal totalSampleSizeBD = new BigDecimal(totalSampleSize);
		
		// get scales of variances of all components
		BigDecimal[] scaleOfVariances = getInfoForUpdatingSampleSizes(totalSingleDist, listSubDists, listExecProbs);
		BigDecimal[] sqrtScales = new BigDecimal[numOfComponents];
		sqrtScales[0] = BigDecimal.ZERO;  // by trace analysis
		BigDecimal sumSqrtScales = BigDecimal.ZERO;
		for(int i = 1; i < numOfComponents; i++) {
			log("   scaleOfVariances["+i+"] = "+scaleOfVariances[i],3);
            sqrtScales[i] = BigFunctions.sqrt(scaleOfVariances[i], contexter.getContext().getPrecision());
			sumSqrtScales = sumSqrtScales.add(sqrtScales[i]);
		}

		// (temporal) new sample sizes of components
		BigDecimal[] listSampleSizesBD = new BigDecimal[numOfComponents];
		listSampleSizesBD[0] = BigDecimal.ZERO;  // by trace analysis
		BigDecimal sumSampleSizes = BigDecimal.ZERO;
		for(int i = 1; i < numOfComponents; i++) {
			if(sumSqrtScales.compareTo(BigDecimal.ZERO) == 0) {
				listSampleSizesBD[i] = BigDecimal.ZERO;
			} else {
				listSampleSizesBD[i] = sqrtScales[i].multiply(totalSampleSizeBD).divide(sumSqrtScales, contexter.getContext().getPrecision(), BigDecimal.ROUND_HALF_UP);
				listSampleSizesBD[i] = listSampleSizesBD[i].max(new BigDecimal(usedSampleSizes[i]));
				sumSampleSizes = sumSampleSizes.add(listSampleSizesBD[i]);
			}
		}
		
		// in case sumSampleSizes is larger than the totalSampleSize
		BigDecimal totalSampleSizeDB = new BigDecimal(totalSampleSize);
		BigDecimal diff = sumSampleSizes.subtract(totalSampleSizeDB);
		if(diff.compareTo(BigDecimal.ZERO) > 0) {
			totalSampleSizeBD = totalSampleSizeDB.subtract(diff);
			for(int i = 1; i < numOfComponents; i++) {
				if(sumSqrtScales.compareTo(BigDecimal.ZERO) == 0) {
					listSampleSizesBD[i] = BigDecimal.ZERO;
				} else {
					listSampleSizesBD[i] = sqrtScales[i].multiply(totalSampleSizeBD).divide(sumSqrtScales, contexter.getContext().getPrecision(), BigDecimal.ROUND_HALF_UP);
					listSampleSizesBD[i] = listSampleSizesBD[i].max(new BigDecimal(usedSampleSizes[i]));
					sumSampleSizes = sumSampleSizes.add(listSampleSizesBD[i]);
				}
			}
		}

		// convert BigDecimal array to BigInteger array
		// and then subtract the numers of used sample sizes
		BigInteger[] listSampleSizes = new BigInteger[numOfComponents];
		for(int i = 0; i < numOfComponents; i++) {
			listSampleSizes[i] = listSampleSizesBD[i].toBigInteger().subtract(usedSampleSizes[i]);
		}
		return listSampleSizes;
	}

    private static void log(Object aObject, int level){
    	Logger loggy = Logger.getInstance();
    	loggy.log(String.valueOf(aObject),level);
    }
    private static void log(Object[] array, String name, int level){
    	Logger loggy = Logger.getInstance();
    	loggy.log(array,name,level);
    }
}
